#include <stdlib.h>
#include <stdbool.h>

#include "auth.h"
#include "eccrsh.h"
#include "eccrsh-internals.h"
#include "crypto.h"
#include "ecc.h"

typedef struct {
	bignum256modm ALIGN(16) c;
	bignum256modm ALIGN(16) r;
} cr;

static void auth_contractCr(unsigned char* buffer, const cr* cr) {
	contract256_modm(buffer     , cr->c);
	contract256_modm(buffer + 32, cr->r);
}

static void auth_expandCr(cr* cr, const unsigned char* buffer) {
	expand256_modm(cr->c, buffer     , 32);
	expand256_modm(cr->r, buffer + 32, 32);
}

static void auth_calcT(const uint8_t basepoint_table[256][96], ge25519* T, const cr* cr, const ge25519* pk) {
	ge25519 ALIGN(16) temp;
	ge25519_scalarmult(&temp, pk, cr->c);
	ge25519_scalarmult_base_niels(T, basepoint_table, cr->r);
	ge25519_add(T, T, &temp);
}

static bool auth_hashC(const ge25519* pk1, const ge25519* pk2, const ge25519* pk3,
	const ge25519* T1, const ge25519* T2, const ge25519* T3,
	const unsigned char* message, size_t msgLen, const unsigned char* associatedData, size_t adLen,
	const unsigned char* implHashTag, size_t ihtLen,
	unsigned char* cBuffer) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));
	Keccak_Must(Keccak_HashUpdate(&hasher, implHashTag, ihtLen*8));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, pk1));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, pk2));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, pk3));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, T1));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, T2));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, T3));
	Keccak_Must(Keccak_HashUpdate(&hasher, message, msgLen*8));
	if (adLen > 0) {
		Keccak_Must(Keccak_HashUpdate(&hasher, associatedData, adLen*8));
	}
	Keccak_Must(Keccak_HashFinal(&hasher, cBuffer));
	return true;
}

bool auth_prove(const uint8_t basepoint_table[256][96],
	const ge25519* pk1, const ge25519* pk2, const ge25519* pk3,
	unsigned char knownKey, const bignum256modm sk,
	const unsigned char* message, size_t msgLen, const unsigned char* associatedData, size_t adLen,
	const unsigned char* implHashTag, size_t ihtLen,
	unsigned char* proof) {

	// Notation:
	//   "1", "2", and "3" refer to the key data slots by public position.
	//   "x" refers to our slot, "y" to the first other slot, and "z" to the second other slot.

	// Create selection bits
	unsigned char select1 = (knownKey & AUTH_KEY1);
	unsigned char notSelect1 = 1 - select1;
	unsigned char select2 = (knownKey & AUTH_KEY2) >> 1;
	unsigned char select3 = (knownKey & AUTH_KEY3) >> 2;
	unsigned char notSelect3 = 1 - select3;

	// Generate challenge/response pairs
	cr crx, cry, crz;
	modm_rand(cry.c);
	modm_rand(cry.r);
	modm_rand(crz.c);
	modm_rand(crz.r);

	bignum256modm ALIGN(16) tx;
	modm_rand(tx);

	ge25519 ALIGN(16) Tx, Ty, Tz;
	ge25519_scalarmult_base_niels(&Tx, basepoint_table, tx);

	// Conditional move into Ay, Az
	//   If key 1: Ay = key 2, Az = key 3
	//   If key 2: Ay = key 1, Az = key 3
	//   If key 3: Ay = key 1, Az = key 2
	ge25519 Ay, Az;
	cond_move(&Ay, pk2, sizeof(ge25519), select1);
	cond_move(&Ay, pk1, sizeof(ge25519), notSelect1);
	cond_move(&Az, pk3, sizeof(ge25519), notSelect3);
	cond_move(&Az, pk2, sizeof(ge25519), select3);

	// Compute Ty, Tz
	auth_calcT(basepoint_table, &Ty, &cry, &Ay);
	auth_calcT(basepoint_table, &Tz, &crz, &Az);

	// Conditional move into T1, T2, T3
	//   If key 1: T1 = Tx, T2 = Ty, T3 = Tz
	//   If key 2: T1 = Ty, T2 = Tx, T3 = Tz
	//   If key 3: T1 = Ty, T2 = Tz, T3 = Tx
	ge25519 ALIGN(16) T1, T2, T3;
	cond_move(&T1, &Tx, sizeof(ge25519), select1);
	cond_move(&T1, &Ty, sizeof(ge25519), notSelect1);
	cond_move(&T2, &Ty, sizeof(ge25519), select1);
	cond_move(&T2, &Tx, sizeof(ge25519), select2);
	cond_move(&T2, &Tz, sizeof(ge25519), select3);
	cond_move(&T3, &Tz, sizeof(ge25519), notSelect3);
	cond_move(&T3, &Tx, sizeof(ge25519), select3);

	// Compute c
	bignum256modm ALIGN(16) c;
	unsigned char cBuffer[32];
	if (!auth_hashC(pk1, pk2, pk3, &T1, &T2, &T3, message, msgLen, associatedData, adLen, implHashTag, ihtLen, cBuffer)) return false;
	expand256_modm(c, cBuffer, 32);

	// Compute crx.c, crx.r
	add256_modm(crx.c, cry.c, crz.c);
	modm_sub(crx.c, c, crx.c);
	mul256_modm(crx.r, crx.c, sk);
	modm_sub(crx.r, tx, crx.r);

	// Conditional swap into cr1, cr2, cr3
	//   If key 1: cr1 = crx, cr2 = cry, cr3 = crz
	//   If key 2: cr1 = cry, cr2 = crx, cr3 = crz
	//   If key 3: cr1 = cry, cr2 = crz, cr3 = crx
	cr cr1, cr2, cr3;
	cond_move(&cr1, &crx, sizeof(cr), select1);
	cond_move(&cr1, &cry, sizeof(cr), notSelect1);
	cond_move(&cr2, &cry, sizeof(cr), select1);
	cond_move(&cr2, &crx, sizeof(cr), select2);
	cond_move(&cr2, &crz, sizeof(cr), select3);
	cond_move(&cr3, &crz, sizeof(cr), notSelect3);
	cond_move(&cr3, &crx, sizeof(cr), select3);

	auth_contractCr(proof      , &cr1);
	auth_contractCr(proof + 64 , &cr2);
	auth_contractCr(proof + 128, &cr3);
	return true;
}

bool auth_verify(const uint8_t basepoint_table[256][96],
	const ge25519* pk1, const ge25519* pk2, const ge25519* pk3,
	const unsigned char* proof,
	const unsigned char* message, size_t msgLen, const unsigned char* associatedData, size_t adLen,
	const unsigned char* implHashTag, size_t ihtLen) {

	cr cr1, cr2, cr3;
	auth_expandCr(&cr1, proof      );
	auth_expandCr(&cr2, proof + 64 );
	auth_expandCr(&cr3, proof + 128);

	ge25519 ALIGN(16) T1, T2, T3;
	auth_calcT(basepoint_table, &T1, &cr1, pk1);
	auth_calcT(basepoint_table, &T2, &cr2, pk2);
	auth_calcT(basepoint_table, &T3, &cr3, pk3);

	bignum256modm ALIGN(16) c;
	unsigned char cBuffer[32];
	if (!auth_hashC(pk1, pk2, pk3, &T1, &T2, &T3, message, msgLen, associatedData, adLen, implHashTag, ihtLen, cBuffer)) return false;
	expand256_modm(c, cBuffer, 32);

	bignum256modm ALIGN(16) cTest;
	add256_modm(cTest, cr1.c, cr2.c);
	add256_modm(cTest, cTest, cr3.c);

	return modm_equals(c, cTest) != 0;
}
