#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include "dakez.h"
#include "crypto.h"
#include "ecc.h"
#include "secmem.h"
#include "auth.h"

typedef struct {
	ge25519 gi;
	// pqI immediately follows
} dakez_publicState_I;

typedef struct {
	bignum256modm i;
	// sqI immediately follows
} dakez_secretState_I;

typedef struct {
	ge25519 gi;
	ge25519 gr;
	// pqI immediately follows
	// qR immediately follows pqI
} dakez_publicState_R;

typedef struct {
	bignum256modm r; // Only stored in offline forgeries
	unsigned char key[32];
} dakez_secretState_R;

static unsigned char implHashTag[] = "DAKEZ key exchange";
static size_t ihtLen = 19;

DAKEZ_STATUS dakez_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk,
	unsigned char* scratch) {

	bignum256modm_element_t* mSk = (bignum256modm_element_t*)scratch;

	ge25519 ALIGN(16) gePk, geFpk;

	RAND_bytes(sk, 32);
	curved25519_make_secret(mSk, sk);
	ge25519_scalarmult_base1(&geFpk, mSk);
	ge25519_pack(forgingPk, &geFpk);
	if (forgingSk != NULL) memcpy(forgingSk, sk, 32);

	RAND_bytes(sk, 32);
	curved25519_make_secret(mSk, sk);
	ge25519_scalarmult_base1(&gePk, mSk);
	ge25519_pack(pk, &gePk);

	secureErase(scratch, sizeof(bignum256modm));

	return DAKEZ_SUCCESS;
}

DAKEZ_STATUS dakez_1_initiate_I(const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* send,
	unsigned char* scratch) {

	unsigned char* sendIdI = send;
	unsigned char* sendGi = send + idLen;
	unsigned char* sendPqI = sendGi + 32;

	dakez_publicState_I* publicStateEl = (dakez_publicState_I*)publicState;
	dakez_secretState_I* secretStateEl = (dakez_secretState_I*)secretState;
	unsigned char* pqI = publicState + sizeof(dakez_publicState_I);
	unsigned char* sqI = secretState + sizeof(dakez_secretState_I) + qrKex->sizeAlign8;
	if (!qrKex->qrGenI(pqI, sqI, scratch)) return DAKEZ_QRCALC_FAILURE;
	memcpy(sendPqI, pqI, qrKex->sizePqI);

	memcpy(sendIdI, idI, idLen);

	modm_rand(secretStateEl->i);

	ge25519_scalarmult_base1(&publicStateEl->gi, secretStateEl->i);
	ge25519_pack(sendGi, &publicStateEl->gi);

	return DAKEZ_SUCCESS;
}

DAKEZ_STATUS dakez_2_receive_R(const unsigned char* received, size_t idLen, unsigned char* idI) {
	memcpy(idI, received, idLen);
	return DAKEZ_SUCCESS;
}

// Scratch space required: sizeof(ge25519) + 32
static bool dakez_key(const ge25519* dhPub, const bignum256modm dhSecret,
	const qrKex* qrKex, const unsigned char* qrKey,
	unsigned char* key,
	unsigned char* scratch) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	ge25519* dhKey = (ge25519*)scratch;
	ge25519_scalarmult(dhKey, dhPub, dhSecret);
	unsigned char* dhKeyPack = scratch + sizeof(ge25519);
	ge25519_pack(dhKeyPack, dhKey);

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, dhKeyPack, 32*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, qrKey, qrKex->sizeKey*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, key) == SUCCESS);

	secureErase(scratch, sizeof(ge25519) + 32);

	return hashSuccess;
}

// Must add 2*idLen + qrKex->sizePqI + qrKex->sizeQR
#define DAKEZ_MSG_OVERHEAD (2*32+1)

#define DAKEZ_TAG_MSG_R 0x42
#define DAKEZ_TAG_MSG_I 0xBD

static void dakez_makeMsg(unsigned char direction,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const ge25519* gi, const ge25519* gr,
	const qrKex* qrKex,
	const unsigned char* pqI, const unsigned char* qR,
	unsigned char* msgR) {

	unsigned char* ptr = msgR;
	*ptr++ = direction;
	memcpy(ptr, idI, idLen);          ptr += idLen;
	memcpy(ptr, idR, idLen);          ptr += idLen;
	ge25519_pack(ptr, gi);            ptr += 32;
	ge25519_pack(ptr, gr);            ptr += 32;
	memcpy(ptr, pqI, qrKex->sizePqI); ptr += qrKex->sizePqI;
	memcpy(ptr, qR, qrKex->sizeQR);   ptr += qrKex->sizeQR;
}

DAKEZ_STATUS dakez_2_respond_R(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idR,
	const unsigned char* forgingPkI, const unsigned char* pkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* send,
	unsigned char* scratch) {

	const unsigned char* receivedIdI = received;
	const unsigned char* receivedGi = received + idLen;
	const unsigned char* receivedPqI = receivedGi + 32;

	unsigned char* sendIdR = send;
	unsigned char* sendGr = sendIdR + idLen;
	unsigned char* sendQR = sendGr + 32;
	unsigned char* sendProof = sendQR + qrKex->sizeQR;

	unsigned char ALIGN(16) qrBuffer[qrKex->sizeQR];

	bignum256modm_element_t* mr = (bignum256modm_element_t*)scratch;
	bignum256modm_element_t* signKey = (bignum256modm_element_t*)(scratch + sizeof(bignum256modm));
	unsigned char* scratchKey = scratch + 2*sizeof(bignum256modm);
	unsigned char* qrKey = scratchKey + sizeof(ge25519) + 32;
	unsigned char* qrScratch = qrKey + qrKex->sizeKey;

	dakez_publicState_R* publicStateEl = (dakez_publicState_R*)publicState;
	dakez_secretState_R* secretStateEl = (dakez_secretState_R*)secretState;

	ge25519 ALIGN(16) geFpkI, gePkR;
	if (!ge25519_unpack_vartime(&geFpkI, forgingPkI)) return DAKEZ_INVALID_KEY;
	if (!ge25519_unpack_vartime(&gePkR, pkR)) return DAKEZ_INVALID_KEY;

	memcpy(sendIdR, idR, idLen);

	if (!qrKex->qrGenR(receivedPqI, qrBuffer, qrKey, qrScratch)) return DAKEZ_QRCALC_FAILURE;
	memcpy(sendQR, qrBuffer, qrKex->sizeQR);

	if (!ge25519_unpack_vartime(&publicStateEl->gi, receivedGi)) {
		secureErase(qrKey, qrKex->sizeKey);
		secureErase(qrScratch, qrKex->sizeScratch);
		return DAKEZ_INVALID_MESSAGE;
	}

	modm_rand(mr);

	ge25519_scalarmult_base1(&publicStateEl->gr, mr);
	ge25519_pack(sendGr, &publicStateEl->gr);

	bool success = dakez_key(&publicStateEl->gi, mr, qrKex, qrKey, secretStateEl->key, scratchKey);

	if (success) {
		size_t msgLen = DAKEZ_MSG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
		unsigned char msg[msgLen];
		dakez_makeMsg(DAKEZ_TAG_MSG_R, idLen, receivedIdI, idR, &publicStateEl->gi, &publicStateEl->gr, qrKex, receivedPqI, sendQR, msg);

		expand256_modm(signKey, signSecret, 32);
		success = auth_prove(ge25519_niels_base1_multiples, &geFpkI, &gePkR, &publicStateEl->gi, signMode, signKey, msg, msgLen, associatedData, adLen, implHashTag, ihtLen, sendProof);
	}

	if (success && signMode == DAKEZ_OFFLINE_FORGE) {
		memcpy(secretStateEl->r, mr, sizeof(bignum256modm));
	}

	secureErase(scratch, qrKex->sizeKey + qrKex->sizeScratch + DAKEZ_SCRATCH_R_OVERHEAD);

	if (!success) {
		secureErase(secretState, sizeof(dakez_secretState_R));
		return DAKEZ_CALC_FAILURE;
	}

	unsigned char* publicStatePqI = publicState + sizeof(dakez_publicState_R);
	unsigned char* publicStateQR = publicStatePqI + qrKex->sizePqI;
	memcpy(publicStatePqI, receivedPqI, qrKex->sizePqI);
	memcpy(publicStateQR, sendQR, qrKex->sizeQR);

	return DAKEZ_SUCCESS;
}

DAKEZ_STATUS dakez_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
	return DAKEZ_SUCCESS;
}

DAKEZ_STATUS dakez_3_respond_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI,
	const unsigned char* pkR, const unsigned char* forgingPkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch) {

	DAKEZ_STATUS result = DAKEZ_SUCCESS;

	const unsigned char* receivedIdR = received;
	const unsigned char* receivedGr = receivedIdR + idLen;
	const unsigned char* receivedQR = receivedGr + 32;
	const unsigned char* receivedProof = receivedQR + qrKex->sizeQR;

	unsigned char* keyScratch = scratch;
	unsigned char* qrKey = keyScratch + sizeof(ge25519) + 32;
	unsigned char* qrScratch = qrKey + qrKex->sizeKey;
	bignum256modm_element_t* signKey = (bignum256modm_element_t*)scratch; // Intentional overlap

	ge25519 ALIGN(16) gr;
	if (!ge25519_unpack_vartime(&gr, receivedGr)) {
		result = DAKEZ_INVALID_MESSAGE;
		goto cleanup;
	}

	unsigned char* pqI = ((unsigned char*)publicState) + sizeof(dakez_publicState_I);

	ge25519 ALIGN(16) gePkI, geFpkI, gePkR, geFpkR;
	bool success = true;
	success = success && ge25519_unpack_vartime(&gePkI, pkI);
	success = success && ge25519_unpack_vartime(&geFpkI, forgingPkI);
	success = success && ge25519_unpack_vartime(&gePkR, pkR);
	success = success && ge25519_unpack_vartime(&geFpkR, forgingPkR);
	if (!success) {
		result = DAKEZ_INVALID_KEY;
		goto cleanup;
	}

	// Using scratch space
	{
		size_t msgLen = DAKEZ_MSG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
		unsigned char msg[msgLen];

		dakez_publicState_I* publicStateEl = (dakez_publicState_I*)publicState;
		dakez_secretState_I* secretStateEl = (dakez_secretState_I*)secretState;

		dakez_makeMsg(DAKEZ_TAG_MSG_R, idLen, idI, receivedIdR, &publicStateEl->gi, &gr, qrKex, pqI, receivedQR, msg);
		if (!auth_verify(ge25519_niels_base1_multiples, &geFpkI, &gePkR, &publicStateEl->gi, receivedProof, msg, msgLen, associatedData, adLen, implHashTag, ihtLen)) {
			result = DAKEZ_INVALID_MESSAGE;
			goto scratchCleanup;
		}

		unsigned char* sqI = secretState + sizeof(dakez_secretState_I) + qrKex->sizeAlign8;
		if (!qrKex->qrKeyI(sqI, receivedQR, qrKey, qrScratch)) {
			result = DAKEZ_QRCALC_FAILURE;
			goto scratchCleanup;
		}

		if (!dakez_key(&gr, secretStateEl->i, qrKex, qrKey, key, keyScratch)) {
			result = DAKEZ_CALC_FAILURE;
			goto scratchCleanup;
		}

		msg[0] = DAKEZ_TAG_MSG_I;
		expand256_modm(signKey, signSecret, 32);
		if (!auth_prove(ge25519_niels_base1_multiples, &gePkI, &geFpkR, &gr, signMode, signKey, msg, msgLen, associatedData, adLen, implHashTag, ihtLen, send)) {
			result = DAKEZ_CALC_FAILURE;
			// Intentional fall-through to erase signKey
		}
scratchCleanup:
		secureErase(scratch, qrKex->sizeKey + sizeof(ge25519) + 32 + qrKex->sizeScratch);
	}

cleanup:
	secureErase(secretState, DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI);
	return result;
}

DAKEZ_STATUS dakez_4_verify_R(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkR,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* key) {

	DAKEZ_STATUS result = DAKEZ_SUCCESS;

	dakez_publicState_R* publicStateEl = (dakez_publicState_R*)publicState;
	dakez_secretState_R* secretStateEl = (dakez_secretState_R*)secretState;

	const unsigned char* pqI = publicState + sizeof(dakez_publicState_R);
	const unsigned char* qR = pqI + qrKex->sizePqI;

	ge25519 ALIGN(16) gePkI, geFpkR;
	bool success = true;
	success = success && ge25519_unpack_vartime(&gePkI, pkI);
	success = success && ge25519_unpack_vartime(&geFpkR, forgingPkR);
	if (!success) {
		result = DAKEZ_INVALID_KEY;
		goto cleanup;
	}

	{
		size_t msgLen = DAKEZ_MSG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
		unsigned char msg[msgLen];

		dakez_makeMsg(DAKEZ_TAG_MSG_I, idLen, idI, idR, &publicStateEl->gi, &publicStateEl->gr, qrKex, pqI, qR, msg);
		if (!auth_verify(ge25519_niels_base1_multiples, &gePkI, &geFpkR, &publicStateEl->gr, received, msg, msgLen, associatedData, adLen, implHashTag, ihtLen)) {
			result = DAKEZ_INVALID_MESSAGE;
			goto cleanup;
		}

		memcpy(key, secretStateEl->key, DAKEZ_KEY_SIZE);
	}

cleanup:
	secureErase(secretState, DAKEZ_SECRETSTATE_R_SIZE);
	return result;
}

void dakez_forgeOffline1(const unsigned char* secretStateI, unsigned char* signSecretR) {
	contract256_modm(signSecretR, ((const dakez_secretState_I*)secretStateI)->i);
}

void dakez_forgeOffline2(const unsigned char* secretStateR, unsigned char* signSecretI) {
	contract256_modm(signSecretI, ((const dakez_secretState_R*)secretStateR)->r);
}
