#pragma once

/*
 * This module implements the DAKEZ key exchange.
 *
 * A key exchange takes place between an initiator, I, and a responder R.
 *
 * dakez_genLongTermKeys is used to generate long-term key pairs. It is assumed
 * that parties distribute long-term public keys to each other and authenticate
 * them somehow. Authentication *must* cover both the "normal" public keys and
 * the "forging" public keys. This authentication is beyond the scope of this
 * module. Parties are given caller-defined identifiers that are used to locate
 * matching long-term public keys. To simplify key distribution, callers may
 * choose to use long-term public keys and forging public keys AS identifiers,
 * rather than as table lookup indexes, at the cost of communication overhead.
 *
 * The "forging" keys produced by dakez_genLongTermKeys can be used to forge
 * key exchanges with others when being coerced by a third party to engage in a
 * session (see below for details about how to perform these forgeries).
 * However, if a malicious party ever compromises a long-term forging secret
 * key, then that party can impersonate others to the owner of the key without
 * detection. To prevent this attack, passing forgingSk = NULL to
 * dakez_genLongTermKeys will prevent this key from ever being generated, at the
 * cost of making the user vulnerable to coercion. In the event that the forging
 * key IS generated, it should be stored more securely than the "normal"
 * long-term secret key (e.g., offline, on paper, in a secure storage facility).
 *
 * Given known long-term public keys, a normal exchange looks like this:
 *   1) I calls dakez_1_initiate_I and sends the message to R.
 *   2) R calls dakez_2_receive_R to learn I's asserted identifier.
 *   3) R looks up I's asserted long-term public key using some external system.
 *   4) R calls dakez_2_respond_R and sends the message to I.
 *   5) I calls dakez_3_receive_I to learn R's asserted identifier.
 *   6) I looks up R's asserted long-term public key using some external system.
 *   7) I calls dakez_3_respond_I. I sends the resulting message to R. I now
 *      trusts R's identity and it knows the shared session key.
 *   8) R calls dakez_4_verify_R. R now trusts I's identity and it knows the
 *      shared session key.
 *
 * dakez_2_respond_R and dakez_3_respond_I involve the production of
 * authenticating proofs. These functions take two parameters---"signMode" and
 * "signSecret"---that specify keys to use to produce these proofs. In the
 * normal situation (where both I and R are honest and truly are who they say
 * they are), the following values are used:
 *   - R passes signMode = DAKEZ_HONEST_R and signSecret = skR to
 *     dakez_2_respond_R.
 *   - I passes signMode = DAKEZ_HONEST_I and signSecret = skI to
 *     dakez_3_respond_I.
 *
 * For plausible deniability purposes, this module provides the ability to forge
 * communication transcripts. An "offline forger" (someone who knows neither skI
 * nor skR) can produce a false transcript in the following way:
 *   1) Call dakez_1_initiate_I to produce the first message.
 *   2) Call dakez_forgeOffline1 to extract signSecretR.
 *   3) Call dakez_2_respond_R using DAKEZ_OFFLINE_FORGE and signSecretR to
 *      produce the second message.
 *   4) Call dakez_forgeOffline2 to extract signSecretI.
 *   5) Call dakez_3_respond_I using DAKEZ_OFFLINE_FORGE and signSecretI to
 *      produce the third message.
 *   6) The shared session key is returned by dakez_3_respond_I. It can be used
 *      to forge subsequent messages in higher-level protocols.
 *
 * There are also "online forger" cases. If I is being coerced by an external
 * entity J, I can forge an exchange with R (without knowing skR) in the
 * following way:
 *   1) J provides a message that I pretends to relay to R.
 *      (If J allows it, then I can call dakez_1_initiate_I normally instead.)
 *   2) I calls dakez_2_receive_R to ensure that the message is valid.
 *   3) I calls dakez_2_respond_R with signMode = DAKEZ_ONLINE_FORGE_R and
 *      signSecret = forgingSkI, then sends the message to J as if R sent it.
 *   4) J provides a message that I pretends to relay to R.
 *      (If J allows it, then I can call dakez_3_respond_I normally instead.)
 *   5) I calls dakez_4_verify_R to recover the shared session key. It can be
 *      used to forge subsequent messages in higher-level protocols.
 *
 * If R is being coerced by an external entity J, R can forge an exchange with I
 * (without knowing skI) in the following way:
 *   1) R calls dakez_1_initiate_I and sends the message to J as if I sent it.
 *   2) J provides a message that R pretends to relay to I.
 *      (If J allows it, then R can call dakez_2_respond_R normally instead.)
 *   3) R calls dakez_3_receive_I to ensure that the message is valid.
 *   4) R calls dakez_3_respond_I with signMode = DAKEZ_ONLINE_FORGE_I and
 *      signSecret = forgingSkR, then sends the message to J as if I sent it.
 *      The function also returns the shared session key. It can be used to
 *      forge subsequent messages in higher-level protocols.
 */

#include <stdlib.h>

#include "qr.h"

#define DAKEZ_PK_SIZE 32
#define DAKEZ_SK_SIZE 32
#define DAKEZ_1_SEND_I_OVERHEAD 32
#define DAKEZ_2_SEND_R_OVERHEAD 224
#define DAKEZ_3_SEND_I_SIZE 192
#define DAKEZ_PUBLICSTATE_I_OVERHEAD 192
#define DAKEZ_SECRETSTATE_I_OVERHEAD 40
#define DAKEZ_PUBLICSTATE_R_OVERHEAD 384
#define DAKEZ_SECRETSTATE_R_SIZE 72
#define DAKEZ_SCRATCH_KEYGEN_SIZE 40
#define DAKEZ_SCRATCH_R_OVERHEAD 304
#define DAKEZ_SCRATCH_I_RESPOND_OVERHEAD 224
#define DAKEZ_KEY_SIZE 32

typedef enum {
	DAKEZ_SUCCESS,
	DAKEZ_CALC_FAILURE,
	DAKEZ_QRCALC_FAILURE,
	DAKEZ_INVALID_MESSAGE,
	DAKEZ_INVALID_KEY,
} DAKEZ_STATUS;

#define DAKEZ_HONEST_I 0x1
#define DAKEZ_HONEST_R 0x2
#define DAKEZ_OFFLINE_FORGE 0x4
#define DAKEZ_ONLINE_FORGE_R 0x1
#define DAKEZ_ONLINE_FORGE_I 0x2

/* Generates long-term key pairs. pk and forgingPk are public, while sk and
 * forgingSk are secret.
 * - pk and forgingPk must be DAKEZ_PK_SIZE.
 * - sk and forgingSk must be DAKEZ_SK_SIZE (but forgingSk may be NULL).
 * - If forgingSk is NULL, then the secret forging key is securely erased. See
 *   the comment at the top of the file for a discussion of the consequences.
 * - scratch is secure scratch space of length DAKEZ_SCRATCH_KEYGEN_SIZE.
 * - Callers should arrange for secure storage of sk, forgingSk, and scratch
 *   (e.g., using secmem.h).
 */
DAKEZ_STATUS dakez_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk,
	unsigned char* scratch);

/* Initiates a DAKEZ key exchange. Called by I (initiator).
 * - qrKex is the quantum-resistant key exchange to use.
 * - idI holds a caller-defined identifier of length idLen for I.
 * - A caller can use long-term public keys as identifiers to simplify key
 *   distribution.
 * - publicState holds temporary public state until R responds. It must point to
 *   a buffer of length DAKEZ_PUBLICSTATE_I_OVERHEAD + qrKex->sizePqI.
 *   publicState should be aligned on a 16 byte boundary.
 * - secretState holds temporary secret state until R responds. It must point to
 *   a buffer of length
 *   DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI.
 *   secretState should be aligned on a 16 byte boundary.
 * - send holds the message to transmit to R and must be of length
 *   DAKEZ_1_SEND_I_OVERHEAD + qrKex->sizePqI + idLen.
 * - scratch is secure scratch space of length qrKex->sizeScratch.
 * - Callers should arrange for secure storage of secretState and scratch
 *   (e.g., using secmem.h).
 */
DAKEZ_STATUS dakez_1_initiate_I(const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* send,
	unsigned char* scratch);

/* Called by R (responder) after receiving an initiation message from I.
 * - This function returns I's asserted identifier in idI, which must be of
 *   length idLen.
 * - Call this before calling dakez_2_respond_R in order to decide whether or
 *   not to accept the connection.
 * - Note that idI is *not* trusted until the key exchange has completed with
 *   dakez_4_verify_R; it should only be used to locate the associated long-term
 *   public key for I.
 */
DAKEZ_STATUS dakez_2_receive_R(const unsigned char* received, size_t idLen, unsigned char* idI);

/* Called by R to respond to an initiation request from I.
 * - qrKex is the quantum-resistant key exchange in use.
 * - R's identifier of length idLen is given in idR.
 * - forgingPkI and pkR are long-term keys for the parties.
 * - signMode and signSecret specify the key pair used for authentication; see
 *   the comment at the top of the file for details.
 * - associatedData, of length adLen, is any common protocol state known to both
 *   parties in the higher-level protocol. If the key exchange completes
 *   successfully after dakez_4_verify_R, then it is guaranteed that both
 *   parties passed the same value to dakez_2_respond_R and dakez_3_respond_I.
 * - publicState holds temporary public state until I responds. It must point to
 *   a buffer of length
 *   DAKEZ_PUBLICSTATE_R_OVERHEAD + qrKex->sizePqI + qrKex->sizeQR. publicState
 *   should be aligned on a 16 byte boundary.
 * - secretState holds temporary secret state until I responds. It must point to
 *   a buffer of length DAKEZ_SECRETSTATE_R_SIZE. secretState should be aligned
 *   on a 16 byte boundary.
 * - send holds the message to transmit to I and must be of length
 *   DAKEZ_2_SEND_R_OVERHEAD + qrKex->sizeQR + idLen.
 * - scratch is secure scratch space of length
 *   DAKEZ_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch, and must be
 *   aligned on a 16 byte boundary.
 * - Callers should arrange for secure storage of secretState and scratch (e.g.,
 *   using secmem.h).
 */
DAKEZ_STATUS dakez_2_respond_R(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idR,
	const unsigned char* forgingPkI, const unsigned char* pkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* send,
	unsigned char* scratch);

/* Called by I after receiving a response message from R.
 * - This function returns R's asserted identifier in idR, which must be of
 *   length idLen.
 * - Call this before calling dakez_3_respond_I in order to decide whether or
 *   not to complete the connection.
 * - Note that idR is *not* trusted until the key exchange has completed with
 *   dakez_3_respond_I; it should only be used to locate the associated
 *   long-term public key for R.
 */
DAKEZ_STATUS dakez_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR);

/* Called by I to respond to a response from R, completing the exchange for I.
 * - qrKex is the quantum-resistant key exchange in use.
 * - I's identifier of length idLen is given in idI.
 * - pkI, forgingPkI, pkR, and forgingPkR are long-term keys for the parties.
 * - signMode and signSecret specify the key pair used for authentication; see
 *   the comment at the top of the file for details.
 * - associatedData, of length adLen, is any common protocol state known to both
 *   parties in the higher-level protocol. If this function returns success,
 *   then it is guaranteed that R passed the same value to dakez_2_respond_R.
 * - publicState holds the temporary public state allocated by
 *   dakez_1_initiate_I.
 * - secretState holds the temporary secret state allocated by
 *   dakez_1_initiate_I. It is securely erased after this function returns (even
 *   if it returns an error).
 * - send holds the message to transmit to I and must be DAKEZ_3_SEND_I_SIZE.
 * - key holds the shared secret for the session and is of size DAKEZ_KEY_SIZE.
 *   It may be used directly or used to derive multiple keys for future
 *   communications. There is no explicit confirmation that R derived the same
 *   secret.
 * - scratch is secure scratch space of length
 *   DAKEZ_SCRATCH_I_RESPOND_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch, and
 *   must be aligned on a 16 byte boundary.
 * - Callers should arrange for secure storage of key and scratch
 *   (e.g., using secmem.h).
 */
DAKEZ_STATUS dakez_3_respond_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI,
	const unsigned char* pkR, const unsigned char* forgingPkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch);

/* Called by R after receiving a response message from I.
 * - qrKex is the quantum-resistant key exchange in use.
 * - The identifiers of I and R, each of length idLen, are given in idI and idR.
 * - pkI and forgingPkR are long-term keys for the parties.
 * - associatedData, of length adLen, should match the data passed to
 *   dakez_2_respond_R. If this function returns success, then it is guaranteed
 *   that I passed the same value to dakez_3_respond_I.
 * - secretState holds the temporary secret state allocated by
 *   dakez_2_respond_R. It is securely erased after this function returns (even
 *   if it returns an error).
 * - key holds the shared secret for the session and is of size DAKEZ_KEY_SIZE.
 *   There is no explicit confirmation that I derived the same secret.
 * - Callers should arrange for secure storage of key (e.g., using secmem.h).
 */
DAKEZ_STATUS dakez_4_verify_R(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkR,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* key);

/* Extracts signing key used for an offline forgery of R.
 * - signSecretR must be of length DAKEZ_SK_SIZE.
 * - To avoid leaving traces of the forgery, callers should use the same
 *   precautions for signSecretR that they use for long-term secret keys
 *   (e.g., using secmem.h), and these values should be securely erased after
 *   the forgery.
 */
void dakez_forgeOffline1(const unsigned char* secretStateI, unsigned char* signSecretR);

/* Extracts signing key used for an offline forgery of I.
 * - signSecretI must be of length DAKEZ_SK_SIZE.
 * - To avoid leaving traces of the forgery, callers should use the same
 *   precautions for signSecretI that they use for long-term secret keys
 *   (e.g., using secmem.h), and these values should be securely erased after
 *   the forgery.
 */
void dakez_forgeOffline2(const unsigned char* secretStateR, unsigned char* signSecretI);

