#include <stdlib.h>
#include <stdbool.h>

#include "dre.h"
#include "ecelgamal.h"
#include "ecelgamal-internals.h"
#include "crypto.h"
#include "ecc.h"

// A unique tag input into the hash for this implementation. This tag MUST be changed if different
// generators are used in ecc-tables.inc! The actual contents of this string don't matter, as long
// as nothing else is likely to hash strings beginning with this tag. To reduce the chance of
// accidental collision with other protocols, it is null-terminated.
static unsigned char implHashTag[] = "DAKEZ implementation hash tag";
static size_t implHashTagLen = 30;

typedef struct {
	ge25519 ALIGN(16) T11, T12, T2;
} Ts;

static bool dre_hashUpdatePk(Keccak_HashInstance* hasher, const ecelgamal_pubkey* pk) {
	Keccak_Must(Keccak_HashUpdatePoint(hasher, &pk->h));
	return true;
}

static bool dre_hashUpdateTs(Keccak_HashInstance* hasher, const Ts* T) {
	Keccak_Must(Keccak_HashUpdatePoint(hasher, &T->T11));
	Keccak_Must(Keccak_HashUpdatePoint(hasher, &T->T12));
	Keccak_Must(Keccak_HashUpdatePoint(hasher, &T->T2));
	return true;
}

static bool dre_hashL(const ecelgamal_pubkey* pk1, const ecelgamal_pubkey* pk2,
		const unsigned char* ct12Buffer,
		const Ts* T,
		const unsigned char* associatedData, size_t adLen,
		unsigned char* LBuffer) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));
	Keccak_Must(Keccak_HashUpdate(&hasher, implHashTag, implHashTagLen*8));
	if (!dre_hashUpdatePk(&hasher, pk1)) return false;
	if (!dre_hashUpdatePk(&hasher, pk2)) return false;
	Keccak_Must(Keccak_HashUpdate(&hasher, ct12Buffer, 2*ECELGAMAL_CT_POINT_SIZE*8));
	if (!dre_hashUpdateTs(&hasher, T)) return false;
	if (adLen > 0) {
		Keccak_Must(Keccak_HashUpdate(&hasher, associatedData, adLen*8));
	}
	Keccak_Must(Keccak_HashFinal(&hasher, LBuffer));
	return true;
}

static bool dre_makeNIZKPK(const ecelgamal_pubkey* pk1, const ecelgamal_pubkey* pk2,
	const unsigned char* ct12Buffer,
	const bignum256modm k1, const bignum256modm k2,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* LBuffer, unsigned char* n1Buffer, unsigned char* n2Buffer) {

	bignum256modm ALIGN(16) t1, t2;
	modm_rand(t1);
	modm_rand(t2);

	Ts T;

	ge25519_scalarmult_base1(&T.T11, t1);
	ge25519_scalarmult_base1(&T.T12, t2);

	ge25519 ALIGN(16) temp;
	ge25519_scalarmult(&temp, &pk1->h, t1);
	ge25519_scalarmult(&T.T2, &pk2->h, t2);
	ge25519_neg(&T.T2, &T.T2);
	ge25519_add(&T.T2, &temp, &T.T2);

	if (!dre_hashL(pk1, pk2, ct12Buffer, &T, associatedData, adLen, LBuffer)) return false;

	bignum256modm ALIGN(16) L, n1, n2;
	curved25519_make_secret(L, LBuffer);

	mul256_modm(n1, L, k1);
	modm_sub(n1, t1, n1);
	mul256_modm(n2, L, k2);
	modm_sub(n2, t2, n2);

	contract256_modm(n1Buffer, n1);
	contract256_modm(n2Buffer, n2);

	return true;
}

static bool dre_verifyNIZKPK(const ecelgamal_pubkey* pk1, const ecelgamal_pubkey* pk2,
	const unsigned char* ct12Buffer,
	const ecelgamal_ciphertext* ct1, const ecelgamal_ciphertext* ct2,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* LBuffer, const unsigned char* n1Buffer, const unsigned char* n2Buffer) {

	bignum256modm ALIGN(16) L, n1, n2;
	unsigned char LTempBuffer[32];
	memcpy(LTempBuffer, LBuffer, 32);
	curved25519_make_secret(L, LTempBuffer);
	expand256_modm(n1, n1Buffer, 32);
	expand256_modm(n2, n2Buffer, 32);

	Ts T;
	ge25519 ALIGN(16) temp;

	// T_1j'
	ge25519_scalarmult_vartime(&temp, &ct1->c1, L);
	ge25519_scalarmult_base1(&T.T11, n1);
	ge25519_add(&T.T11, &T.T11, &temp);

	ge25519_scalarmult_vartime(&temp, &ct2->c1, L);
	ge25519_scalarmult_base1(&T.T12, n2);
	ge25519_add(&T.T12, &T.T12, &temp);

	// T_2'
	ge25519_scalarmult_vartime(&temp, &pk1->h, n1);
	ge25519_scalarmult_vartime(&T.T2, &pk2->h, n2);
	ge25519_neg(&T.T2, &T.T2);
	ge25519_add(&T.T2, &temp, &T.T2);
	ge25519_neg(&temp, &ct2->c2);
	ge25519_add(&temp, &ct1->c2, &temp);
	ge25519_scalarmult_vartime(&temp, &temp, L);
	ge25519_add(&T.T2, &T.T2, &temp);

	// Verification steps
	bignum256modm ALIGN(16) LPrime;
	if (!dre_hashL(pk1, pk2, ct12Buffer, &T, associatedData, adLen, LTempBuffer)) return false;
	curved25519_make_secret(LPrime, LTempBuffer);
	return memcmp(L, LPrime, 32) == 0;
}

bool dre_encrypt(const ecelgamal_pubkey* pk1, const ecelgamal_pubkey* pk2,
	const unsigned char* plaintext, size_t ptLen, const unsigned char* associatedData, size_t adLen,
	unsigned char* ciphertext, size_t ctMax, size_t* ctUsed) {

	*ctUsed = DRE_CT_OVERHEAD + ptLen;
	if (ctMax < *ctUsed) return false;

	ge25519 ALIGN(16) KappaPoint;
	ge25519_rand(&KappaPoint);
	unsigned char Kappa[32];
	ge25519_pack(Kappa, &KappaPoint);
	SHA3_256(Kappa, Kappa, 32);

	ecelgamal_ciphertext ct1, ct2;
	bignum256modm ALIGN(16) k1, k2;
	if (!ecelgamal_encryptRaw(pk1, &KappaPoint, &ct1, k1)) return false;
	if (!ecelgamal_encryptRaw(pk2, &KappaPoint, &ct2, k2)) return false;

	unsigned char* ct12Buffer = ciphertext;
	unsigned char* ct1Buffer = ct12Buffer;
	unsigned char* ct2Buffer = ct1Buffer + ECELGAMAL_CT_POINT_SIZE;
	unsigned char* tagBuffer = ct2Buffer + ECELGAMAL_CT_POINT_SIZE;
	unsigned char* LBuffer = tagBuffer + AEAD_TAG_SIZE;
	unsigned char* n1Buffer = LBuffer + 32;
	unsigned char* n2Buffer = n1Buffer + 32;
	unsigned char* aeadCtBuffer = n2Buffer + 32;

	unsigned char* nonce = LBuffer; // reuse L as a nonce

	ecelgamal_packCiphertext(ct1Buffer, &ct1);
	ecelgamal_packCiphertext(ct2Buffer, &ct2);

	if (!dre_makeNIZKPK(pk1, pk2, ct12Buffer, k1, k2, associatedData, adLen, LBuffer, n1Buffer, n2Buffer)) return false;

	if (!aead_encrypt(aeadCtBuffer, tagBuffer, plaintext, ptLen, Kappa, associatedData, adLen, nonce)) return false;

	return true;
}

bool dre_decrypt(const ecelgamal_pubkey* pk1, const ecelgamal_pubkey* pk2, unsigned char knownKey, const ecelgamal_seckey* sk,
	const unsigned char* ciphertext, size_t ctLen, const unsigned char* associatedData, size_t adLen,
	unsigned char* plaintext, size_t ptMax, size_t* ptUsed) {

	if (ctLen < DRE_CT_OVERHEAD) return false;
	*ptUsed = ctLen - DRE_CT_OVERHEAD;
	if (*ptUsed > ptMax) return false;

	const unsigned char* ct12Buffer = ciphertext;
	const unsigned char* ct1Buffer = ct12Buffer;
	const unsigned char* ct2Buffer = ct1Buffer + ECELGAMAL_CT_POINT_SIZE;
	const unsigned char* tagBuffer = ct2Buffer + ECELGAMAL_CT_POINT_SIZE;
	const unsigned char* LBuffer = tagBuffer + AEAD_TAG_SIZE;
	const unsigned char* n1Buffer = LBuffer + 32;
	const unsigned char* n2Buffer = n1Buffer + 32;
	const unsigned char* aeadCtBuffer = n2Buffer + 32;

	const unsigned char* nonce = LBuffer; // reuse L as a nonce

	ecelgamal_ciphertext ct1, ct2, myct;
	ecelgamal_unpackCiphertext(&ct1, ct1Buffer);
	ecelgamal_unpackCiphertext(&ct2, ct2Buffer);
	cond_move(&myct, &ct1, sizeof(ecelgamal_ciphertext), knownKey & DRE_KEY1);
	cond_move(&myct, &ct2, sizeof(ecelgamal_ciphertext), (knownKey & DRE_KEY2) >> 1);

	if (!dre_verifyNIZKPK(pk1, pk2, ct12Buffer, &ct1, &ct2, associatedData, adLen, LBuffer, n1Buffer, n2Buffer)) return false;

	ge25519 ALIGN(16) KappaPoint;
	if (!ecelgamal_decryptRaw(sk, &myct, &KappaPoint)) return false;

	unsigned char Kappa[32];
	ge25519_pack(Kappa, &KappaPoint);
	SHA3_256(Kappa, Kappa, 32);

	if (!aead_decrypt(plaintext, aeadCtBuffer, *ptUsed, Kappa, tagBuffer, associatedData, adLen, nonce)) return false;

	return true;
}
