#include <stdbool.h>

#include "ecc.h"

#include "crypto.h"

#include "ecc-tables.inc"

#define bignum256modm_reduce_mask (((uint64_t)1 << bignum256modm_bits_per_limb) - 1)
#define bignum256modm_used_bits 256
#define bignum256modm_storage_bits (bignum256modm_limb_size * bignum256modm_bits_per_limb)

// Must divide bignum256modm_bits_per_limb
#define scalarmult_window_size 4
#define scalarmult_table_size (1 << scalarmult_window_size)
#define scalarmult_window_mask (((bignum256modm_element_t)1 << scalarmult_window_size) - 1)
#define scalarmult_limb_windows (bignum256modm_bits_per_limb / scalarmult_window_size)

void ge25519_scalarmult(ge25519* r, const ge25519* p, const bignum256modm s) {
	// TODO: use signed windows

	ge25519_pniels ALIGN(16) T[scalarmult_table_size];

	ge25519 ALIGN(16) result;
	// Set to neutral element initially
	memset(&result, 0, sizeof(ge25519));
	result.y[0] = 1;
	result.z[0] = 1;

	ge25519_full_to_pniels(&T[0], &result);
	for (int i = 0; i < scalarmult_table_size-1; ++i) {
		ge25519_pnielsadd(&T[i+1], p, &T[i]);
	}

	ge25519_pniels ALIGN(16) addPoint;
	ge25519_p1p1 ALIGN(16) rp1p1;

	bool firstWindow = true;
	for (int limb = bignum256modm_limb_size - 1; limb >= 0; --limb) {
		for (int window = scalarmult_limb_windows - 1; window >= 0; --window) {
			if (firstWindow) {
				window -= ((bignum256modm_storage_bits - bignum256modm_used_bits) / scalarmult_window_size);
				firstWindow = false;
			}

			unsigned int bits = (unsigned int)((s[limb] >> (window * scalarmult_window_size)) & scalarmult_window_mask);
			for (unsigned int i = 0; i < scalarmult_table_size; ++i) {
				unsigned char select = cond_eq(i, bits);
				cond_move(&addPoint, &T[i], sizeof(ge25519_pniels), select);
			}

			for (int i = 0; i < scalarmult_window_size; ++i) {
				ge25519_double_p1p1(&rp1p1, &result);
				if (i < scalarmult_window_size - 1) {
					ge25519_p1p1_to_partial(&result, &rp1p1);
				} else {
					ge25519_p1p1_to_full(&result, &rp1p1);
				}
			}
			ge25519_pnielsadd_p1p1(&rp1p1, &result, &addPoint, 0);

			if (limb > 0 || window > 0) {
				ge25519_p1p1_to_partial(&result, &rp1p1);
			}
		}
	}
	ge25519_p1p1_to_full(r, &rp1p1);
}

void ge25519_scalarmult_vartime(ge25519* r, const ge25519* p, const bignum256modm s) {
	// Adapted from ge25519_double_scalarmult_vartime (SSE2) with s2 = 0
	signed char slide[256];
	ge25519_pniels ALIGN(16) pre[S1_TABLE_SIZE];
	ge25519 ALIGN(16) d;
	ge25519_p1p1 ALIGN(16) t;
	int32_t i;

	contract256_slidingwindow_modm(slide, s, S1_SWINDOWSIZE);

	ge25519_double(&d, p);
	ge25519_full_to_pniels(pre, p);
	for (i = 0; i < S1_TABLE_SIZE - 1; i++)
		ge25519_pnielsadd(&pre[i+1], &d, &pre[i]);

	/* set neutral */
	memset(r, 0, sizeof(ge25519));
	r->y[0] = 1;
	r->z[0] = 1;

	i = 255;
	while ((i >= 0) && !slide[i])
		i--;

	for (; i >= 0; i--) {
		ge25519_double_p1p1(&t, r);

		if (slide[i]) {
			ge25519_p1p1_to_full(r, &t);
			ge25519_pnielsadd_p1p1(&t, r, &pre[abs(slide[i]) / 2], (unsigned char)slide[i] >> 7);
		}

		if (i > 0) {
			ge25519_p1p1_to_partial(r, &t);
		} else {
			ge25519_p1p1_to_full(r, &t);
		}
	}
}

void ge25519_scalarmult_base1(ge25519* r, const bignum256modm s) {
	ge25519_scalarmult_base_niels(r, ge25519_niels_base1_multiples, s);
}

void ge25519_scalarmult_base2(ge25519* r, const bignum256modm s) {
	ge25519_scalarmult_base_niels(r, ge25519_niels_base2_multiples, s);
}

void ge25519_double_scalarmult_bases_vartime(ge25519* r, const bignum256modm s1, const bignum256modm s2) {
	// Adapted from ge25519_double_scalarmult_vartime with two tables
	signed char slide1[256], slide2[256];
	ge25519_p1p1 ALIGN(16) t;
	int32_t i;

	contract256_slidingwindow_modm(slide1, s1, S1_SWINDOWSIZE);
	contract256_slidingwindow_modm(slide2, s2, S2_SWINDOWSIZE);

	/* set neutral */
	memset(r, 0, sizeof(ge25519));
	r->y[0] = 1;
	r->z[0] = 1;

	i = 255;
	while ((i >= 0) && !(slide1[i] | slide2[i]))
		i--;

	for (; i >= 0; i--) {
		ge25519_double_p1p1(&t, r);

		if (slide1[i]) {
			ge25519_p1p1_to_full(r, &t);
			ge25519_nielsadd2_p1p1(&t, r, &ge25519_niels_base1_sliding_multiples[abs(slide1[i]) / 2], (unsigned char)slide1[i] >> 7);
		}

		if (slide2[i]) {
			ge25519_p1p1_to_full(r, &t);
			ge25519_nielsadd2_p1p1(&t, r, &ge25519_niels_base2_sliding_multiples[abs(slide2[i]) / 2], (unsigned char)slide2[i] >> 7);
		}

		if (i > 0) {
			ge25519_p1p1_to_partial(r, &t);
		} else {
			ge25519_p1p1_to_full(r, &t);
		}
	}
}

void ge25519_double_scalarmult_bases(ge25519* r, const bignum256modm s1, const bignum256modm s2) {
	ge25519 ALIGN(16) r2;
	ge25519_scalarmult_base1(r, s1);
	ge25519_scalarmult_base2(&r2, s2);
	ge25519_add(r, r, &r2);
}

void ge25519_full_to_affine(bignum25519 x, bignum25519 y, const ge25519* p) {
	bignum25519 ALIGN(16) zi;
	curve25519_recip(zi, p->z);
	curve25519_mul(x, p->x, zi);
	curve25519_mul(y, p->y, zi);
}

void ge25519_affine_to_niels(ge25519_niels* p, const bignum25519 x, const bignum25519 y) {
	curve25519_sub_reduce(p->ysubx, y, x);
	curve25519_add_reduce(p->xaddy, y, x);
	bignum25519 ALIGN(16) xy2d;
	curve25519_mul(xy2d, x, y);
	curve25519_mul(p->t2d, xy2d, ge25519_ec2d);
}


void ge25519_full_to_niels(ge25519_niels* p, const ge25519 *r) {
	bignum25519 ALIGN(16) x, y;
	ge25519_full_to_affine(x, y, r);
	ge25519_affine_to_niels(p, x, y);
}

void ge25519_neg(ge25519* r, const ge25519* p) {
	bignum25519 ALIGN(16) t;
	curve25519_copy(t, p->x);
	curve25519_neg(r->x, t);
	if (r != p) {
		curve25519_copy(r->y, p->y);
		curve25519_copy(r->z, p->z);
	}
	curve25519_copy(t, p->t);
	curve25519_neg(r->t, t);
}

// TODO: constant-time needed?
int ge25519_unpack_vartime(ge25519* r, const unsigned char p[32]) {
	unsigned char positive[32];
	memcpy(positive, p, 32);
	positive[31] ^= 0x80;
	return ge25519_unpack_negative_vartime(r, positive);
}

void ge25519_rand(ge25519* r) {
	bignum256modm ALIGN(16) exp;
	modm_rand(exp);
	ge25519_scalarmult_base1(r, exp);
}

int ge25519_equals(const ge25519* p1, const ge25519* p2) {
	unsigned char pack1[32];
	unsigned char pack2[32];
	ge25519_pack(pack1, p1);
	ge25519_pack(pack2, p2);
	return ed25519_verify(pack1, pack2, 32);
}

void modm_rand(bignum256modm m) {
	unsigned char entropy[32];
	RAND_bytes(entropy, 32);
	curved25519_make_secret(m, entropy);
}

void modm_neg(bignum256modm r, const bignum256modm x) {
	bignum256modm_element_t c = 0;
	r[0] = (modm_m[0] - x[0])    ; c = (r[0] >> 63); r[0] &= 0xffffffffffffff;
	r[1] = (modm_m[1] - x[1]) - c; c = (r[1] >> 63); r[1] &= 0xffffffffffffff;
	r[2] = (modm_m[2] - x[2]) - c; c = (r[2] >> 63); r[2] &= 0xffffffffffffff;
	r[3] = (modm_m[3] - x[3]) - c; c = (r[3] >> 63); r[3] &= 0xffffffffffffff;
	r[4] = (modm_m[4] - x[4]) - c;
}

void modm_sub(bignum256modm r, const bignum256modm a, const bignum256modm b) {
	bignum256modm_element_t c; // carry
	bignum256modm sub;         // a - b
	bignum256modm addm;        // (a - b) + m

	c = 0;
	sub[0] = a[0] - b[0]    ; c = sub[0] >> 63; sub[0] &= bignum256modm_reduce_mask;
	sub[1] = a[1] - b[1] - c; c = sub[1] >> 63; sub[1] &= bignum256modm_reduce_mask;
	sub[2] = a[2] - b[2] - c; c = sub[2] >> 63; sub[2] &= bignum256modm_reduce_mask;
	sub[3] = a[3] - b[3] - c; c = sub[3] >> 63; sub[3] &= bignum256modm_reduce_mask;
	sub[4] = a[4] - b[4] - c; c = sub[4] >> 63; sub[4] &= bignum256modm_reduce_mask;

	bignum256modm_element_t selectSub = c - 1;  // Return (a - b) if c = 0
	bignum256modm_element_t selectAddm = 0 - c; // Return ((a - b) + m) if c = 1

	c = 0;
	addm[0] = sub[0] + modm_m[0]    ; c = addm[0] >> bignum256modm_bits_per_limb; addm[0] &= bignum256modm_reduce_mask;
	addm[1] = sub[1] + modm_m[1] + c; c = addm[1] >> bignum256modm_bits_per_limb; addm[1] &= bignum256modm_reduce_mask;
	addm[2] = sub[2] + modm_m[2] + c; c = addm[2] >> bignum256modm_bits_per_limb; addm[2] &= bignum256modm_reduce_mask;
	addm[3] = sub[3] + modm_m[3] + c; c = addm[3] >> bignum256modm_bits_per_limb; addm[3] &= bignum256modm_reduce_mask;
	addm[4] = sub[4] + modm_m[4] + c; c = addm[4] >> bignum256modm_bits_per_limb; addm[4] &= bignum256modm_reduce_mask;

	r[0] = (sub[0] & selectSub) | (addm[0] & selectAddm);
	r[1] = (sub[1] & selectSub) | (addm[1] & selectAddm);
	r[2] = (sub[2] & selectSub) | (addm[2] & selectAddm);
	r[3] = (sub[3] & selectSub) | (addm[3] & selectAddm);
	r[4] = (sub[4] & selectSub) | (addm[4] & selectAddm);
}

int modm_equals(const bignum256modm a, const bignum256modm b) {
	uint64_t differentbits = 0;
	for (int i = 0; i < bignum256modm_limb_size; ++i) {
		differentbits |= (a[i] ^ b[i]);
	}
	const uint64_t msb = 0x8000000000000000;
	uint64_t large = differentbits & msb;
	uint64_t subLarge = (differentbits-1) & msb;
	uint64_t overflow = (~large) & subLarge;
	return (int)(overflow >> 63);
}

void curved25519_make_secret(bignum256modm m, unsigned char bytes[32]) {
	bytes[0] &= 248;
	bytes[31] &= 127;
	bytes[31] |= 64;
	expand256_modm(m, bytes, 32);
}

void ed25519_secretkey(ed25519_secret_key sk, const unsigned char bytes[32]) {
	hash_512bits scratch;
	ed25519_hash(scratch, bytes, 32);
	scratch[0] &= 248;
	scratch[31] &= 127;
	scratch[31] |= 64;
	memcpy(sk, scratch, 32);
}

HashReturn Keccak_HashUpdatePoint(Keccak_HashInstance* hashInstance, const ge25519* p) {
	unsigned char buffer[32];
	ge25519_pack(buffer, p);
	return Keccak_HashUpdate(hashInstance, buffer, 32*8);
}

unsigned char cond_eq(unsigned int a, unsigned int b) {
	uint64_t t = a ^ b;
	t = (t - 1) >> 63;
	return (unsigned char)t;
}

void cond_move(void* dst, const void* src, size_t len, unsigned char condition) {
	unsigned char* dstBuf = dst;
	const unsigned char* srcBuf = src;
	unsigned char selectSrc = -condition;
	unsigned char selectDst = -(1 - condition);
	for (size_t i = 0; i < len; ++i) {
		dstBuf[i] = (dstBuf[i] & selectDst) | (srcBuf[i] & selectSrc);
	}
}

void cond_swap(void* a, void* b, size_t len, unsigned char condition) {
	unsigned char* buf1 = a;
	unsigned char* buf2 = b;
	unsigned char swap = -condition;
	for (size_t i = 0; i < len; ++i) {
		unsigned char temp = swap & (buf1[i] ^ buf2[i]);
		buf1[i] ^= temp;
		buf2[i] ^= temp;
	}
}
