#pragma once

#include "crypto.h"

// Computes r = s*p in constant time
void ge25519_scalarmult(ge25519* r, const ge25519* p, const bignum256modm s);

// Computes r = s*p in variable time
void ge25519_scalarmult_vartime(ge25519* r, const ge25519* p, const bignum256modm s);

// Computes r = s*g1
void ge25519_scalarmult_base1(ge25519* r, const bignum256modm s);

// Computes r = s*g2
void ge25519_scalarmult_base2(ge25519* r, const bignum256modm s);

// Computes r = s1*g1 + s2*g2 in variable time
void ge25519_double_scalarmult_bases_vartime(ge25519* r, const bignum256modm s1, const bignum256modm s2);

// Computes r = s1*g1 + s2*g2 in constant time (slower)
void ge25519_double_scalarmult_bases(ge25519* r, const bignum256modm s1, const bignum256modm s2);

// Converts a point in (X,Y,Z,T) format to one in (x,y) format
void ge25519_full_to_affine(bignum25519 x, bignum25519 y, const ge25519* p);

// Converts a point in (x,y) format to a precomputation in (y-x,y+x,2dxy) format
void ge25519_affine_to_niels(ge25519_niels* p, const bignum25519 x, const bignum25519 y);

// Converts a point in (X,Y,Z,T) format to a precomputation in (y-x,y+x,2dxy) format
void ge25519_full_to_niels(ge25519_niels* p, const ge25519* r);

// Negate an element
void ge25519_neg(ge25519* r, const ge25519* p);

// Unpacks the positive value of an element
int ge25519_unpack_vartime(ge25519* r, const unsigned char p[32]);

// Generate a random point in the group
void ge25519_rand(ge25519* r);

// Compare two points
int ge25519_equals(const ge25519* p1, const ge25519* p2);

// Generate a random exponent mod m (the group order of the base points)
void modm_rand(bignum256modm m);

// Subtract elements mod m
void modm_sub(bignum256modm out, const bignum256modm a, const bignum256modm b);

// Negate an element mod m
void modm_neg(bignum256modm r, const bignum256modm x);

// Compare two elements mod m. Assumes that the elements are reduced.
int modm_equals(const bignum256modm a, const bignum256modm b);

// Converts 32 bytes into a Curve25519 secret key. Modifies bytes.
void curved25519_make_secret(bignum256modm m, unsigned char bytes[32]);

// Converts 32 bytes into an Ed25519 secret key.
void ed25519_secretkey(ed25519_secret_key sk, const unsigned char bytes[32]);

// Adds a point as input to a Keccak instance
HashReturn Keccak_HashUpdatePoint(Keccak_HashInstance* hashInstance, const ge25519* p);

#define Keccak_Must(X) if ((X) != SUCCESS) { return false; }

// If a == b, returns 1. Otherwise, returns 0. Constant time.
unsigned char cond_eq(unsigned int a, unsigned int b);

// If condition == 1, *dst = *src. If condition == 0, does nothing. Constant time.
void cond_move(void* dst, const void* src, size_t len, unsigned char condition);

// If condition == 1, swaps a and b. If condition == 0, does nothing. Constant time.
void cond_swap(void* a, void* b, size_t len, unsigned char condition);

const ge25519 ge25519_basepoint1;
const ge25519 ge25519_basepoint2;
const ge25519_niels ALIGN(16) ge25519_niels_base1_sliding_multiples[32];
const ge25519_niels ALIGN(16) ge25519_niels_base2_sliding_multiples[32];
const uint8_t ALIGN(16) ge25519_niels_base1_multiples[256][96];
const uint8_t ALIGN(16) ge25519_niels_base2_multiples[256][96];
