#include <stdlib.h>
#include <stdbool.h>

#include "eccrsh.h"
#include "eccrsh-internals.h"
#include "crypto.h"
#include "ecc.h"
#include "aead.h"

eccrsh_pubkey* eccrsh_newPubKey(void) { return malloc(sizeof(eccrsh_pubkey)); }
eccrsh_seckey* eccrsh_newSecKey(void) { return malloc(sizeof(eccrsh_seckey)); }

void eccrsh_freePubKey(eccrsh_pubkey* pk) { free(pk); }
void eccrsh_freeSecKey(eccrsh_seckey* sk) { free(sk); }

void eccrsh_genKeys(eccrsh_pubkey* pk, eccrsh_seckey* sk) {
	modm_rand(sk->x1);
	modm_rand(sk->x2);
	modm_rand(sk->y1);
	modm_rand(sk->y2);
	modm_rand(sk->z);

	ge25519_double_scalarmult_bases(&pk->c, sk->x1, sk->x2);
	ge25519_double_scalarmult_bases(&pk->d, sk->y1, sk->y2);
	ge25519_scalarmult_base1(&pk->h, sk->z);
}


void eccrsh_packPubKey(unsigned char* buffer, const eccrsh_pubkey* pk) {
	ge25519_pack(buffer       , &pk->c);
	ge25519_pack(buffer + 32  , &pk->d);
	ge25519_pack(buffer + 32*2, &pk->h);
}

void eccrsh_unpackPubKey(eccrsh_pubkey* pk, const unsigned char* buffer) {
	ge25519_unpack_vartime(&pk->c, buffer       );
	ge25519_unpack_vartime(&pk->d, buffer + 32  );
	ge25519_unpack_vartime(&pk->h, buffer + 32*2);
}

void eccrsh_packSecKey(unsigned char* buffer, const eccrsh_seckey* sk) {
	contract256_modm(buffer       , sk->x1);
	contract256_modm(buffer + 32  , sk->x2);
	contract256_modm(buffer + 32*2, sk->y1);
	contract256_modm(buffer + 32*3, sk->y2);
	contract256_modm(buffer + 32*4, sk->z );
}

void eccrsh_unpackSecKey(eccrsh_seckey* sk, const unsigned char* buffer) {
	expand256_modm(sk->x1, buffer       , 32);
	expand256_modm(sk->x2, buffer + 32  , 32);
	expand256_modm(sk->y1, buffer + 32*2, 32);
	expand256_modm(sk->y2, buffer + 32*3, 32);
	expand256_modm(sk->z , buffer + 32*4, 32);
}

void eccrsh_packCiphertext(unsigned char* buffer, const eccrsh_ciphertext* ct) {
	ge25519_pack(buffer       , &ct->u1);
	ge25519_pack(buffer + 32  , &ct->u2);
	ge25519_pack(buffer + 32*2, &ct->e);
	ge25519_pack(buffer + 32*3, &ct->v);
}

void eccrsh_unpackCiphertext(eccrsh_ciphertext* ct, const unsigned char** vBuffer, const unsigned char* buffer) {
	ge25519_unpack_vartime(&ct->u1, buffer       );
	ge25519_unpack_vartime(&ct->u2, buffer + 32  );
	ge25519_unpack_vartime(&ct->e , buffer + 32*2);
	*vBuffer = buffer + 32*3;
}

bool eccrsh_computeAlpha(bignum256modm alpha, const ge25519* u1, const ge25519* u2, const ge25519* e) {
	unsigned char alphaHash[32];
	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, u1));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, u2));
	Keccak_Must(Keccak_HashUpdatePoint(&hasher, e));
	Keccak_Must(Keccak_HashFinal(&hasher, alphaHash));
	curved25519_make_secret(alpha, alphaHash);
	return true;
}

bool eccrsh_encrypt(const eccrsh_pubkey* pk,
	const unsigned char* plaintext, size_t ptLen, const unsigned char* associatedData, size_t adLen,
	unsigned char* ciphertext, size_t ctMax, size_t* ctUsed) {

	*ctUsed = ECCRSH_CT_OVERHEAD + ptLen;
	if (ctMax < *ctUsed) return false;

	// Select a random point as an encryption key
	ge25519 ALIGN(16) KappaPoint;
	ge25519_rand(&KappaPoint);
	unsigned char Kappa[32];
	ge25519_pack(Kappa, &KappaPoint);

	eccrsh_ciphertext ct;

	bignum256modm ALIGN(16) k;
	ge25519 ALIGN(16) cdAlpha;
	if (!eccrsh_encryptRaw(pk, &KappaPoint, &ct, k, &cdAlpha)) return false;

	eccrsh_packCiphertext(ciphertext, &ct);

	unsigned char* nonce = ciphertext + 32*3;
	unsigned char* tag = nonce + 32;
	unsigned char* aeadCiphertext = tag + AEAD_TAG_SIZE;
	return aead_encrypt(aeadCiphertext, tag, plaintext, ptLen, Kappa, associatedData, adLen, nonce);
}

bool eccrsh_encryptRaw(const eccrsh_pubkey* pk, const ge25519* p, eccrsh_ciphertext* ct, bignum256modm k, ge25519* cdAlpha) {
	// TODO: check if pk elements are neutral

	modm_rand(k);
	ge25519_scalarmult_base1(&ct->u1, k);
	ge25519_scalarmult_base2(&ct->u2, k);
	ge25519_scalarmult(&ct->e, &pk->h, k);
	ge25519_add(&ct->e, &ct->e, p);

	bignum256modm ALIGN(16) alpha;
	if (!eccrsh_computeAlpha(alpha, &ct->u1, &ct->u2, &ct->e)) return false;

	// v = k*(c + alpha*d)
	ge25519_scalarmult(cdAlpha, &pk->d, alpha);
	ge25519_add(cdAlpha, &pk->c, cdAlpha);
	ge25519_scalarmult(&ct->v, cdAlpha, k);

	return true;
}

bool eccrsh_decrypt(const eccrsh_seckey* sk,
	const unsigned char* ciphertext, size_t ctLen, const unsigned char* associatedData, size_t adLen,
	unsigned char* plaintext, size_t ptMax, size_t* ptUsed) {

	if (ctLen < ECCRSH_CT_OVERHEAD) return false;
	*ptUsed = ctLen - ECCRSH_CT_OVERHEAD;
	if (ptMax < *ptUsed) return false;

	eccrsh_ciphertext ct;
	const unsigned char* vBuffer;
	eccrsh_unpackCiphertext(&ct, &vBuffer, ciphertext);

	bignum256modm ALIGN(16) alpha;
	if (!eccrsh_computeAlpha(alpha, &ct.u1, &ct.u2, &ct.e)) return false;

	ge25519 ALIGN(16) KappaPoint;
	if (!eccrsh_decryptRaw(sk, &ct, vBuffer, alpha, &KappaPoint)) return false;

	unsigned char Kappa[32];
	ge25519_pack(Kappa, &KappaPoint);

	const unsigned char* nonce = ciphertext + 32*3;
	const unsigned char* tag = nonce + 32;
	const unsigned char* aeadCiphertext = tag + AEAD_TAG_SIZE;
	return aead_decrypt(plaintext, aeadCiphertext, *ptUsed, Kappa, tag, associatedData, adLen, nonce);
}

bool eccrsh_decryptRaw(const eccrsh_seckey* sk, const eccrsh_ciphertext* ct, const unsigned char* vBuffer, const bignum256modm alpha, ge25519* p) {
	// TODO: constant-time double scalarmult add?

	// Compute x1*u1 + x2*u2 + alpha*(y1*u1 + y2*u2)
	ge25519 ALIGN(16) temp1, temp2;
	ge25519_scalarmult(&temp1, &ct->u1, sk->y1);
	ge25519_scalarmult(&temp2, &ct->u2, sk->y2);
	ge25519_add(&temp1, &temp1, &temp2);
	ge25519_scalarmult(&temp1, &temp1, alpha);
	ge25519_scalarmult(&temp2, &ct->u1, sk->x1);
	ge25519_add(&temp1, &temp1, &temp2);
	ge25519_scalarmult(&temp2, &ct->u2, sk->x2);
	ge25519_add(&temp1, &temp1, &temp2);

	unsigned char testPack[32];
	ge25519_pack(testPack, &temp1);
	if (!ed25519_verify(testPack, vBuffer, 32)) return false;

	ge25519_scalarmult(&temp1, &ct->u1, sk->z);
	ge25519_neg(&temp1, &temp1);
	ge25519_add(p, &ct->e, &temp1);

	return true;
}
