#include <stdlib.h>
#include <stdbool.h>

#include "ecelgamal.h"
#include "ecelgamal-internals.h"
#include "crypto.h"
#include "ecc.h"
#include "aead.h"

ecelgamal_pubkey* ecelgamal_newPubKey(void) { return malloc(sizeof(ecelgamal_pubkey)); }
ecelgamal_seckey* ecelgamal_newSecKey(void) { return malloc(sizeof(ecelgamal_seckey)); }

void ecelgamal_freePubKey(ecelgamal_pubkey* pk) { free(pk); }
void ecelgamal_freeSecKey(ecelgamal_seckey* sk) { free(sk); }

void ecelgamal_genKeys(ecelgamal_pubkey* pk, ecelgamal_seckey* sk) {
	modm_rand(sk->x);
	ge25519_scalarmult_base1(&pk->h, sk->x);
}

void ecelgamal_genPackedKeys(unsigned char* secBuffer, ecelgamal_pubkey* public, ecelgamal_seckey* secret) {
	RAND_bytes(secBuffer, 32);
	curved25519_make_secret(secret->x, secBuffer);
	ge25519_scalarmult_base1(&public->h, secret->x);
}

void ecelgamal_packPubKey(unsigned char* buffer, const ecelgamal_pubkey* pk) {
	ge25519_pack(buffer, &pk->h);
}

void ecelgamal_unpackPubKey(ecelgamal_pubkey* pk, const unsigned char* buffer) {
	ge25519_unpack_vartime(&pk->h, buffer);
}

void ecelgamal_packSecKey(unsigned char* buffer, const ecelgamal_seckey* sk) {
	contract256_modm(buffer, sk->x);
}

void ecelgamal_unpackSecKey(ecelgamal_seckey* sk, const unsigned char* buffer) {
	expand256_modm(sk->x, buffer, 32);
}

void ecelgamal_packCiphertext(unsigned char* buffer, const ecelgamal_ciphertext* ct) {
	ge25519_pack(buffer     , &ct->c1);
	ge25519_pack(buffer + 32, &ct->c2);
}

void ecelgamal_unpackCiphertext(ecelgamal_ciphertext* ct, const unsigned char* buffer) {
	ge25519_unpack_vartime(&ct->c1, buffer     );
	ge25519_unpack_vartime(&ct->c2, buffer + 32);
}

bool ecelgamal_encrypt(const ecelgamal_pubkey* pk,
	const unsigned char* plaintext, size_t ptLen, const unsigned char* associatedData, size_t adLen, const unsigned char* nonce,
	unsigned char* ciphertext, size_t ctMax, size_t* ctUsed) {

	*ctUsed = ECELGAMAL_CT_OVERHEAD + ptLen;
	if (ctMax < *ctUsed) return false;

	// Select a random point as an encryption key
	ge25519 ALIGN(16) KappaPoint;
	ge25519_rand(&KappaPoint);
	unsigned char Kappa[32];
	ge25519_pack(Kappa, &KappaPoint);

	ecelgamal_ciphertext ct;

	bignum256modm ALIGN(16) y;
	if (!ecelgamal_encryptRaw(pk, &KappaPoint, &ct, y)) return false;

	ecelgamal_packCiphertext(ciphertext, &ct);

	unsigned char* tag = ciphertext + 32*2;
	unsigned char* aeadCiphertext = tag + AEAD_TAG_SIZE;
	return aead_encrypt(aeadCiphertext, tag, plaintext, ptLen, Kappa, associatedData, adLen, nonce);
}

bool ecelgamal_encryptRaw(const ecelgamal_pubkey* pk, const ge25519* p, ecelgamal_ciphertext* ct, bignum256modm y) {
	// TODO: check if pk elements are neutral
	modm_rand(y);
	ge25519_scalarmult_base1(&ct->c1, y);
	ge25519_scalarmult(&ct->c2, &pk->h, y);
	ge25519_add(&ct->c2, &ct->c2, p);
	return true;
}

bool ecelgamal_decrypt(const ecelgamal_seckey* sk,
	const unsigned char* ciphertext, size_t ctLen, const unsigned char* associatedData, size_t adLen, const unsigned char* nonce,
	unsigned char* plaintext, size_t ptMax, size_t* ptUsed) {

	if (ctLen < ECELGAMAL_CT_OVERHEAD) return false;
	*ptUsed = ctLen - ECELGAMAL_CT_OVERHEAD;
	if (ptMax < *ptUsed) return false;

	ecelgamal_ciphertext ct;
	ecelgamal_unpackCiphertext(&ct, ciphertext);

	ge25519 ALIGN(16) KappaPoint;
	if (!ecelgamal_decryptRaw(sk, &ct, &KappaPoint)) return false;

	unsigned char Kappa[32];
	ge25519_pack(Kappa, &KappaPoint);

	const unsigned char* tag = ciphertext + 32*2;
	const unsigned char* aeadCiphertext = tag + AEAD_TAG_SIZE;
	return aead_decrypt(plaintext, aeadCiphertext, *ptUsed, Kappa, tag, associatedData, adLen, nonce);
}

bool ecelgamal_decryptRaw(const ecelgamal_seckey* sk, const ecelgamal_ciphertext* ct, ge25519* p) {
	ge25519 ALIGN(16) temp;
	ge25519_scalarmult(&temp, &ct->c1, sk->x);
	ge25519_neg(&temp, &temp);
	ge25519_add(p, &ct->c2, &temp);
	return true;
}
