#include "kmac.h"

static size_t kmac_encode(unsigned char* output, size_t x, bool left) {
	const size_t sizeBits = sizeof(size_t) * 8;
	size_t clearBytes = 0;
	for (size_t mask = (size_t)0xFF << (sizeBits - 8); mask; mask >>= 8) {
		if (mask & x) break;
		++clearBytes;
	}
	unsigned char writeBytes = (unsigned char)(sizeof(size_t) - clearBytes);
	size_t i = 0;
	if (left) output[i++] = writeBytes;
	if (writeBytes > 0) {
		size_t shift = (writeBytes - 1) * 8;
		for (size_t mask = (size_t)0xFF << shift; mask; mask >>= 8) {
			output[i++] = (unsigned char)((x & mask) >> shift);
			shift -= 8;
		}
	}
	if (!left) output[i++] = writeBytes;
	return i;
}

static size_t kmac_encodeString(unsigned char* output, const unsigned char* input, size_t inputByteLen) {
	size_t lenSize = kmac_encode(output, inputByteLen, true);
	memcpy(output + lenSize, input, inputByteLen);
	return lenSize + inputByteLen;
}

static size_t kmac_padString(unsigned char* output, size_t wrote, size_t padMultiple) {
	size_t padSize = (padMultiple - (wrote % padMultiple)) % padMultiple;
	for (size_t i = 0; i < padSize; ++i) output[i] = 0;
	return padSize;
}

static size_t kmac_encodePadString(unsigned char* output, const unsigned char* input, size_t inputByteLen, size_t padMultiple) {
	size_t contentSize = kmac_encodeString(output, input, inputByteLen);
	size_t padSize = kmac_padString(output + contentSize, contentSize, padMultiple);
	return contentSize + padSize;
}

static int kmac(unsigned char *output, size_t outputByteLen,
	const unsigned char *key, size_t keyByteLen,
	const unsigned char *input, size_t inputByteLen,
	const unsigned char* customStr, size_t customStrLen,
	Keccak_HashInstance* hashInstance, size_t rate) {

	int success = 0;
	size_t len;

	unsigned char* buffer = malloc(2 + 2*sizeof(size_t) + MAX(keyByteLen, customStrLen + 4) + rate);

	// Write bytepad(encode_string(S) || encode_string(N), rate)
	len = kmac_encodeString(buffer, customStr, customStrLen);
	len += kmac_encodeString(buffer + len, (const unsigned char*)"KMAC", 4);
	len += kmac_padString(buffer + len, len, rate);
	if (Keccak_HashUpdate(hashInstance, buffer, len*8) != SUCCESS) goto cleanup;

	// Write bytepad(encode_string(K), rate) || X || right_encode(L)
	len = kmac_encodePadString(buffer, key, keyByteLen, rate);
	if (Keccak_HashUpdate(hashInstance, buffer, len*8) != SUCCESS) goto cleanup;
	if (Keccak_HashUpdate(hashInstance, input, inputByteLen*8) != SUCCESS) return 1;
	len = kmac_encode(buffer, outputByteLen * 8, false);
	if (Keccak_HashUpdate(hashInstance, buffer, len*8) != SUCCESS) goto cleanup;

	// Trailing "00" bits
	buffer[0] = 0;
	if (Keccak_HashUpdate(hashInstance, buffer, 2) != SUCCESS) goto cleanup;

	if (Keccak_HashFinal(hashInstance, output) != SUCCESS) goto cleanup;

	success = 1;
cleanup:
	free(buffer);
	return success;
}

int kmac128(unsigned char *output, size_t outputByteLen,
	const unsigned char *key, size_t keyByteLen,
	const unsigned char *input, size_t inputByteLen,
	const unsigned char* customStr, size_t customStrLen) {

	Keccak_HashInstance hashInstance;
	Keccak_HashInitialize(&hashInstance, 1344, 256, (unsigned int)outputByteLen*8, 0x1F);

	return kmac(output, outputByteLen, key, keyByteLen, input, inputByteLen, customStr, customStrLen, &hashInstance, 168);
}

int kmac256(unsigned char *output, size_t outputByteLen,
	const unsigned char *key, size_t keyByteLen,
	const unsigned char *input, size_t inputByteLen,
	const unsigned char* customStr, size_t customStrLen) {

	Keccak_HashInstance hashInstance;
	Keccak_HashInitialize(&hashInstance, 1088, 512, (unsigned int)outputByteLen*8, 0x1F);

	return kmac(output, outputByteLen, key, keyByteLen, input, inputByteLen, customStr, customStrLen, &hashInstance, 136);
}
