#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>

#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "crypto.h"
#include "dakez.h"
#include "spawnplus.h"
#include "zdh.h"
#include "xzdh.h"
#include "qr.h"
#include "other-ake.h"

#define UND "\33[4m"
#define NRM "\33[0m"
#define STARTARG(X) UND X NRM
#define ARG(X) " " STARTARG(X)
#define UNUSED(X) (void)(X)

static void panic(const char* message) {
	fprintf(stderr, "Error: %s\n", message);
	exit(254);
}

struct scheme;

typedef struct {
	void* (*setup)(struct scheme*, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen);
	void (*keygen)(struct scheme*, void* state, unsigned char* pubKey, unsigned char* fpubKey);
	void (*step)(struct scheme*, void* state, int step, const unsigned char* pubKeyOther, const unsigned char* fpubKeyOther, const unsigned char* recv, unsigned char* send);
	void (*cleanup)(struct scheme*, void* state);
} partyFuncs;

typedef struct scheme {
	size_t pubKeySize;
	size_t* transmitSizes;
	int freeSteps;

	qrKex* qrKex;

	partyFuncs funcI, funcR;
} scheme;



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                  QRSHELL                                   *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

static void* qrshell_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(id); UNUSED(idLen); UNUSED(ad); UNUSED(adLen);
	return malloc(scheme->qrKex->sizeSqI);
}

static void* qrshell_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(id); UNUSED(idLen); UNUSED(ad); UNUSED(adLen);
	return NULL;
}

static void qrshell_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme); UNUSED(state); UNUSED(pubKey); UNUSED(fpubKey);
}

static void qrshell_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	if (state != NULL) free(state);
}

static void qrshell_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(pubKeyR); UNUSED(fpubKeyR);
	unsigned char* sqI = (unsigned char*)state;
	unsigned char ALIGN(32) scratch[scheme->qrKex->sizeScratch];
	if (step == 1) {
		if (!scheme->qrKex->qrGenI(send, sqI, scratch)) {
			panic("Failed to generate I's message in QR shell");
		}
	} else if (step == 3) {
		unsigned char key[scheme->qrKex->sizeKey];
		if (!scheme->qrKex->qrKeyI(sqI, recv, key, scratch)) {
			panic("Failed to receive R's message in QR shell");
		}
	} else {
		panic("Invalid QR shell I step");
	}
}

static void qrshell_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	UNUSED(state); UNUSED(pubKeyI); UNUSED(fpubKeyI);
	unsigned char ALIGN(32) scratch[scheme->qrKex->sizeScratch];
	if (step == 2) {
		unsigned char key[scheme->qrKex->sizeKey];
		if (!scheme->qrKex->qrGenR(recv, send, key, scratch)) {
			panic("Failed to generate R's message in QR shell");
		}
	} else {
		panic("Invalid QR shell R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                   DAKEZ                                    *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[DAKEZ_PK_SIZE];
	unsigned char secKey[DAKEZ_SK_SIZE];
	unsigned char fpubKey[DAKEZ_PK_SIZE];
	unsigned char* pubState;
	unsigned char* secState;
	const unsigned char* id;
	unsigned char* peerId;
	size_t idLen;
	const unsigned char* ad;
	size_t adLen;
} dakez_state;

static void* dakez_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	dakez_state* dstate = malloc(sizeof(dakez_state));
	dstate->pubState = aligned_alloc(16, DAKEZ_PUBLICSTATE_I_OVERHEAD + scheme->qrKex->sizePqI);
	dstate->secState = aligned_alloc(16, DAKEZ_SECRETSTATE_I_OVERHEAD + scheme->qrKex->sizeAlign8 + scheme->qrKex->sizeSqI);
	dstate->id = id;
	dstate->peerId = malloc(idLen);
	dstate->idLen = idLen;
	dstate->ad = ad;
	dstate->adLen = adLen;
	return dstate;
}

static void* dakez_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	dakez_state* dstate = malloc(sizeof(dakez_state));
	dstate->pubState = aligned_alloc(16, DAKEZ_PUBLICSTATE_R_OVERHEAD + scheme->qrKex->sizePqI + scheme->qrKex->sizeQR);
	dstate->secState = aligned_alloc(16, DAKEZ_SECRETSTATE_R_SIZE);
	dstate->id = id;
	dstate->peerId = malloc(idLen);
	dstate->idLen = idLen;
	dstate->ad = ad;
	dstate->adLen = adLen;
	return dstate;
}

static void dakez_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme);
	dakez_state* dstate = (dakez_state*)state;
	unsigned char scratch[DAKEZ_SCRATCH_KEYGEN_SIZE];
	if (dakez_genLongTermKeys(dstate->pubKey, dstate->secKey, dstate->fpubKey, NULL, scratch) != DAKEZ_SUCCESS) {
		panic("Failed to generate long-term keys");
	}
	if (pubKey != NULL) memcpy(pubKey, dstate->pubKey, DAKEZ_PK_SIZE);
	if (fpubKey != NULL) memcpy(fpubKey, dstate->fpubKey, DAKEZ_PK_SIZE);
}

static void dakez_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	dakez_state* dstate = (dakez_state*)state;
	free(dstate->pubState);
	free(dstate->secState);
	free(dstate->peerId);
	free(dstate);
}

static void dakez_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	dakez_state* dstate = (dakez_state*)state;
	if (step == 1) {
		unsigned char scratch[scheme->qrKex->sizeScratch];
		if (dakez_1_initiate_I(scheme->qrKex, dstate->idLen, dstate->id, dstate->pubState, dstate->secState, send, scratch) != DAKEZ_SUCCESS) {
			panic("Failed to initiate DAKEZ exchange");
		}
	} else if (step == 3) {
		if (dakez_3_receive_I(recv, dstate->idLen, dstate->peerId) != DAKEZ_SUCCESS) {
			panic("Failed to receive R's identifier in DAKEZ");
		}
		unsigned char ALIGN(16) scratch[DAKEZ_SCRATCH_I_RESPOND_OVERHEAD + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[DAKEZ_KEY_SIZE];
		if (dakez_3_respond_I(recv, scheme->qrKex, dstate->idLen, dstate->id, dstate->pubKey, dstate->fpubKey, pubKeyR, fpubKeyR, DAKEZ_HONEST_I, dstate->secKey, dstate->ad, dstate->adLen, dstate->pubState, dstate->secState, send, key, scratch) != DAKEZ_SUCCESS) {
			panic("Failed to respond to R in DAKEZ");
		}
	} else {
		panic("Invalid DAKEZ I step");
	}
}

static void dakez_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	dakez_state* dstate = (dakez_state*)state;
	if (step == 2) {
		if (dakez_2_receive_R(recv, dstate->idLen, dstate->peerId) != DAKEZ_SUCCESS) {
			panic("Failed to receive I's identifier in DAKEZ");
		}
		unsigned char ALIGN(16) scratch[DAKEZ_SCRATCH_R_OVERHEAD + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		if (dakez_2_respond_R(recv, scheme->qrKex, dstate->idLen, dstate->id, fpubKeyI, dstate->pubKey, DAKEZ_HONEST_R, dstate->secKey, dstate->ad, dstate->adLen, dstate->pubState, dstate->secState, send, scratch) != DAKEZ_SUCCESS) {
			panic("Failed to respond to I in DAKEZ");
		}
	} else if (step == 4) {
		unsigned char key[DAKEZ_KEY_SIZE];
		if (dakez_4_verify_R(recv, scheme->qrKex, dstate->idLen, dstate->peerId, dstate->id, pubKeyI, dstate->fpubKey, dstate->ad, dstate->adLen, dstate->pubState, dstate->secState, key) != DAKEZ_SUCCESS) {
			panic("Failed to verify in DAKEZ");
		}
	} else {
		panic("Invalid DAKEZ R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                   SPAWN+                                   *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[SPAWNPLUS_PK_SIZE];
	unsigned char secKey[SPAWNPLUS_SK_SIZE];
	unsigned char fpubKey[SPAWNPLUS_PK_SIZE];
	unsigned char* pubState;
	unsigned char* secState;
	const unsigned char* id;
	unsigned char* peerId;
	size_t idLen;
	const unsigned char* ad;
	size_t adLen;
} spawnplus_state;

static void* spawnplus_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	spawnplus_state* sstate = malloc(sizeof(spawnplus_state));
	sstate->pubState = aligned_alloc(16, SPAWNPLUS_PUBLICSTATE_OVERHEAD + scheme->qrKex->sizePqI);
	sstate->secState = aligned_alloc(16, SPAWNPLUS_SECRETSTATE_OVERHEAD + scheme->qrKex->sizeAlign8 + scheme->qrKex->sizeSqI);
	sstate->id = id;
	sstate->peerId = malloc(idLen);
	sstate->idLen = idLen;
	sstate->ad = ad;
	sstate->adLen = adLen;
	return sstate;
}

static void* spawnplus_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme);
	spawnplus_state* sstate = malloc(sizeof(spawnplus_state));
	sstate->pubState = NULL;
	sstate->secState = NULL;
	sstate->id = id;
	sstate->peerId = malloc(idLen);
	sstate->idLen = idLen;
	sstate->ad = ad;
	sstate->adLen = adLen;
	return sstate;
}

static void spawnplus_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme);
	spawnplus_state* sstate = (spawnplus_state*)state;
	unsigned char scratch[SPAWNPLUS_SCRATCH_KEYGEN_SIZE];
	if (spawnplus_genLongTermKeys(sstate->pubKey, sstate->secKey, sstate->fpubKey, NULL, scratch) != SPAWNPLUS_SUCCESS) {
		panic("Failed to generate long-term keys");
	}
	if (pubKey != NULL) memcpy(pubKey, sstate->pubKey, SPAWNPLUS_PK_SIZE);
	if (fpubKey != NULL) memcpy(fpubKey, sstate->fpubKey, SPAWNPLUS_PK_SIZE);
}

static void spawnplus_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	spawnplus_state* sstate = (spawnplus_state*)state;
	if (sstate->pubState != NULL) free(sstate->pubState);
	if (sstate->secState != NULL) free(sstate->secState);
	free(sstate->peerId);
	free(sstate);
}

static void spawnplus_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	spawnplus_state* sstate = (spawnplus_state*)state;
	if (step == 1) {
		memcpy(send, sstate->id, sstate->idLen);
		unsigned char scratch[scheme->qrKex->sizeScratch];
		if (spawnplus_1_makePreKey_I(scheme->qrKex, sstate->pubState, sstate->secState, send+sstate->idLen, scratch) != SPAWNPLUS_SUCCESS) {
			panic("Failed to generate Spawn+ prekey");
		}
	} else if (step == 3) {
		if (spawnplus_3_receive_I(recv, sstate->idLen, sstate->peerId) != SPAWNPLUS_SUCCESS) {
			panic("Failed to receive R's identifier in Spawn+");
		}
		unsigned char ALIGN(16) scratch[SPAWNPLUS_SCRATCH_I_OVERHEAD + scheme->qrKex->sizeQR + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[SPAWNPLUS_KEY_SIZE];
		if (spawnplus_3_exchange_I(recv, scheme->qrKex, sstate->idLen, sstate->id, sstate->pubKey, sstate->fpubKey, pubKeyR, fpubKeyR, SPAWNPLUS_FINISH_HONEST, sstate->secKey, sstate->ad, sstate->adLen, sstate->pubState, sstate->secState, key, scratch) != SPAWNPLUS_SUCCESS) {
			panic("Failed to receive R's message in Spawn+");
		}
	} else {
		panic("Invalid Spawn+ I step");
	}
}

static void spawnplus_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	spawnplus_state* sstate = (spawnplus_state*)state;
	if (step == 2) {
		memcpy(sstate->peerId, recv, sstate->idLen);
		unsigned char ALIGN(16) scratch[SPAWNPLUS_SCRATCH_R_OVERHEAD + scheme->qrKex->sizeQR + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[SPAWNPLUS_KEY_SIZE];
		if (spawnplus_2_exchange_R(recv+sstate->idLen, scheme->qrKex, sstate->idLen, sstate->peerId, sstate->id, pubKeyI, fpubKeyI, sstate->pubKey, sstate->fpubKey, SPAWNPLUS_RESP_HONEST, sstate->secKey, sstate->ad, sstate->adLen, send, key, scratch) != SPAWNPLUS_SUCCESS) {
			panic("Failed to send message to I in Spawn+");
		}
	} else {
		panic("Invalid Spawn+ R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                    ZDH                                     *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[ZDH_PK_SIZE];
	unsigned char secKey[ZDH_SK_SIZE];
	unsigned char fpubKey[ZDH_PK_SIZE];
	unsigned char* pubState;
	unsigned char* secState;
	const unsigned char* id;
	unsigned char* peerId;
	size_t idLen;
	const unsigned char* ad;
	size_t adLen;
} zdh_state;

static void* zdh_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	zdh_state* zstate = malloc(sizeof(zdh_state));
	zstate->pubState = aligned_alloc(16, ZDH_PUBLICSTATE_OVERHEAD + scheme->qrKex->sizePqI);
	zstate->secState = aligned_alloc(16, ZDH_SECRETSTATE_OVERHEAD + scheme->qrKex->sizeAlign8 + scheme->qrKex->sizeSqI);
	zstate->id = id;
	zstate->peerId = malloc(idLen);
	zstate->idLen = idLen;
	zstate->ad = ad;
	zstate->adLen = adLen;
	return zstate;
}

static void* zdh_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme);
	zdh_state* zstate = malloc(sizeof(zdh_state));
	zstate->pubState = NULL;
	zstate->secState = NULL;
	zstate->id = id;
	zstate->peerId = malloc(idLen);
	zstate->idLen = idLen;
	zstate->ad = ad;
	zstate->adLen = adLen;
	return zstate;
}

static void zdh_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme);
	zdh_state* zstate = (zdh_state*)state;
	unsigned char scratch[ZDH_SCRATCH_KEYGEN_SIZE];
	if (zdh_genLongTermKeys(zstate->pubKey, zstate->secKey, zstate->fpubKey, NULL, scratch) != ZDH_SUCCESS) {
		panic("Failed to generate long-term keys");
	}
	if (pubKey != NULL) memcpy(pubKey, zstate->pubKey, ZDH_PK_SIZE);
	if (fpubKey != NULL) memcpy(fpubKey, zstate->fpubKey, ZDH_PK_SIZE);
}

static void zdh_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	zdh_state* zstate = (zdh_state*)state;
	if (zstate->pubState != NULL) free(zstate->pubState);
	if (zstate->secState != NULL) free(zstate->secState);
	free(zstate->peerId);
	free(zstate);
}

static void zdh_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(fpubKeyR);
	zdh_state* zstate = (zdh_state*)state;
	if (step == 1) {
		memcpy(send, zstate->id, zstate->idLen);
		unsigned char scratch[scheme->qrKex->sizeScratch];
		if (zdh_1_makePreKey_I(scheme->qrKex, zstate->pubState, zstate->secState, send+zstate->idLen, scratch) != ZDH_SUCCESS) {
			panic("Failed to generate ZDH prekey");
		}
	} else if (step == 3) {
		if (zdh_3_receive_I(recv, zstate->idLen, zstate->peerId) != ZDH_SUCCESS) {
			panic("Failed to receive R's identifier in ZDH");
		}
		unsigned char ALIGN(16) scratch[ZDH_SCRATCH_I_OVERHEAD + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[ZDH_KEY_SIZE];
		if (zdh_3_exchange_I(recv, scheme->qrKex, zstate->idLen, zstate->id, zstate->pubKey, zstate->fpubKey, pubKeyR, zstate->secKey, zstate->ad, zstate->adLen, zstate->pubState, zstate->secState, key, scratch) != ZDH_SUCCESS) {
			panic("Failed to receive R's message in ZDH");
		}
	} else {
		panic("Invalid ZDH I step");
	}
}

static void zdh_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	zdh_state* zstate = (zdh_state*)state;
	if (step == 2) {
		memcpy(zstate->peerId, recv, zstate->idLen);
		unsigned char ALIGN(16) scratch[ZDH_SCRATCH_R_OVERHEAD + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[ZDH_KEY_SIZE];
		if (zdh_2_exchange_R(recv+zstate->idLen, scheme->qrKex, zstate->idLen, zstate->peerId, zstate->id, pubKeyI, fpubKeyI, zstate->pubKey, ZDH_HONEST, zstate->secKey, zstate->ad, zstate->adLen, send, key, scratch) != ZDH_SUCCESS) {
			panic("Failed to send message to I in ZDH");
		}
	} else {
		panic("Invalid ZDH R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                    XZDH                                    *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[XZDH_PK_SIZE];
	unsigned char secKey[XZDH_SK_SIZE];
	unsigned char fpubKey[XZDH_PK_SIZE];
	unsigned char* pubSignState;
	unsigned char* pubOneState;
	unsigned char* secSignState;
	unsigned char* secOneState;
	unsigned char* signedPrekey;
	const unsigned char* id;
	unsigned char* peerId;
	size_t idLen;
	const unsigned char* ad;
	size_t adLen;
} xzdh_state;

static void* xzdh_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	xzdh_state* zstate = malloc(sizeof(xzdh_state));
	zstate->pubSignState = malloc(XZDH_SIGNED_PUBLICSTATE_SIZE);
	zstate->pubOneState = aligned_alloc(16, XZDH_ONETIME_PUBLICSTATE_OVERHEAD + scheme->qrKex->sizePqI);
	zstate->secSignState = aligned_alloc(16, XZDH_SIGNED_SECRETSTATE_SIZE);
	zstate->secOneState = aligned_alloc(16, XZDH_ONETIME_SECRETSTATE_OVERHEAD + scheme->qrKex->sizeAlign8 + scheme->qrKex->sizeSqI);
	zstate->signedPrekey = NULL;
	zstate->id = id;
	zstate->peerId = malloc(idLen);
	zstate->idLen = idLen;
	zstate->ad = ad;
	zstate->adLen = adLen;
	return zstate;
}

static void* xzdh_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme);
	xzdh_state* zstate = malloc(sizeof(xzdh_state));
	zstate->pubSignState = NULL;
	zstate->pubOneState = NULL;
	zstate->secSignState = NULL;
	zstate->secOneState = NULL;
	zstate->signedPrekey = malloc(XZDH_SIGNED_PREKEY_SIZE);
	zstate->id = id;
	zstate->peerId = malloc(idLen);
	zstate->idLen = idLen;
	zstate->ad = ad;
	zstate->adLen = adLen;
	return zstate;
}

static void xzdh_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme);
	xzdh_state* zstate = (xzdh_state*)state;
	if (xzdh_genLongTermKeys(zstate->pubKey, zstate->secKey, zstate->fpubKey, NULL) != XZDH_SUCCESS) {
		panic("Failed to generate long-term keys");
	}
	if (pubKey != NULL) memcpy(pubKey, zstate->pubKey, XZDH_PK_SIZE);
	if (fpubKey != NULL) memcpy(fpubKey, zstate->fpubKey, XZDH_PK_SIZE);
}

static void xzdh_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	xzdh_state* zstate = (xzdh_state*)state;
	if (zstate->pubSignState != NULL) free(zstate->pubSignState);
	if (zstate->pubOneState != NULL) free(zstate->pubOneState);
	if (zstate->secSignState != NULL) free(zstate->secSignState);
	if (zstate->secOneState != NULL) free(zstate->secOneState);
	if (zstate->signedPrekey != NULL) free(zstate->signedPrekey);
	free(zstate->peerId);
	free(zstate);
}

static void xzdh_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(fpubKeyR);
	xzdh_state* zstate = (xzdh_state*)state;
	if (step == 1) {
		if (xzdh_genSignedPrekey(zstate->pubKey, zstate->secKey, zstate->pubSignState, zstate->secSignState, send) != XZDH_SUCCESS) {
			panic("Failed to generate XZDH signed prekey");
		}
	} else if (step == 3) {
		memcpy(send, zstate->id, zstate->idLen);
		unsigned char scratch[scheme->qrKex->sizeScratch];
		if (xzdh_1_makeOneTimePrekey_I(scheme->qrKex, zstate->pubOneState, zstate->secOneState, send+zstate->idLen, scratch) != XZDH_SUCCESS) {
			panic("Failed to generate XZDH one-time prekey");
		}
	} else if (step == 5) {
		if (zdh_3_receive_I(recv, zstate->idLen, zstate->peerId) != ZDH_SUCCESS) {
			panic("Failed to receive R's identifier in ZDH");
		}
		unsigned char ALIGN(16) scratch[XZDH_SCRATCH_I_OVERHEAD + scheme->qrKex->sizeAlign8 + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[XZDH_KEY_SIZE];
		if (xzdh_3_exchange_I(recv, scheme->qrKex, zstate->idLen, zstate->id, zstate->pubKey, zstate->fpubKey, pubKeyR, zstate->secKey, zstate->ad, zstate->adLen, zstate->pubSignState, zstate->pubOneState, zstate->secSignState, zstate->secOneState, key, scratch) != XZDH_SUCCESS) {
			panic("Failed to receive R's message in XZDH");
		}
	} else {
		panic("Invalid XZDH I step");
	}
}

static void xzdh_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	xzdh_state* zstate = (xzdh_state*)state;
	if (step == 2) {
		memcpy(zstate->signedPrekey, recv, XZDH_SIGNED_PREKEY_SIZE);
		if (xzdh_2_verifySignedPrekey_R(pubKeyI, zstate->signedPrekey) != XZDH_SUCCESS) {
			panic("Failed to verify I's signed prekey in XZDH");
		}
	} else if (step == 4) {
		memcpy(zstate->peerId, recv, zstate->idLen);
		unsigned char ALIGN(16) scratch[XZDH_SCRATCH_R_OVERHEAD + scheme->qrKex->sizeKey + scheme->qrKex->sizeScratch];
		unsigned char key[XZDH_KEY_SIZE];
		if (xzdh_2_exchange_R(zstate->signedPrekey, recv+zstate->idLen, scheme->qrKex, zstate->idLen, zstate->peerId, zstate->id, pubKeyI, fpubKeyI, zstate->pubKey, XZDH_HONEST, zstate->secKey, zstate->ad, zstate->adLen, send, key, scratch) != XZDH_SUCCESS) {
			panic("Failed to send message to I in XZDH");
		}
	} else {
		panic("Invalid XZDH R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                    ECDH                                    *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

static void* ecdh_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(id); UNUSED(idLen); UNUSED(ad); UNUSED(adLen);
	return malloc(ECDH_SECRETSTATE_SIZE);
}

static void* ecdh_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(id); UNUSED(idLen); UNUSED(ad); UNUSED(adLen);
	return NULL;
}

static void ecdh_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme); UNUSED(state); UNUSED(pubKey); UNUSED(fpubKey);
}

static void ecdh_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	if (state != NULL) free(state);
}

static void ecdh_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(pubKeyR); UNUSED(fpubKeyR);
	unsigned char* secretState = (unsigned char*)state;
	if (step == 1) {
		ecdh_1_initiate_I(secretState, send);
	} else if (step == 3) {
		unsigned char key[ECDH_KEY_SIZE];
		if (!ecdh_3_finish_I(recv, secretState, key)) {
			panic("Failed to receive R's message 2 in ECDH");
		}
	} else {
		panic("Invalid ECDH I step");
	}
}

static void ecdh_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(state); UNUSED(pubKeyI); UNUSED(fpubKeyI);
	if (step == 2) {
		unsigned char key[ECDH_KEY_SIZE];
		if (!ecdh_2_respond_R(recv, send, key)) {
			panic("Failed to generate ECDH message 2");
		}
	} else {
		panic("Invalid ECDH R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                    3DH                                     *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[THREEDH_PK_SIZE];
	unsigned char secKey[THREEDH_SK_SIZE];
	unsigned char* secState;
	const unsigned char* id;
	unsigned char* peerId;
	size_t idLen;
} threedh_state;

static void* threedh_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(ad); UNUSED(adLen);
	threedh_state* tstate = malloc(sizeof(threedh_state));
	tstate->secState = aligned_alloc(16, THREEDH_SECRETSTATE_SIZE);
	tstate->id = id;
	tstate->peerId = malloc(idLen);
	tstate->idLen = idLen;
	return tstate;
}

static void* threedh_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(ad); UNUSED(adLen);
	threedh_state* tstate = malloc(sizeof(threedh_state));
	tstate->secState = NULL;
	tstate->id = id;
	tstate->peerId = malloc(idLen);
	tstate->idLen = idLen;
	return tstate;
}

static void threedh_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme); UNUSED(fpubKey);
	threedh_state* tstate = (threedh_state*)state;
	unsigned char scratch[THREEDH_SCRATCH_KEYGEN_SIZE];
	threedh_genLongTermKeys(tstate->pubKey, tstate->secKey, scratch);
	if (pubKey != NULL) memcpy(pubKey, tstate->pubKey, THREEDH_PK_SIZE);
}

static void threedh_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	threedh_state* tstate = (threedh_state*)state;
	if (tstate->secState != NULL) free(tstate->secState);
	free(tstate->peerId);
	free(tstate);
}

static void threedh_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(fpubKeyR);
	threedh_state* tstate = (threedh_state*)state;
	if (step == 1) {
		memcpy(send, tstate->id, tstate->idLen);
		threedh_1_makePreKey_I(tstate->secState, send+tstate->idLen);
	} else if (step == 3) {
		threedh_3_receive_I(recv, tstate->idLen, tstate->peerId);
		unsigned char key[THREEDH_KEY_SIZE];
		if (!threedh_3_exchange_I(recv, tstate->idLen, pubKeyR, tstate->secKey, tstate->secState, key)) {
			panic("Failed to receive R's message in 3DH");
		}
	} else {
		panic("Invalid 3DH I step");
	}
}

static void threedh_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(fpubKeyI);
	threedh_state* tstate = (threedh_state*)state;
	if (step == 2) {
		memcpy(tstate->peerId, recv, tstate->idLen);
		unsigned char key[THREEDH_KEY_SIZE];
		if (!threedh_2_exchange_R(recv+tstate->idLen, tstate->idLen, tstate->id, pubKeyI, tstate->secKey, send, key)) {
			panic("Failed to send message to I in 3DH");
		}
	} else {
		panic("Invalid 3DH R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                    X3DH                                    *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[XTHREEDH_PK_SIZE];
	unsigned char secKey[XTHREEDH_SK_SIZE];
	unsigned char* secSignState;
	unsigned char* secOneState;
	unsigned char* signedPrekey;
	const unsigned char* id;
	unsigned char* peerId;
	size_t idLen;
} xthreedh_state;

static void* xthreedh_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(ad); UNUSED(adLen);
	xthreedh_state* tstate = malloc(sizeof(xthreedh_state));
	tstate->secSignState = aligned_alloc(16, XTHREEDH_SECRETSTATE_SIZE);
	tstate->secOneState = aligned_alloc(16, XTHREEDH_SECRETSTATE_SIZE);
	tstate->signedPrekey = NULL;
	tstate->id = id;
	tstate->peerId = malloc(idLen);
	tstate->idLen = idLen;
	return tstate;
}

static void* xthreedh_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(ad); UNUSED(adLen);
	xthreedh_state* tstate = malloc(sizeof(xthreedh_state));
	tstate->secSignState = NULL;
	tstate->secOneState = NULL;
	tstate->signedPrekey = malloc(XTHREEDH_SIGNED_PREKEY_SIZE);
	tstate->id = id;
	tstate->peerId = malloc(idLen);
	tstate->idLen = idLen;
	return tstate;
}

static void xthreedh_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme); UNUSED(fpubKey);
	xthreedh_state* tstate = (xthreedh_state*)state;
	xthreedh_genLongTermKeys(tstate->pubKey, tstate->secKey);
	if (pubKey != NULL) memcpy(pubKey, tstate->pubKey, XTHREEDH_PK_SIZE);
}

static void xthreedh_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	xthreedh_state* tstate = (xthreedh_state*)state;
	if (tstate->secSignState != NULL) free(tstate->secSignState);
	if (tstate->secOneState != NULL) free(tstate->secOneState);
	if (tstate->signedPrekey != NULL) free(tstate->signedPrekey);
	free(tstate->peerId);
	free(tstate);
}

static void xthreedh_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(fpubKeyR);
	xthreedh_state* tstate = (xthreedh_state*)state;
	if (step == 1) {
		xthreedh_genSignedPrekey(tstate->pubKey, tstate->secKey, tstate->secSignState, send);
	} else if (step == 3) {
		memcpy(send, tstate->id, tstate->idLen);
		xthreedh_1_makeOneTimePreKey_I(tstate->secOneState, send+tstate->idLen);
	} else if (step == 5) {
		xthreedh_3_receive_I(recv, tstate->idLen, tstate->peerId);
		unsigned char key[XTHREEDH_KEY_SIZE];
		if (!xthreedh_3_exchange_I(recv, tstate->idLen, pubKeyR, tstate->secKey, tstate->secSignState, tstate->secOneState, key)) {
			panic("Failed to receive R's message in X3DH");
		}
	} else {
		panic("Invalid X3DH I step");
	}
}

static void xthreedh_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(fpubKeyI);
	xthreedh_state* tstate = (xthreedh_state*)state;
	if (step == 2) {
		memcpy(tstate->signedPrekey, recv, XTHREEDH_SIGNED_PREKEY_SIZE);
		if (!xthreedh_2_verifySignedPrekey_R(pubKeyI, tstate->signedPrekey)) {
			panic("Failed to verify I's signed prekey in X3DH");
		}
	} else if (step == 4) {
		memcpy(tstate->peerId, recv, tstate->idLen);
		unsigned char key[XTHREEDH_KEY_SIZE];
		if (!xthreedh_2_exchange_R(tstate->signedPrekey, recv+tstate->idLen, tstate->idLen, tstate->id, pubKeyI, tstate->secKey, send, key)) {
			panic("Failed to send message to I in X3DH");
		}
	} else {
		panic("Invalid X3DH R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                  SIGMA-R                                   *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

typedef struct {
	unsigned char pubKey[SIGMAR_PK_SIZE];
	unsigned char secKey[SIGMAR_SK_SIZE];
	unsigned char* secState;
	const unsigned char* id;
	size_t idLen;
} sigmar_state;

static void* sigmar_setupI(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(ad); UNUSED(adLen);
	sigmar_state* rstate = malloc(sizeof(sigmar_state));
	rstate->secState = aligned_alloc(16, SIGMAR_SECRETSTATE_I_SIZE);
	rstate->id = id;
	rstate->idLen = idLen;
	return rstate;
}

static void* sigmar_setupR(struct scheme* scheme, const unsigned char* id, size_t idLen, const unsigned char* ad, size_t adLen) {
	UNUSED(scheme); UNUSED(ad); UNUSED(adLen);
	sigmar_state* rstate = malloc(sizeof(sigmar_state));
	rstate->secState = aligned_alloc(16, SIGMAR_SECRETSTATE_R_SIZE);
	rstate->id = id;
	rstate->idLen = idLen;
	return rstate;
}

static void sigmar_keygen(struct scheme* scheme, void* state, unsigned char* pubKey, unsigned char* fpubKey) {
	UNUSED(scheme); UNUSED(fpubKey);
	sigmar_state* rstate = (sigmar_state*)state;
	sigmar_genLongTermKeys(rstate->pubKey, rstate->secKey);
	if (pubKey != NULL) memcpy(pubKey, rstate->pubKey, SIGMAR_PK_SIZE);
}

static void sigmar_cleanup(struct scheme* scheme, void* state) {
	UNUSED(scheme);
	sigmar_state* rstate = (sigmar_state*)state;
	free(rstate->secState);
	free(rstate);
}

static void sigmar_stepI(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyR, const unsigned char* fpubKeyR, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(fpubKeyR);
	sigmar_state* rstate = (sigmar_state*)state;
	if (step == 1) {
		sigmar_1_initiateI(rstate->secState, send);
	} else if (step == 3) {
		if (!sigmar_3_respond_I(recv, rstate->idLen, rstate->id, rstate->pubKey, rstate->secKey, rstate->secState, send)) {
			panic("Failed to respond to R in SIGMA-R");
		}
	} else if (step == 5) {
		unsigned char peerId[rstate->idLen];
		sigmar_5_receive_I(recv, rstate->idLen, peerId);
		unsigned char key[SIGMAR_KEY_SIZE];
		if (!sigmar_5_complete_I(recv, rstate->idLen, peerId, pubKeyR, rstate->secState, key)) {
			panic("Failed to complete exchange with R in SIGMA-R");
		}
	} else {
		panic("Invalid SIGMA-R I step");
	}
}

static void sigmar_stepR(struct scheme* scheme, void* state, int step, const unsigned char* pubKeyI, const unsigned char* fpubKeyI, const unsigned char* recv, unsigned char* send) {
	UNUSED(scheme); UNUSED(fpubKeyI);
	sigmar_state* rstate = (sigmar_state*)state;
	if (step == 2) {
		sigmar_2_respond_R(recv, rstate->secState, send);
	} else if (step == 4) {
		unsigned char peerId[rstate->idLen];
		sigmar_4_receive_R(recv, rstate->idLen, peerId);
		unsigned char key[SIGMAR_KEY_SIZE];
		if (!sigmar_4_complete_R(recv, rstate->idLen, peerId, rstate->id, pubKeyI, rstate->pubKey, rstate->secKey, rstate->secState, send, key)) {
			panic("Failed to complete exchange with I in SIGMA-R");
		}
	} else {
		panic("Invalid SIGMA-R R step");
	}
}



/******************************************************************************\
|******************************************************************************|
|******************************************************************************|
|*                                    MAIN                                    *|
|******************************************************************************|
|******************************************************************************|
\******************************************************************************/

static void exitUsage(const char* progName) {
	fprintf(stderr, "Usage: %s init" ARG("bindip") ARG("remoteip") ARG("port") ARG("dake") ARG("qr") "\n"
	                "       %s resp" ARG("listenip") ARG("port") ARG("dake") ARG("qr") "\n"
		            "       %s keygen" ARG("dake") ARG("qr") "\n"
		            "       %s kex" ARG("dake") ARG("qr") "\n"
		            "       %s size" ARG("dake") ARG("qr") "\n"
		            "\n"
		            STARTARG("bindip") "," ARG("remoteip") ", and" ARG("listenip") " are all IPv4 addresses.\n"
		            STARTARG("port") " is a TCP port.\n"
		            STARTARG("dake") " is one of \"none\", \"dakez\", \"spawn+\", \"zdh\", \"xzdh\", \"ecdh\", \"3dh\", \"x3dh\", or \"sigma-r\"\n"
		            STARTARG("qr") " is one of \"none\", \"sidh\", or \"newhope\"\n"
		            "\n"
		            "Only \"dakez\", \"spawn+\", \"zdh\", and \"xzdh\" support QR schemes other than \"none\".\n"
		            "If \"none\" is selected as the DAKE, then the QR scheme must not be \"none\".\n"
		            "\n"
		            "In all modes except \"size\", the program outputs \"?\" when setup is complete.\n"
		            "When it receives a newline character, it runs the benchmark.\n"
		            "The program outputs the number of nanoseconds when finished, followed by \"!\".\n"
		            , progName, progName, progName, progName, progName);
	exit(255);
}

#define IDLEN 8

static void makeScheme(scheme* scheme, const char* dake, const char* qr, const char* progName, size_t idLen) {
	if (strcmp(qr, "none") == 0) {
		scheme->qrKex = qrKexNone;
	} else if (strcmp(qr, "sidh") == 0) {
		scheme->qrKex = qrKexSIDH;
	} else if (strcmp(qr, "newhope") == 0) {
		scheme->qrKex = qrKexNewHope;
	} else {
		exitUsage(progName);
	}
	scheme->qrKex->qrInit();

	if (strcmp(dake, "none") == 0) {
		if (scheme->qrKex == qrKexNone) { panic("Quantum resistance must be used when not using a DAKE."); }
		scheme->funcI.setup = &qrshell_setupI;
		scheme->funcR.setup = &qrshell_setupR;
		scheme->pubKeySize = 0;
		scheme->funcI.keygen = scheme->funcR.keygen = &qrshell_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*4);
		scheme->transmitSizes[0] = scheme->qrKex->sizePqI;
		scheme->transmitSizes[1] = scheme->qrKex->sizeQR;
		scheme->transmitSizes[2] = 0;
		scheme->transmitSizes[3] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &qrshell_stepI;
		scheme->funcR.step = &qrshell_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &qrshell_cleanup;
	} else if (strcmp(dake, "dakez") == 0) {
		scheme->funcI.setup = &dakez_setupI;
		scheme->funcR.setup = &dakez_setupR;
		scheme->pubKeySize = DAKEZ_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &dakez_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*5);
		scheme->transmitSizes[0] = DAKEZ_1_SEND_I_OVERHEAD + scheme->qrKex->sizePqI + idLen;
		scheme->transmitSizes[1] = DAKEZ_2_SEND_R_OVERHEAD + scheme->qrKex->sizeQR + idLen;
		scheme->transmitSizes[2] = DAKEZ_3_SEND_I_SIZE;
		scheme->transmitSizes[3] = 0;
		scheme->transmitSizes[4] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &dakez_stepI;
		scheme->funcR.step = &dakez_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &dakez_cleanup;
	} else if (strcmp(dake, "spawn+") == 0) {
		scheme->funcI.setup = &spawnplus_setupI;
		scheme->funcR.setup = &spawnplus_setupR;
		scheme->pubKeySize = SPAWNPLUS_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &spawnplus_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*4);
		scheme->transmitSizes[0] = idLen + SPAWNPLUS_PREKEY_OVERHEAD + scheme->qrKex->sizePqI;
		scheme->transmitSizes[1] = SPAWNPLUS_RESP_OVERHEAD + scheme->qrKex->sizeQR + idLen;
		scheme->transmitSizes[2] = 0;
		scheme->transmitSizes[3] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &spawnplus_stepI;
		scheme->funcR.step = &spawnplus_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &spawnplus_cleanup;
	} else if (strcmp(dake, "zdh") == 0) {
		scheme->funcI.setup = &zdh_setupI;
		scheme->funcR.setup = &zdh_setupR;
		scheme->pubKeySize = ZDH_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &zdh_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*4);
		scheme->transmitSizes[0] = idLen + ZDH_PREKEY_OVERHEAD + scheme->qrKex->sizePqI;
		scheme->transmitSizes[1] = ZDH_RESP_OVERHEAD + scheme->qrKex->sizeQR + idLen;
		scheme->transmitSizes[2] = 0;
		scheme->transmitSizes[3] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &zdh_stepI;
		scheme->funcR.step = &zdh_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &zdh_cleanup;
	} else if (strcmp(dake, "xzdh") == 0) {
		scheme->funcI.setup = &xzdh_setupI;
		scheme->funcR.setup = &xzdh_setupR;
		scheme->pubKeySize = XZDH_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &xzdh_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*6);
		scheme->transmitSizes[0] = XZDH_SIGNED_PREKEY_SIZE;
		scheme->transmitSizes[1] = 0;
		scheme->transmitSizes[2] = idLen + XZDH_ONETIME_PREKEY_OVERHEAD + scheme->qrKex->sizePqI;
		scheme->transmitSizes[3] = XZDH_RESP_OVERHEAD + scheme->qrKex->sizeQR + idLen;
		scheme->transmitSizes[4] = 0;
		scheme->transmitSizes[5] = 0;
		scheme->freeSteps = 2;
		scheme->funcI.step = &xzdh_stepI;
		scheme->funcR.step = &xzdh_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &xzdh_cleanup;
	} else if (strcmp(dake, "ecdh") == 0) {
		if (scheme->qrKex != qrKexNone) { panic("ECDH does not support quantum resistance."); }
		scheme->funcI.setup = &ecdh_setupI;
		scheme->funcR.setup = &ecdh_setupR;
		scheme->pubKeySize = 0;
		scheme->funcI.keygen = scheme->funcR.keygen = &ecdh_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*4);
		scheme->transmitSizes[0] = ECDH_MSG_SIZE;
		scheme->transmitSizes[1] = ECDH_MSG_SIZE;
		scheme->transmitSizes[2] = 0;
		scheme->transmitSizes[3] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &ecdh_stepI;
		scheme->funcR.step = &ecdh_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &ecdh_cleanup;
	} else if (strcmp(dake, "3dh") == 0) {
		if (scheme->qrKex != qrKexNone) { panic("3DH does not support quantum resistance."); }
		scheme->funcI.setup = &threedh_setupI;
		scheme->funcR.setup = &threedh_setupR;
		scheme->pubKeySize = THREEDH_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &threedh_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*4);
		scheme->transmitSizes[0] = idLen + THREEDH_PREKEY_SIZE;
		scheme->transmitSizes[1] = THREEDH_MSG_OVERHEAD + idLen;
		scheme->transmitSizes[2] = 0;
		scheme->transmitSizes[3] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &threedh_stepI;
		scheme->funcR.step = &threedh_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &threedh_cleanup;
	} else if (strcmp(dake, "x3dh") == 0) {
		if (scheme->qrKex != qrKexNone) { panic("X3DH does not support quantum resistance."); }
		scheme->funcI.setup = &xthreedh_setupI;
		scheme->funcR.setup = &xthreedh_setupR;
		scheme->pubKeySize = XTHREEDH_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &xthreedh_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*6);
		scheme->transmitSizes[0] = XTHREEDH_SIGNED_PREKEY_SIZE;
		scheme->transmitSizes[1] = 0;
		scheme->transmitSizes[2] = idLen + XTHREEDH_ONETIME_PREKEY_SIZE;
		scheme->transmitSizes[3] = THREEDH_MSG_OVERHEAD + idLen;
		scheme->transmitSizes[4] = 0;
		scheme->transmitSizes[5] = 0;
		scheme->freeSteps = 2;
		scheme->funcI.step = &xthreedh_stepI;
		scheme->funcR.step = &xthreedh_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &xthreedh_cleanup;
	} else if (strcmp(dake, "sigma-r") == 0) {
		if (scheme->qrKex != qrKexNone) { panic("SIGMA-R does not support quantum resistance."); }
		scheme->funcI.setup = &sigmar_setupI;
		scheme->funcR.setup = &sigmar_setupR;
		scheme->pubKeySize = SIGMAR_PK_SIZE;
		scheme->funcI.keygen = scheme->funcR.keygen = &sigmar_keygen;
		scheme->transmitSizes = malloc(sizeof(size_t)*6);
		scheme->transmitSizes[0] = SIGMAR_EX1_MSG_SIZE;
		scheme->transmitSizes[1] = SIGMAR_EX1_MSG_SIZE;
		scheme->transmitSizes[2] = SIGMAR_EX2_MSG_OVERHEAD + idLen;
		scheme->transmitSizes[3] = SIGMAR_EX2_MSG_OVERHEAD + idLen;
		scheme->transmitSizes[4] = 0;
		scheme->transmitSizes[5] = 0;
		scheme->freeSteps = 0;
		scheme->funcI.step = &sigmar_stepI;
		scheme->funcR.step = &sigmar_stepR;
		scheme->funcI.cleanup = scheme->funcR.cleanup = &sigmar_cleanup;
	} else {
		exitUsage(progName);
	}
}

static void cleanupScheme(scheme* scheme) {
	scheme->qrKex->qrCleanup();
}

static void startTimer(struct timespec* timer, bool realTime) {
	putc('?', stdout);
	fflush(stdout);
	while (getc(stdin) != '\n');
	clock_gettime(realTime ? CLOCK_MONOTONIC : CLOCK_THREAD_CPUTIME_ID, timer);
}

static void endTimer(struct timespec* timer, bool realTime)
{
	struct timespec end;
	clock_gettime(realTime ? CLOCK_MONOTONIC : CLOCK_THREAD_CPUTIME_ID, &end);

	if ((end.tv_nsec-timer->tv_nsec)<0) {
		end.tv_sec = end.tv_sec-timer->tv_sec-1;
		end.tv_nsec = 1000000000+end.tv_nsec-timer->tv_nsec;
	} else {
		end.tv_sec = end.tv_sec-timer->tv_sec;
		end.tv_nsec = end.tv_nsec-timer->tv_nsec;
	}
	printf("%lu.%09lu\n", end.tv_sec, end.tv_nsec);
}

typedef enum {
	MEASURE_KEYGEN,
	MEASURE_KEX,
	MEASURE_SIZE,
} simMeasurement;

static void simExchange(scheme* scheme, simMeasurement measure) {
	struct timespec timer;
	if (measure == MEASURE_SIZE) {
		printf("Public key size: %lu\n", scheme->pubKeySize);
	}
	unsigned char idI[IDLEN];
	unsigned char idR[IDLEN];
	unsigned char pubKeyI[scheme->pubKeySize];
	unsigned char pubKeyR[scheme->pubKeySize];
	unsigned char fpubKeyI[scheme->pubKeySize];
	unsigned char fpubKeyR[scheme->pubKeySize];

	RAND_bytes(idI, IDLEN);
	RAND_bytes(idR, IDLEN);

	void* stateI = scheme->funcI.setup(scheme, idI, IDLEN, NULL, 0);
	void* stateR = NULL;

	if (measure == MEASURE_KEYGEN) startTimer(&timer, false);
	scheme->funcI.keygen(scheme, stateI, pubKeyI, fpubKeyI);
	if (measure == MEASURE_KEYGEN) {
		endTimer(&timer, false);
		goto cleanup;
	}

	stateR = scheme->funcR.setup(scheme, idR, IDLEN, NULL, 0);
	scheme->funcR.keygen(scheme, stateR, pubKeyR, fpubKeyR);

	{
		int step = 0;
		size_t sizeSendI, sizeSendR;
		sizeSendI = sizeSendR = 1024*8;
		unsigned char* sendI = malloc(sizeSendI);
		unsigned char* sendR = malloc(sizeSendR);
		size_t sentData = 0;
		size_t totalData = 0;
		size_t freeData = 0;
		do {
			if (step == scheme->freeSteps && measure == MEASURE_KEX) startTimer(&timer, false);
			sentData = scheme->transmitSizes[step];
			++step;
			if (step & 1) {
				if (sentData > sizeSendI) sendI = realloc(sendI, sentData);
				scheme->funcI.step(scheme, stateI, step, pubKeyR, fpubKeyR, sendR, sendI);
			} else {
				if (sentData > sizeSendR) sendR = realloc(sendR, sentData);
				scheme->funcR.step(scheme, stateR, step, pubKeyI, fpubKeyI, sendI, sendR);
			}
			if (measure == MEASURE_SIZE && sentData > 0) {
				printf("Transmission %s%s: %lu\n", (step & 1) ? "I -> R" : "R -> I", (step <= scheme->freeSteps ? " (free)" : ""), sentData);
			}
			totalData += sentData;
			if (step <= scheme->freeSteps) freeData += sentData;
		} while (sentData > 0 || step <= scheme->freeSteps);
		if (measure == MEASURE_KEX) endTimer(&timer, false);
		else if (measure == MEASURE_SIZE) {
			printf("Total transmission data: %lu", totalData);
			if (freeData > 0) {
				printf(" (%lu free, %lu non-free)", freeData, totalData - freeData);
			}
			printf("\n");
		}
		free(sendI);
		free(sendR);
	}

cleanup:
	scheme->funcI.cleanup(scheme, stateI);
	if (stateR != NULL) scheme->funcR.cleanup(scheme, stateR);
}

static int main_sim(int argc, char** argv, simMeasurement measure) {
	if (argc != 4) exitUsage(argv[0]);
	scheme scheme;
	makeScheme(&scheme, argv[2], argv[3], argv[0], IDLEN);
	simExchange(&scheme, measure);
	cleanupScheme(&scheme);
	return 0;
}

static bool send_all(int sockfd, const void* buf, size_t len, int flags) {
	char* nextBuf = (char*)buf;
	while (len > 0) {
		ssize_t sent = send(sockfd, nextBuf, len, flags);
		if (sent < 0) return false;
		len -= (size_t)sent;
		nextBuf += (size_t)sent;
	}
	return true;
}

static bool recv_all(int sockfd, void* buf, size_t len, int flags) {
	ssize_t totalReceived = 0;
	char* nextBuf = (char*)buf;
	while (len > 0) {
		ssize_t received = recv(sockfd, nextBuf, len, flags);
		if (received < 0) return false;
		if (received == 0) return false;
		len -= (size_t)received;
		nextBuf += (size_t)received;
		totalReceived += received;
	}
	return true;
}

#define NET_DO(X) if (!(X)) { fprintf(stderr, "WARNING: connection severed before completion\n"); return; }

static void runNetProtocol(int sock, scheme* scheme, partyFuncs* funcs, bool speakFirst) {
	struct timespec timer;
	unsigned char id[IDLEN];
	unsigned char pubKey[scheme->pubKeySize];
	unsigned char fpubKey[scheme->pubKeySize];
	unsigned char peerPubKey[scheme->pubKeySize];
	unsigned char peerFPubKey[scheme->pubKeySize];

	RAND_bytes(id, IDLEN);

	void* state = funcs->setup(scheme, id, IDLEN, NULL, 0);
	funcs->keygen(scheme, state, pubKey, fpubKey);

	NET_DO(send_all(sock, pubKey, scheme->pubKeySize, 0));
	NET_DO(send_all(sock, fpubKey, scheme->pubKeySize, 0));
	NET_DO(recv_all(sock, peerPubKey, scheme->pubKeySize, 0));
	NET_DO(recv_all(sock, peerFPubKey, scheme->pubKeySize, 0));

	size_t sizeSend, sizeRecv;
	sizeSend = sizeRecv = 1024*8;
	unsigned char* sendBuf = malloc(sizeSend);
	unsigned char* recvBuf = malloc(sizeRecv);

	startTimer(&timer, true);

	int step = 0;
	bool speak = speakFirst;
	while (true) {
		if (speak) {
			size_t sentData = scheme->transmitSizes[step++];
			if (sentData > sizeSend) sendBuf = realloc(sendBuf, sentData);
			funcs->step(scheme, state, step, peerPubKey, peerFPubKey, recvBuf, sendBuf);
			if (sentData == 0 && step > scheme->freeSteps) break;
			NET_DO(send_all(sock, sendBuf, sentData, 0));
		}
		speak = true;

		size_t recvData = scheme->transmitSizes[step++];
		if (recvData == 0 && step > scheme->freeSteps) break;
		if (recvData > sizeRecv) recvBuf = realloc(recvBuf, recvData);
		NET_DO(recv_all(sock, recvBuf, recvData, 0));
	}
	endTimer(&timer, true);
}

static int main_client(int argc, char** argv) {
	if (argc != 7) exitUsage(argv[0]);
	scheme scheme;
	makeScheme(&scheme, argv[5], argv[6], argv[0], IDLEN);

	long port = strtol(argv[4], NULL, 10);
	if (port < 1 || port > 65535) exitUsage(argv[0]);
	in_addr_t clientIp = inet_addr(argv[2]);
	if (clientIp == INADDR_NONE) exitUsage(argv[0]);
	in_addr_t servIp = inet_addr(argv[3]);
	if (servIp == INADDR_NONE) exitUsage(argv[0]);

	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1) panic("Could not allocate client socket");

	struct sockaddr_in clientAddr;
	clientAddr.sin_family = AF_INET;
	clientAddr.sin_port = 0;
	clientAddr.sin_addr.s_addr = clientIp;
	memset(&clientAddr.sin_zero, 0, sizeof(clientAddr.sin_zero));
	if (bind(sock, (struct sockaddr*)&clientAddr, sizeof(clientAddr)) == -1) panic("Could not bind to client address");

	struct sockaddr_in servAddr;
	servAddr.sin_family = AF_INET;
	servAddr.sin_port = htons((uint16_t)port);
	servAddr.sin_addr.s_addr = servIp;
	memset(&servAddr.sin_zero, 0, sizeof(servAddr.sin_zero));

	if (connect(sock, (struct sockaddr*)&servAddr, sizeof(servAddr)) != 0) panic("Could not connect to server");

	runNetProtocol(sock, &scheme, &scheme.funcI, true);

	close(sock);
	cleanupScheme(&scheme);
	return 0;
}

static int main_server(int argc, char** argv) {
	if (argc != 6) exitUsage(argv[0]);
	scheme scheme;
	makeScheme(&scheme, argv[4], argv[5], argv[0], IDLEN);

	long port = strtol(argv[3], NULL, 10);
	if (port < 1 || port > 65535) exitUsage(argv[0]);
	in_addr_t ip = inet_addr(argv[2]);
	if (ip == INADDR_NONE) exitUsage(argv[0]);

	int listenSock = socket(AF_INET, SOCK_STREAM, 0);
	if (listenSock == -1) panic("Could not allocate server socket");
	if (setsockopt(listenSock, SOL_SOCKET, SO_REUSEADDR, &(int){1}, sizeof(int)) == -1) panic("Could not set server socket to reuse address");
	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons((uint16_t)port);
	addr.sin_addr.s_addr = ip;
	memset(&addr.sin_zero, 0, sizeof(addr.sin_zero));
	if (bind(listenSock, (struct sockaddr*)&addr, sizeof(addr)) == -1) panic("Could not bind to listen address");
	if (listen(listenSock, 0) != 0) panic("Could not listen on server socket");

	struct sockaddr_in remoteIP;
	socklen_t addrSize = sizeof(remoteIP);
	int sock = accept(listenSock, (struct sockaddr*)&remoteIP, &addrSize);
	if (sock == -1) panic("Failed to accept remote connection");

	runNetProtocol(sock, &scheme, &scheme.funcR, false);

	close(sock);
	close(listenSock);
	cleanupScheme(&scheme);
	return 0;
}

int main(int argc, char** argv) {
	const char* progName = (argc > 0 ? argv[0] : "dakes-benchmark");
	if (argc < 2) exitUsage(progName);
	if (strcmp(argv[1], "init") == 0) return main_client(argc, argv);
	if (strcmp(argv[1], "resp") == 0) return main_server(argc, argv);
	if (strcmp(argv[1], "keygen") == 0) return main_sim(argc, argv, MEASURE_KEYGEN);
	if (strcmp(argv[1], "kex") == 0) return main_sim(argc, argv, MEASURE_KEX);
	if (strcmp(argv[1], "size") == 0) return main_sim(argc, argv, MEASURE_SIZE);
	exitUsage(progName);
}
