#include <stdbool.h>
#include <stdio.h>

#include "crypto.h"
#include "ecc.h"

static void printBigNum(bignum25519 num) {
	printf("{");
	for (int i = 0; i < 10; ++i) {
		if (i > 0) printf(",");
		printf("0x%08x", num[i]);
	}
	printf("}");
}

static bool isGenerator(ge25519* p) {
	ge25519 ALIGN(16) mul;
	ge25519_scalarmult_vartime(&mul, p, modm_m);

	// Adapted from ge25519_is_neutral_vartime
	static const unsigned char zero[32] = {0};
	unsigned char point_buffer[3][32];
	curve25519_contract(point_buffer[0], mul.x);
	curve25519_contract(point_buffer[1], mul.y);
	curve25519_contract(point_buffer[2], mul.z);
	return (memcmp(point_buffer[0], zero, 32) == 0) && (memcmp(point_buffer[1], point_buffer[2], 32) == 0);
}

static void pointFromString(ge25519* p, const char* str) {
	unsigned char hash[32] = {0};
	unsigned char hashBuffer[32] = {0};

	SHA3_256(hash, (const unsigned char*)str, strlen(str));
	uint64_t i = 0;
	while (true) {
		++i;
		ge25519_unpack_negative_vartime(p, hash);
		if (isGenerator(p)) break;
		SHA3_256(hashBuffer, hash, 32);
		memcpy(hash, hashBuffer, 32);
	}
	printf("Found generator in %lu hash steps\n\n", i);
}

int main(int argc, char** argv) {
	if (argc != 2) {
		fprintf(stderr, "Usage: dakes-makegens <string>\n\nCreates the Ed25519-donna cached tables for a group generator derived from the given string.\n");
		return 1;
	}

	ge25519 ALIGN(16) generator;
	pointFromString(&generator, argv[1]);

	printf("{\n");
	printf("\t"); printBigNum(generator.x); printf(",\n");
	printf("\t"); printBigNum(generator.y); printf(",\n");
	printf("\t"); printBigNum(generator.z); printf(",\n");
	printf("\t"); printBigNum(generator.t); printf("\n");
	printf("};\n\n");

	ge25519 ALIGN(16) multiple;
	ge25519_niels ALIGN(16) multipleNiels;

	printf("[32] = {\n");
	bignum256modm ALIGN(16) scalar = {0,0,0,0,0};
	for (uint64_t i = 1; i < 64; i += 2) {
		scalar[0] = i;
		ge25519_scalarmult_vartime(&multiple, &generator, scalar);
		ge25519_full_to_niels(&multipleNiels, &multiple);
		printf("\t{");
		printBigNum(multipleNiels.ysubx); printf(",");
		printBigNum(multipleNiels.xaddy); printf(",");
		printBigNum(multipleNiels.t2d);
		printf("},\n");
	}
	printf("};\n\n");

	bignum25519 ALIGN(16) invd;
	curve25519_recip(invd, ge25519_ecd);
	bignum256modm ALIGN(16) radix2 = {16*16,0,0,0,0};
	bignum256modm ALIGN(16) radixBase = {1,0,0,0,0};
	bignum256modm ALIGN(16) riModM = {0,0,0,0,0};
	bignum256modm ALIGN(16) ri16i = {1,0,0,0,0};

	unsigned char packed[96];
	printf("[256][96] = {\n");
	for (int i = 0; i < 64; i += 2) {
		for (int ri = 1; ri <= 8; ++ri) {
			printf("\t{");
			riModM[0] = (uint64_t)ri;
			mul256_modm(ri16i, radixBase, riModM);
			ge25519_scalarmult_vartime(&multiple, &generator, ri16i);
			ge25519_full_to_niels(&multipleNiels, &multiple);
			if (i == 0) curve25519_mul(multipleNiels.t2d, multipleNiels.t2d, invd);
			curve25519_contract(packed, multipleNiels.ysubx);
			curve25519_contract(&packed[32], multipleNiels.xaddy);
			curve25519_contract(&packed[64], multipleNiels.t2d);
			for (int i = 0; i < 96; ++i) {
				if (i > 0) printf(",");
				printf("0x%02x", packed[i]);
			}
			printf("},\n");
		}
		mul256_modm(radixBase, radixBase, radix2);
	}
	printf("};\n");

	return 0;
}
