#include <stdio.h>

#include <openssl/evp.h>
#include <openssl/err.h>

#include "crypto.h"

#include "ecc.h"
#include "eccrsh.h"
#include "eccrsh-internals.h"
#include "ecelgamal.h"
#include "ecelgamal-internals.h"
#include "dre.h"
#include "auth.h"
#include "secmem.h"
#include "qr.h"
#include "dakez.h"
#include "spawnplus.h"
#include "zdh.h"
#include "xzdh.h"
#include "other-ake.h"

static void printHex(const unsigned char* data, size_t len) {
	for (size_t i = 0; i < len; ++i) {
		printf("%02x", data[i]);
		if (i % 8 == 7) printf(" ");
	}
}

static void printDesc(const char* desc) {
	size_t l = strlen(desc);
	for (size_t i = l; i < 27; ++i) printf(" ");
	printf("%s: ", desc);
}

static void printBigNum(const bignum25519 num) {
	printf("{");
	for (int i = 0; i < 10; ++i) {
		if (i > 0) printf(",");
		printf("%08x", num[i]);
	}
	printf("}");
}

static void printPoint(const ge25519* p, const char* desc) {
	unsigned char packed[32];
	ge25519_pack(packed, p);
	printDesc(desc);
	printHex(packed, 32);
	printf("(t=");
	printBigNum(p->z);
	printf(")\n");
}

static void printBigNumModM(const bignum256modm num) {
	printf("{");
	for (int i = 0; i < 5; ++i) {
		if (i > 0) printf(",");
		printf("%lu", num[i]);
	}
	printf("}");
}

static void prepCanary(unsigned char* buffer, size_t size) {
	for (size_t i = 0; i < size; ++i) buffer[i] = 0xCE;
	buffer[size] = 0xDE;
	buffer[size+1] = 0xAD;
	buffer[size+2] = 0xBE;
	buffer[size+3] = 0xEF;
}

static bool checkCanary(unsigned char* buffer, size_t size, bool cleared) {
	if (cleared) {
		for (size_t i = 0; i < size; ++i) {
			if (buffer[i] != 0x00) return false;
		}
	}
	if (buffer[size] != 0xDE) return false;
	if (buffer[size+1] != 0xAD) return false;
	if (buffer[size+2] != 0xBE) return false;
	if (buffer[size+3] != 0xEF) return false;
	return true;
}

static void comparePoints(const ge25519* p1, const ge25519* p2, const bignum256modm s1, const bignum256modm s2, const char* desc) {
	unsigned char pack1[32];
	unsigned char pack2[32];
	ge25519_pack(pack1, p1);
	ge25519_pack(pack2, p2);
	if (memcmp(pack1, pack2, 32) != 0) {
		printf("\nComparison failure for %s!\n", desc);
		printDesc("Scalar 1"); printBigNumModM(s1); printf("\n");
		printDesc("Scalar 2"); printBigNumModM(s2); printf("\n");
		printPoint(p1, "Obtained");
		printPoint(p2, "Expected");
		exit(1);
	}
}

static void compareModM(const bignum256modm test1, const bignum256modm test2, const bignum256modm s1, const bignum256modm s2, const char* desc) {
	unsigned char pack1[32];
	unsigned char pack2[32];
	contract256_modm(pack1, test1);
	contract256_modm(pack2, test2);
	if (memcmp(pack1, pack2, 32) != 0) {
		printf("\nComparison failure for %s!\n", desc);
		printDesc("Scalar 1"); printBigNumModM(s1); printf("\n");
		printDesc("Scalar 2"); printBigNumModM(s2); printf("\n");
		printDesc("Obtained"); printBigNumModM(test1); printf("\n");
		printDesc("Expected"); printBigNumModM(test2); printf("\n");
		exit(1);
	}
}

static void cosmicRay(unsigned char* buffer, size_t bufLen, int strikeChance, const char* desc, bool* shouldSucceed, char* change) {
	if (!*shouldSucceed) return;
	if ((rand() % 100) + 1 > strikeChance) return;
	int shift = rand() % 8;
	size_t idx = (size_t)rand() % bufLen;
	unsigned char mask = (unsigned char)(1 << shift);
	buffer[idx] ^= mask;
	sprintf(change, "%s. Op: [%lu] ^= 0x%02X", desc, idx, mask);
	*shouldSucceed = false;
}

static void fuzz(void) {
	bignum256modm ALIGN(16) zero = {0,0,0,0,0};
	bignum256modm ALIGN(16) s1, s2, st;
	ge25519 ALIGN(16) r1, r2;
	ge25519 ALIGN(16) identity = {{0},{1,0},{1,0},{0}};
	unsigned char packBuffer[32];

	unsigned char* securePage = mallocSecurePage(2);
	qrKexNone->qrInit();
	qrKexSIDH->qrInit();
	qrKexNewHope->qrInit();

	uint64_t longCount = 0;
	for (int i = 0; true; ++i) {
		if ((i % 100) == 99) printf(".");
		if (i >= 9999) {
			i = 0;
			printf(" %lu0000\n", ++longCount);
		}

		// ECC tests
		{
			modm_rand(s1);
			modm_rand(s2);

			ge25519_scalarmult_base1(&r1, s1);
			ge25519_double_scalarmult_vartime(&r2, &ge25519_basepoint1, s1, zero);
			comparePoints(&r1, &r2, s1, s2, "s1*g1");

			ge25519_scalarmult_base2(&r1, s2);
			ge25519_double_scalarmult_vartime(&r2, &ge25519_basepoint2, s2, zero);
			comparePoints(&r1, &r2, s1, s2, "s2*g2");

			ge25519_scalarmult_base1(&r1, s1);
			ge25519_scalarmult_base2(&r2, s2);
			ge25519_add(&r1, &r1, &r2);
			ge25519_double_scalarmult_bases_vartime(&r2, s1, s2);
			comparePoints(&r1, &r2, s1, s2, "(Exp vs Var) s1*g1+s2*g2");

			ge25519_double_scalarmult_bases(&r1, s1, s2);
			comparePoints(&r1, &r2, s1, s2, "(Const vs Var) s1*g1+s2*g2");

			ge25519_neg(&r1, &r2); // Intentionally using vartime result
			ge25519_add(&r1, &r1, &r2);
			comparePoints(&r1, &identity, s1, s2, "(s1*g1+s2*g2)-(s1*g1+s2*g2)");

			ge25519_scalarmult_base1(&r1, s1);
			ge25519_scalarmult_vartime(&r1, &r1, s2);
			mul256_modm(st, s1, s2);
			ge25519_double_scalarmult_vartime(&r2, &ge25519_basepoint1, st, zero);
			comparePoints(&r1, &r2, s1, s2, "s2*(s1*g1)");

			ge25519_scalarmult_base1(&r2, s1);
			ge25519_scalarmult(&r2, &r2, s2);
			comparePoints(&r2, &r1, s1, s2, "s2*(s1*g1) with constant time scalarmult");

			ge25519_neg(&r2, &r1); // Intentionally using vartime result
			ge25519_add(&r2, &r1, &r2);
			comparePoints(&r2, &identity, s1, s2, "s2*(s1*g1)-s2*(s1*g1)");

			ge25519_scalarmult_base1(&r1, s1);
			ge25519_neg(&r2, &r1);
			ge25519_neg(&r2, &r2);
			comparePoints(&r1, &r2, s1, s2, "--(s1*g1)");

			ge25519_pack(packBuffer, &r1);
			ge25519_unpack_vartime(&r2, packBuffer);
			comparePoints(&r1, &r2, s1, s2, "Unpack(Pack(s1*g1))");

			modm_sub(st, s1, s1);
			compareModM(st, zero, s1, s2, "s1-s1");

			modm_neg(st, s1);
			add256_modm(st, st, s1);
			compareModM(st, zero, s1, s2, "s1+(-s1)");

			add256_modm(st, s1, s2);
			modm_sub(st, st, s2);
			compareModM(st, s1, s1, s2, "(s1+s2)-s2");
		}

		// ElGamal tests
		if ((i % 5) == 0) {
			ecelgamal_pubkey pk;
			ecelgamal_seckey sk;
			unsigned char pt[1024];
			size_t ptLen = rand() % 1024;
			RAND_bytes(pt, (int)ptLen);
			unsigned char ad[1024];
			size_t adLen = rand() % 1024;
			RAND_bytes(ad, (int)adLen);
			unsigned char nonce[AEAD_NONCE_SIZE];
			RAND_bytes(nonce, AEAD_NONCE_SIZE);
			unsigned char ct[ECELGAMAL_CT_OVERHEAD + ptLen];
			size_t ctUsed;
			unsigned char ptRecover[ptLen];
			size_t ptUsed;
			ecelgamal_genKeys(&pk, &sk);
			bool succeeded = true;
			if (!ecelgamal_encrypt(&pk, pt, ptLen, ad, adLen, nonce, ct, sizeof(ct), &ctUsed)) {
				printf("Failed to encrypt!\n");
				succeeded = false;
			} else if (!ecelgamal_decrypt(&sk, ct, sizeof(ct), ad, adLen, nonce, ptRecover, ptLen, &ptUsed)) {
				printf("Failed to decrypt\n");
				succeeded = false;
			} else if (!ed25519_verify(pt, ptRecover, ptLen)) {
				printf("Decryption didn't match plaintext!\n");
				succeeded = false;
			}
			if (!succeeded) {
				printf("*** ElGamal test failure ***\n");
				printDesc("Plaintext Len"); printf("%lu\n", ptLen);
				printDesc("AD Len"); printf("%lu\n", adLen);
				printDesc("Plaintext"); printHex(pt, ptLen); printf("\n");
				printDesc("AD"); printHex(ad, adLen); printf("\n");
				printDesc("Nonce"); printHex(nonce, AEAD_NONCE_SIZE); printf("\n");
				printDesc("Ciphertext"); printHex(ct, ctUsed); printf("\n");
				printPoint(&pk.h, "H");
				printDesc("X"); printBigNumModM(sk.x); printf("\n");
				exit(1);
			}
		}

		// Cramer-Shoup tests
		if ((i % 5) == 0) {
			eccrsh_pubkey pk;
			eccrsh_seckey sk;
			unsigned char pt[1024];
			size_t ptLen = rand() % 1024;
			RAND_bytes(pt, (int)ptLen);
			unsigned char ad[1024];
			size_t adLen = rand() % 1024;
			RAND_bytes(ad, (int)adLen);
			unsigned char ct[ECCRSH_CT_OVERHEAD + ptLen];
			size_t ctUsed;
			unsigned char ptRecover[ptLen];
			size_t ptUsed;
			eccrsh_genKeys(&pk, &sk);
			bool succeeded = true;
			if (!eccrsh_encrypt(&pk, pt, ptLen, ad, adLen, ct, sizeof(ct), &ctUsed)) {
				printf("Failed to encrypt!\n");
				succeeded = false;
			} else if (!eccrsh_decrypt(&sk, ct, sizeof(ct), ad, adLen, ptRecover, ptLen, &ptUsed)) {
				printf("Failed to decrypt\n");
				succeeded = false;
			} else if (!ed25519_verify(pt, ptRecover, ptLen)) {
				printf("Decryption didn't match plaintext!\n");
				succeeded = false;
			}
			if (!succeeded) {
				printf("*** Cramer-Shoup test failure ***\n");
				printDesc("Plaintext Len"); printf("%lu\n", ptLen);
				printDesc("AD Len"); printf("%lu\n", adLen);
				printDesc("Plaintext"); printHex(pt, ptLen); printf("\n");
				printDesc("AD"); printHex(ad, adLen); printf("\n");
				printDesc("Ciphertext"); printHex(ct, ctUsed); printf("\n");
				printPoint(&pk.c, "C");
				printPoint(&pk.d, "D");
				printPoint(&pk.h, "H");
				printDesc("X1"); printBigNumModM(sk.x1); printf("\n");
				printDesc("X2"); printBigNumModM(sk.x2); printf("\n");
				printDesc("Y1"); printBigNumModM(sk.y1); printf("\n");
				printDesc("Y2"); printBigNumModM(sk.y2); printf("\n");
				printDesc("Z"); printBigNumModM(sk.z); printf("\n");
				exit(1);
			}
		}

		// DRE tests
		if (i % 10 == 0) {
			ecelgamal_pubkey pk1, pk2;
			ecelgamal_seckey sk1, sk2;
			ecelgamal_genKeys(&pk1, &sk1);
			ecelgamal_genKeys(&pk2, &sk2);

			unsigned char pt[1024];
			size_t ptLen = rand() % 1024;
			RAND_bytes(pt, (int)ptLen);
			unsigned char ad[1024];
			size_t adLen = rand() % 1024;
			RAND_bytes(ad, (int)adLen);
			unsigned char ct[DRE_CT_OVERHEAD + ptLen];
			size_t ctUsed;
			unsigned char ptRecover[ptLen];
			size_t ptUsed;

			bool succeeded = true;
			if (!dre_encrypt(&pk1, &pk2, pt, ptLen, ad, adLen, ct, sizeof(ct), &ctUsed)) {
				printf("Failed to encrypt DRE!\n");
				succeeded = false;
			} else if (!dre_decrypt(&pk1, &pk2, DRE_KEY1, &sk1, ct, ctUsed, ad, adLen, ptRecover, sizeof(ptRecover), &ptUsed)) {
				printf("Failed to decrypt DRE with sk1!\n");
				succeeded = false;
			} else if (!dre_decrypt(&pk1, &pk2, DRE_KEY2, &sk2, ct, ctUsed, ad, adLen, ptRecover, sizeof(ptRecover), &ptUsed)) {
				printf("Failed to decrypt DRE with sk2!\n");
				succeeded = false;
			}
			if (!succeeded) {
				printf("*** DRE test failure ***\n");
				printDesc("Plaintext Len"); printf("%lu\n", ptLen);
				printDesc("AD Len"); printf("%lu\n", adLen);
				printDesc("Plaintext"); printHex(pt, ptLen); printf("\n");
				printDesc("AD"); printHex(ad, adLen); printf("\n");
				printDesc("Ciphertext"); printHex(ct, ctUsed); printf("\n");
				printPoint(&pk1.h, "H1");
				printPoint(&pk2.h, "H2");
				printDesc("X1"); printBigNumModM(sk1.x); printf("\n");
				printDesc("X2"); printBigNumModM(sk2.x); printf("\n");
				exit(1);
			}
		}

		// Auth tests
		if ((i % 5) == 0) {
			bignum256modm ALIGN(16) sk1, sk2, sk3;
			ge25519 ALIGN(16) pk1, pk2, pk3;
			modm_rand(sk1);
			modm_rand(sk2);
			modm_rand(sk3);
			ge25519_scalarmult_base1(&pk1, sk1);
			ge25519_scalarmult_base1(&pk2, sk2);
			ge25519_scalarmult_base1(&pk3, sk3);

			unsigned char msg[1024];
			size_t msgLen = rand() % 1024;
			RAND_bytes(msg, (int)msgLen);
			unsigned char ad[1024];
			size_t adLen = rand() % 1024;
			RAND_bytes(ad, (int)adLen);
			unsigned char implHashTag[1024];
			size_t ihtLen = rand() % 1024;
			RAND_bytes(implHashTag, (int)ihtLen);

			unsigned char proof[AUTH_PROOF_SIZE];
			unsigned char selectedKey = rand() % 3;
			unsigned char keyMask;
			bignum256modm* sk;
			switch (selectedKey) {
			case 0: keyMask = AUTH_KEY1; sk = &sk1; break;
			case 1: keyMask = AUTH_KEY2; sk = &sk2; break;
			case 2: keyMask = AUTH_KEY3; sk = &sk3; break;
			}

			bool succeeded = true;
			if (!auth_prove(ge25519_niels_base1_multiples, &pk1, &pk2, &pk3, keyMask, *sk, msg, msgLen, ad, adLen, implHashTag, ihtLen, proof)) {
				printf("Failed to create Auth proof!\n");
				succeeded = false;
			} else if (!auth_verify(ge25519_niels_base1_multiples, &pk1, &pk2, &pk3, proof, msg, msgLen, ad, adLen,implHashTag, ihtLen)) {
				printf("Failed to verify proof with Verif!\n");
				succeeded = false;
			}
			if (!succeeded) {
				printf("*** Auth/Verif test failure ***\n");
				printDesc("Message Len"); printf("%lu\n", msgLen);
				printDesc("AD Len"); printf("%lu\n", adLen);
				printDesc("IHT Len"); printf("%lu\n", ihtLen);
				printDesc("Message"); printHex(msg, msgLen); printf("\n");
				printDesc("AD"); printHex(ad, adLen); printf("\n");
				printDesc("IHT"); printHex(implHashTag, ihtLen); printf("\n");
				printDesc("Selected key"); printf("%d\n", selectedKey);
				printDesc("Proof"); printHex(proof, AUTH_PROOF_SIZE); printf("\n");
				printPoint(&pk1, "pk1");
				printPoint(&pk2, "pk2");
				printPoint(&pk3, "pk3");
				printDesc("sk1"); printBigNumModM(sk1); printf("\n");
				printDesc("sk2"); printBigNumModM(sk2); printf("\n");
				printDesc("sk3"); printBigNumModM(sk3); printf("\n");
				exit(1);
			}
		}

		// Key exchange tests
		if ((i % 20) == 0) {
			qrKex* qrKex;
			switch (rand() % 3) {
			case 0: qrKex = qrKexNone; break;
			case 1: qrKex = qrKexSIDH; break;
			case 2: qrKex = qrKexNewHope; break;
			}

			char change[1024];
			const char* problem = "(unknown)";

			// DAKEZ
			{
				bool shouldSucceed = true;

				unsigned char ad[1024];
				size_t adLen = rand() % 1024;
				RAND_bytes(ad, (int)adLen);
				unsigned char idI[1024];
				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idI, (int)idLen);
				RAND_bytes(idR, (int)idLen);

				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + DAKEZ_SK_SIZE;
				unsigned char* fskI = skR+4 + DAKEZ_SK_SIZE;
				unsigned char* fskR = fskI+4 + DAKEZ_SK_SIZE;
				unsigned char* secretStateI = fskR+4 + DAKEZ_SK_SIZE;
				secretStateI += (16 - ((size_t)secretStateI % 16)) % 16;
				unsigned char* secretStateR = secretStateI+4 + DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI;
				secretStateR += (16 - ((size_t)secretStateR % 16)) % 16;
				unsigned char* keyI = secretStateR+4 + DAKEZ_SECRETSTATE_R_SIZE;
				unsigned char* keyR = keyI+4 + DAKEZ_KEY_SIZE;
				unsigned char* offlineForgeKey = keyR+4 + DAKEZ_KEY_SIZE;
				unsigned char* scratch = offlineForgeKey+4 + DAKEZ_SK_SIZE;
				scratch += (16 - ((size_t)scratch % 16)) % 16;

				prepCanary(skI, DAKEZ_SK_SIZE);
				prepCanary(skR, DAKEZ_SK_SIZE);
				prepCanary(fskI, DAKEZ_SK_SIZE);
				prepCanary(fskR, DAKEZ_SK_SIZE);
				prepCanary(secretStateI, DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI);
				prepCanary(secretStateR, DAKEZ_SECRETSTATE_R_SIZE);
				prepCanary(keyI, DAKEZ_KEY_SIZE);
				prepCanary(keyR, DAKEZ_KEY_SIZE);
				prepCanary(offlineForgeKey, DAKEZ_SK_SIZE);

				unsigned char pkI[DAKEZ_PK_SIZE];
				unsigned char pkR[DAKEZ_PK_SIZE];
				unsigned char fpkI[DAKEZ_PK_SIZE];
				unsigned char fpkR[DAKEZ_PK_SIZE];

				unsigned char ALIGN(16) publicStateI[DAKEZ_PUBLICSTATE_I_OVERHEAD + qrKex->sizePqI];
				unsigned char ALIGN(16) publicStateR[DAKEZ_PUBLICSTATE_R_OVERHEAD + qrKex->sizePqI + qrKex->sizeQR];

				unsigned char message1[DAKEZ_1_SEND_I_OVERHEAD + qrKex->sizePqI + idLen];
				unsigned char message2[DAKEZ_2_SEND_R_OVERHEAD + qrKex->sizeQR + idLen];
				unsigned char message3[DAKEZ_3_SEND_I_SIZE];

				unsigned char receivedIdI[idLen];
				unsigned char receivedIdR[idLen];

				unsigned char signMode;
				const unsigned char* signSecret;

				bool succeeded = false;
				size_t scratchSize = 0;

				// Generate key pair for I
				scratchSize = DAKEZ_SCRATCH_KEYGEN_SIZE;
				prepCanary(scratch, scratchSize);
				if (dakez_genLongTermKeys(pkI, skI, fpkI, fskI, scratch) != DAKEZ_SUCCESS) {
					problem = "Failed to generate key pair I";
					goto dakezCleanup;
				}
				if (   !checkCanary(scratch, scratchSize, true)
				    || !checkCanary(skI, DAKEZ_SK_SIZE, false)
				    || !checkCanary(fskI, DAKEZ_SK_SIZE, false)) {
					problem = "Memory corruption from dakez_genLongTermKeys for I";
					goto dakezCleanup;
				}
				scratchSize = 0;

				// Generate key pair for R
				scratchSize = DAKEZ_SCRATCH_KEYGEN_SIZE;
				prepCanary(scratch, scratchSize);
				if (dakez_genLongTermKeys(pkR, skR, fpkR, fskR, scratch) != DAKEZ_SUCCESS) {
					problem = "Failed to generate key pair R";
					goto dakezCleanup;
				}
				if (   !checkCanary(scratch, scratchSize, true)
				    || !checkCanary(skR, DAKEZ_SK_SIZE, false)
				    || !checkCanary(fskR, DAKEZ_SK_SIZE, false)) {
					problem = "Memory corruption from dakez_genLongTermKeys for R";
					goto dakezCleanup;
				}
				scratchSize = 0;

				// Generate first message
				scratchSize = qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (dakez_1_initiate_I(qrKex, idLen, idI, publicStateI, secretStateI, message1, scratch) != DAKEZ_SUCCESS) {
					problem = "Failed to generate message 1";
					goto dakezCleanup;
				}
				if (   !checkCanary(secretStateI, DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI, false)
				    || !checkCanary(scratch, scratchSize, false)) {
					problem = "Memory corruption from dakez_1_initiate_I";
					goto dakezCleanup;
				}
				scratchSize = 0;

				cosmicRay(message1, DAKEZ_1_SEND_I_OVERHEAD + qrKex->sizePqI + idLen, 10, "Alter message 1", &shouldSucceed, change);

				// Receive first message
				if (dakez_2_receive_R(message1, idLen, receivedIdI) != DAKEZ_SUCCESS) {
					problem = "Failed to extract ID from message 1";
					goto dakezCleanup;
				}
				if (memcmp(idI, receivedIdI, idLen) != 0) {
					problem = "Extracted ID from message 1 was incorrect";
					goto dakezCleanup;
				}

				// Generate second message
				bool offlineForging = false;
				switch (rand() % 3) {
				case 0:
					signMode = DAKEZ_HONEST_R;
					signSecret = skR;
					break;
				case 1:
					dakez_forgeOffline1(secretStateI, offlineForgeKey);
					if (!checkCanary(offlineForgeKey, DAKEZ_SK_SIZE, false)) {
						problem = "Memory corruption from dakez_forgeOffline1";
						goto dakezCleanup;
					}
					signMode = DAKEZ_OFFLINE_FORGE;
					signSecret = offlineForgeKey;
					offlineForging = true;
					break;
				case 2:
					signMode = DAKEZ_ONLINE_FORGE_R;
					signSecret = fskI;
					break;
				}
				scratchSize = DAKEZ_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (dakez_2_respond_R(message1, qrKex, idLen, idR, fpkI, pkR, signMode, signSecret, ad, adLen, publicStateR, secretStateR, message2, scratch) != DAKEZ_SUCCESS) {
					problem = "Failed to generate message 2";
					goto dakezCleanup;
				}
				if (   !checkCanary(secretStateR, DAKEZ_SECRETSTATE_R_SIZE, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from dakez_2_respond_R";
					goto dakezCleanup;
				}
				scratchSize = 0;

				cosmicRay(message2, DAKEZ_2_SEND_R_OVERHEAD + qrKex->sizeQR + idLen, 10, "Alter message 2", &shouldSucceed, change);

				// Receive second message
				if (dakez_3_receive_I(message2, idLen, receivedIdR) != DAKEZ_SUCCESS) {
					problem = "Failed to extract ID from message 2";
					goto dakezCleanup;
				}
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from message 2 was incorrect";
					goto dakezCleanup;
				}

				// Generate third message
				if (offlineForging) {
					dakez_forgeOffline2(secretStateR, offlineForgeKey);
					if (!checkCanary(offlineForgeKey, DAKEZ_SK_SIZE, false)) {
						problem = "Memory corruption from dakez_forgeOffline2";
						goto dakezCleanup;
					}
					signMode = DAKEZ_OFFLINE_FORGE;
					signSecret = offlineForgeKey;
				} else {
					switch (rand() % 2) {
					case 0:
						signMode = DAKEZ_HONEST_I;
						signSecret = skI;
						break;
					case 1:
						signMode = DAKEZ_ONLINE_FORGE_I;
						signSecret = fskR;
						break;
					}
				}
				scratchSize = DAKEZ_SCRATCH_I_RESPOND_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (dakez_3_respond_I(message2, qrKex, idLen, idI, pkI, fpkI, pkR, fpkR, signMode, signSecret, ad, adLen, publicStateI, secretStateI, message3, keyI, scratch) != DAKEZ_SUCCESS) {
					problem = "Failed to generate message 3";
					goto dakezCleanup;
				}
				if (   !checkCanary(secretStateI, DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI, true)
				    || !checkCanary(keyI, DAKEZ_KEY_SIZE, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from dakez_3_respond_I";
					goto dakezCleanup;
				}
				scratchSize = 0;

				cosmicRay(message3, DAKEZ_3_SEND_I_SIZE, 10, "Alter message 3", &shouldSucceed, change);

				// Receive third message
				if (dakez_4_verify_R(message3, qrKex, idLen, receivedIdI, idR, pkI, fpkR, ad, adLen, publicStateR, secretStateR, keyR) != DAKEZ_SUCCESS) {
					problem = "Failed to verify message 3";
					goto dakezCleanup;
				}
				if (   !checkCanary(secretStateR, DAKEZ_SECRETSTATE_R_SIZE, true)
				    || !checkCanary(keyR, DAKEZ_KEY_SIZE, false)) {
					problem = "Memory corruption from dakez_4_verify_R";
					goto dakezCleanup;
				}

				if (memcmp(keyI, keyR, DAKEZ_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto dakezCleanup;
				}

				succeeded = true;
dakezCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** DAKEZ test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("AD Len"); printf("%lu\n", adLen);
					printDesc("AD"); printHex(ad, adLen); printf("\n");
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID I"); printHex(idI, idLen); printf("\n");
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, DAKEZ_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, DAKEZ_PK_SIZE); printf("\n");
					printDesc("FPK I"); printHex(fpkI, DAKEZ_PK_SIZE); printf("\n");
					printDesc("FPK R"); printHex(fpkR, DAKEZ_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, DAKEZ_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, DAKEZ_SK_SIZE+4); printf("\n");
					printDesc("FSK I + canary"); printHex(fskI, DAKEZ_SK_SIZE+4); printf("\n");
					printDesc("FSK R + canary"); printHex(fskR, DAKEZ_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, DAKEZ_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, DAKEZ_KEY_SIZE+4); printf("\n");
					printDesc("Secret State I + canary"); printHex(secretStateI, DAKEZ_SECRETSTATE_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI + 4); printf("\n");
					printDesc("Secret State R + canary"); printHex(secretStateR, DAKEZ_SECRETSTATE_R_SIZE + 4); printf("\n");
					printDesc("Signing mode"); printf("0x%02X\n", signMode);
					if (scratchSize > 0) {
						printDesc("Scratch Space + canary"); printHex(scratch, scratchSize+4); printf("\n");
					}
					exit(1);
				}
			}

			// Spawn+
			{
				bool shouldSucceed = true;

				unsigned char ad[1024];
				size_t adLen = rand() % 1024;
				RAND_bytes(ad, (int)adLen);
				unsigned char idI[1024];
				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idI, (int)idLen);
				RAND_bytes(idR, (int)idLen);

				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + SPAWNPLUS_SK_SIZE;
				unsigned char* fskI = skR+4 + SPAWNPLUS_SK_SIZE;
				unsigned char* fskR = fskI+4 + SPAWNPLUS_SK_SIZE;
				unsigned char* secretState = fskR+4 + SPAWNPLUS_SK_SIZE;
				secretState += (16 - ((size_t)secretState % 16)) % 16;
				unsigned char* keyI = secretState+4 + SPAWNPLUS_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI;
				unsigned char* keyR = keyI+4 + SPAWNPLUS_KEY_SIZE;
				unsigned char* offlineForgeKey = keyR+4 + SPAWNPLUS_KEY_SIZE;
				unsigned char* scratch = offlineForgeKey+4 + SPAWNPLUS_SK_SIZE;
				scratch += (16 - ((size_t)scratch % 16)) % 16;

				prepCanary(skI, SPAWNPLUS_SK_SIZE);
				prepCanary(skR, SPAWNPLUS_SK_SIZE);
				prepCanary(fskI, SPAWNPLUS_SK_SIZE);
				prepCanary(fskR, SPAWNPLUS_SK_SIZE);
				prepCanary(secretState, SPAWNPLUS_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI);
				prepCanary(keyI, SPAWNPLUS_KEY_SIZE);
				prepCanary(keyR, SPAWNPLUS_KEY_SIZE);
				prepCanary(offlineForgeKey, SPAWNPLUS_SK_SIZE);

				unsigned char pkI[SPAWNPLUS_PK_SIZE];
				unsigned char pkR[SPAWNPLUS_PK_SIZE];
				unsigned char fpkI[SPAWNPLUS_PK_SIZE];
				unsigned char fpkR[SPAWNPLUS_PK_SIZE];

				unsigned char ALIGN(16) publicState[SPAWNPLUS_PUBLICSTATE_OVERHEAD + qrKex->sizePqI];

				unsigned char prekey[SPAWNPLUS_PREKEY_OVERHEAD + qrKex->sizePqI + 4];
				unsigned char response[SPAWNPLUS_RESP_OVERHEAD + qrKex->sizeQR + idLen + 4];
				prepCanary(prekey, SPAWNPLUS_PREKEY_OVERHEAD + qrKex->sizePqI);
				prepCanary(response, SPAWNPLUS_RESP_OVERHEAD + qrKex->sizeQR + idLen);

				unsigned char receivedIdR[idLen];

				unsigned char signMode;
				const unsigned char* signSecret;
				unsigned char finishMode;
				const unsigned char* finishSecret;

				bool succeeded = false;
				size_t scratchSize = 0;

				// Generate key pair for I
				scratchSize = SPAWNPLUS_SCRATCH_KEYGEN_SIZE;
				prepCanary(scratch, scratchSize);
				if (spawnplus_genLongTermKeys(pkI, skI, fpkI, fskI, scratch) != SPAWNPLUS_SUCCESS) {
					problem = "Failed to generate key pair I";
					goto spawnPlusCleanup;
				}
				if (   !checkCanary(scratch, scratchSize, true)
				    || !checkCanary(skI, SPAWNPLUS_SK_SIZE, false)
				    || !checkCanary(fskI, SPAWNPLUS_SK_SIZE, false)) {
					problem = "Memory corruption from spawnplus_genLongTermKeys for I";
					goto spawnPlusCleanup;
				}
				scratchSize = 0;

				// Generate key pair for R
				scratchSize = SPAWNPLUS_SCRATCH_KEYGEN_SIZE;
				prepCanary(scratch, scratchSize);
				if (spawnplus_genLongTermKeys(pkR, skR, fpkR, fskR, scratch) != SPAWNPLUS_SUCCESS) {
					problem = "Failed to generate key pair R";
					goto spawnPlusCleanup;
				}
				if (   !checkCanary(scratch, scratchSize, true)
				    || !checkCanary(skR, SPAWNPLUS_SK_SIZE, false)
				    || !checkCanary(fskR, SPAWNPLUS_SK_SIZE, false)) {
					problem = "Memory corruption from spawnplus_genLongTermKeys for R";
					goto spawnPlusCleanup;
				}
				scratchSize = 0;

				// Generate prekey
				scratchSize = qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (spawnplus_1_makePreKey_I(qrKex, publicState, secretState, prekey, scratch) != SPAWNPLUS_SUCCESS) {
					problem = "Failed to generate prekey";
					goto spawnPlusCleanup;
				}
				if (   !checkCanary(secretState, SPAWNPLUS_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI, false)
				    || !checkCanary(prekey, SPAWNPLUS_PREKEY_OVERHEAD + qrKex->sizePqI, false)
				    || !checkCanary(scratch, scratchSize, false)) {
					problem = "Memory corruption from spawnplus_1_makePreKey_I";
					goto spawnPlusCleanup;
				}
				scratchSize = 0;

				cosmicRay(prekey, SPAWNPLUS_PREKEY_OVERHEAD + qrKex->sizePqI, 15, "Alter prekey", &shouldSucceed, change);

				// Generate response
				switch (rand() % 3) {
				case 0:
					signMode = SPAWNPLUS_RESP_HONEST;
					signSecret = skR;
					break;
				case 1:
					spawnplus_forgeOffline(secretState, offlineForgeKey);
					if (!checkCanary(offlineForgeKey, SPAWNPLUS_SK_SIZE, false)) {
						problem = "Memory corruption from spawnplus_forgeOffline";
						goto spawnPlusCleanup;
					}
					signMode = SPAWNPLUS_RESP_OFFLINE_FORGE;
					signSecret = offlineForgeKey;
					break;
				case 2:
					signMode = SPAWNPLUS_RESP_ONLINE_FORGE;
					signSecret = fskI;
					break;
				}
				scratchSize = SPAWNPLUS_SCRATCH_R_OVERHEAD + qrKex->sizeQR + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (spawnplus_2_exchange_R(prekey, qrKex, idLen, idI, idR, pkI, fpkI, pkR, fpkR, signMode, signSecret, ad, adLen, response, keyR, scratch) != SPAWNPLUS_SUCCESS) {
					problem = "Failed to generate response";
					goto spawnPlusCleanup;
				}
				if (   !checkCanary(response, SPAWNPLUS_RESP_OVERHEAD + qrKex->sizeQR + idLen, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from spawnplus_2_exchange_R";
					goto spawnPlusCleanup;
				}
				scratchSize = 0;

				cosmicRay(response, SPAWNPLUS_RESP_OVERHEAD + qrKex->sizeQR + idLen, 15, "Alter response", &shouldSucceed, change);

				// Receive response
				if (spawnplus_3_receive_I(response, idLen, receivedIdR) != SPAWNPLUS_SUCCESS) {
					problem = "Failed to extract ID from response";
					goto spawnPlusCleanup;
				}
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from response was incorrect";
					goto spawnPlusCleanup;
				}

				// Complete exchange
				switch (rand() % 2) {
				case 0:
					finishMode = SPAWNPLUS_FINISH_HONEST;
					finishSecret = skI;
					break;
				case 1:
					finishMode = SPAWNPLUS_FINISH_ONLINE_FORGE;
					finishSecret = fskR;
					break;
				}
				scratchSize = SPAWNPLUS_SCRATCH_I_OVERHEAD + qrKex->sizeQR + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (spawnplus_3_exchange_I(response, qrKex, idLen, idI, pkI, fpkI, pkR, fpkR, finishMode, finishSecret, ad, adLen, publicState, secretState, keyI, scratch) != SPAWNPLUS_SUCCESS) {
					problem = "Failed to complete exchange";
					goto spawnPlusCleanup;
				}
				if (   !checkCanary(secretState, SPAWNPLUS_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI, true)
				    || !checkCanary(keyI, SPAWNPLUS_KEY_SIZE, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from spawnplus_3_exchange_I";
					goto spawnPlusCleanup;
				}
				scratchSize = 0;

				if (memcmp(keyI, keyR, SPAWNPLUS_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto spawnPlusCleanup;
				}

				succeeded = true;
spawnPlusCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** Spawn+ test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("AD Len"); printf("%lu\n", adLen);
					printDesc("AD"); printHex(ad, adLen); printf("\n");
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID I"); printHex(idI, idLen); printf("\n");
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, SPAWNPLUS_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, SPAWNPLUS_PK_SIZE); printf("\n");
					printDesc("FPK I"); printHex(fpkI, SPAWNPLUS_PK_SIZE); printf("\n");
					printDesc("FPK R"); printHex(fpkR, SPAWNPLUS_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, SPAWNPLUS_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, SPAWNPLUS_SK_SIZE+4); printf("\n");
					printDesc("FSK I + canary"); printHex(fskI, SPAWNPLUS_SK_SIZE+4); printf("\n");
					printDesc("FSK R + canary"); printHex(fskR, SPAWNPLUS_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, SPAWNPLUS_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, SPAWNPLUS_KEY_SIZE+4); printf("\n");
					printDesc("Secret State + canary"); printHex(secretState, SPAWNPLUS_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI + 4); printf("\n");
					printDesc("Signing mode"); printf("0x%02X\n", signMode);
					printDesc("Finish mode"); printf("0x%02X\n", finishMode);
					if (scratchSize > 0) {
						printDesc("Scratch Space + canary"); printHex(scratch, scratchSize+4); printf("\n");
					}
					exit(1);
				}
			}

			// ZDH
			{
				bool shouldSucceed = true;

				unsigned char ad[1024];
				size_t adLen = rand() % 1024;
				RAND_bytes(ad, (int)adLen);
				unsigned char idI[1024];
				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idI, (int)idLen);
				RAND_bytes(idR, (int)idLen);
				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + ZDH_SK_SIZE;
				unsigned char* fskI = skR+4 + ZDH_SK_SIZE;
				unsigned char* fskR = fskI+4 + ZDH_SK_SIZE;
				unsigned char* secretState = fskR+4 + ZDH_SK_SIZE;
				secretState += (16 - ((size_t)secretState % 16)) % 16;
				unsigned char* keyI = secretState+4 + ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI;
				unsigned char* keyR = keyI+4 + ZDH_KEY_SIZE;
				unsigned char* offlineForgeKey = keyR+4 + ZDH_KEY_SIZE;
				unsigned char* scratch = offlineForgeKey+4 + ZDH_SK_SIZE;
				scratch += (16 - ((size_t)scratch % 16)) % 16;

				prepCanary(skI, ZDH_SK_SIZE);
				prepCanary(skR, ZDH_SK_SIZE);
				prepCanary(fskI, ZDH_SK_SIZE);
				prepCanary(fskR, ZDH_SK_SIZE);
				prepCanary(secretState, ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI);
				prepCanary(keyI, ZDH_KEY_SIZE);
				prepCanary(keyR, ZDH_KEY_SIZE);
				prepCanary(offlineForgeKey, ZDH_SK_SIZE);

				unsigned char pkI[ZDH_PK_SIZE];
				unsigned char pkR[ZDH_PK_SIZE];
				unsigned char fpkI[ZDH_PK_SIZE];
				unsigned char fpkR[ZDH_PK_SIZE];

				unsigned char ALIGN(16) publicState[ZDH_PUBLICSTATE_OVERHEAD + qrKex->sizePqI];

				unsigned char prekey[ZDH_PREKEY_OVERHEAD + qrKex->sizePqI + 4];
				unsigned char response[ZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen + 4];
				prepCanary(prekey, ZDH_PREKEY_OVERHEAD + qrKex->sizePqI);
				prepCanary(response, ZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen);

				unsigned char receivedIdR[idLen];

				unsigned char signMode;
				const unsigned char* signSecret;

				bool succeeded = false;
				size_t scratchSize = 0;

				// Generate key pair for I
				scratchSize = ZDH_SCRATCH_KEYGEN_SIZE;
				prepCanary(scratch, scratchSize);
				if (zdh_genLongTermKeys(pkI, skI, fpkI, fskI, scratch) != ZDH_SUCCESS) {
					problem = "Failed to generate key pair I";
					goto zdhCleanup;
				}
				if (   !checkCanary(scratch, scratchSize, true)
				    || !checkCanary(skI, ZDH_SK_SIZE, false)
				    || !checkCanary(fskI, ZDH_SK_SIZE, false)) {
					problem = "Memory corruption from zdh_genLongTermKeys for I";
					goto zdhCleanup;
				}
				scratchSize = 0;

				// Generate key pair for R
				scratchSize = ZDH_SCRATCH_KEYGEN_SIZE;
				prepCanary(scratch, scratchSize);
				if (zdh_genLongTermKeys(pkR, skR, fpkR, fskR, scratch) != ZDH_SUCCESS) {
					problem = "Failed to generate key pair R";
					goto zdhCleanup;
				}
				if (   !checkCanary(scratch, scratchSize, true)
				    || !checkCanary(skR, ZDH_SK_SIZE, false)
				    || !checkCanary(fskR, ZDH_SK_SIZE, false)) {
					problem = "Memory corruption from zdh_genLongTermKeys for R";
					goto zdhCleanup;
				}
				scratchSize = 0;

				// Generate prekey
				scratchSize = qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (zdh_1_makePreKey_I(qrKex, publicState, secretState, prekey, scratch) != ZDH_SUCCESS) {
					problem = "Failed to generate prekey";
					goto zdhCleanup;
				}
				if (   !checkCanary(secretState, ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI, false)
				    || !checkCanary(prekey, ZDH_PREKEY_OVERHEAD + qrKex->sizePqI, false)
				    || !checkCanary(scratch, scratchSize, false)) {
					problem = "Memory corruption from zdh_1_makePreKey_I";
					goto zdhCleanup;
				}
				scratchSize = 0;

				cosmicRay(prekey, ZDH_PREKEY_OVERHEAD + qrKex->sizePqI, 15, "Alter prekey", &shouldSucceed, change);

				// Generate response
				switch (rand() % 3) {
				case 0:
					signMode = ZDH_HONEST;
					signSecret = skR;
					break;
				case 1:
					zdh_forgeOffline(secretState, offlineForgeKey);
					if (!checkCanary(offlineForgeKey, ZDH_SK_SIZE, false)) {
						problem = "Memory corruption from zdh_forgeOffline";
						goto zdhCleanup;
					}
					signMode = ZDH_OFFLINE_FORGE;
					signSecret = offlineForgeKey;
					break;
				case 2:
					signMode = ZDH_ONLINE_FORGE;
					signSecret = fskI;
					break;
				}
				scratchSize = ZDH_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (zdh_2_exchange_R(prekey, qrKex, idLen, idI, idR, pkI, fpkI, pkR, signMode, signSecret, ad, adLen, response, keyR, scratch) != ZDH_SUCCESS) {
					problem = "Failed to generate response";
					goto zdhCleanup;
				}
				if (   !checkCanary(response, ZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from zdh_2_exchange_R";
					goto zdhCleanup;
				}
				scratchSize = 0;

				cosmicRay(response, ZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen, 15, "Alter response", &shouldSucceed, change);

				// Receive response
				if (zdh_3_receive_I(response, idLen, receivedIdR) != ZDH_SUCCESS) {
					problem = "Failed to extract ID from response";
					goto zdhCleanup;
				}
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from response was incorrect";
					goto zdhCleanup;
				}

				// Complete exchange
				scratchSize = ZDH_SCRATCH_I_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (zdh_3_exchange_I(response, qrKex, idLen, idI, pkI, fpkI, pkR, skI, ad, adLen, publicState, secretState, keyI, scratch) != ZDH_SUCCESS) {
					problem = "Failed to complete exchange";
					goto zdhCleanup;
				}
				if (   !checkCanary(secretState, ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI, true)
				    || !checkCanary(keyI, ZDH_KEY_SIZE, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from zdh_3_exchange_I";
					goto zdhCleanup;
				}
				scratchSize = 0;

				if (memcmp(keyI, keyR, ZDH_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto zdhCleanup;
				}

				succeeded = true;
zdhCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** ZDH test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("AD Len"); printf("%lu\n", adLen);
					printDesc("AD"); printHex(ad, adLen); printf("\n");
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID I"); printHex(idI, idLen); printf("\n");
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, ZDH_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, ZDH_PK_SIZE); printf("\n");
					printDesc("FPK I"); printHex(fpkI, ZDH_PK_SIZE); printf("\n");
					printDesc("FPK R"); printHex(fpkR, ZDH_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, ZDH_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, ZDH_SK_SIZE+4); printf("\n");
					printDesc("FSK I + canary"); printHex(fskI, ZDH_SK_SIZE+4); printf("\n");
					printDesc("FSK R + canary"); printHex(fskR, ZDH_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, ZDH_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, ZDH_KEY_SIZE+4); printf("\n");
					printDesc("Secret State + canary"); printHex(secretState, ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI + 4); printf("\n");
					printDesc("Signing mode"); printf("0x%02X\n", signMode);
					if (scratchSize > 0) {
						printDesc("Scratch Space + canary"); printHex(scratch, scratchSize+4); printf("\n");
					}
					exit(1);
				}
			}

			// XZDH
			{
				bool shouldSucceed = true;

				unsigned char ad[1024];
				size_t adLen = rand() % 1024;
				RAND_bytes(ad, (int)adLen);
				unsigned char idI[1024];
				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idI, (int)idLen);
				RAND_bytes(idR, (int)idLen);
				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + XZDH_SK_SIZE;
				unsigned char* fskI = skR+4 + XZDH_SK_SIZE;
				unsigned char* fskR = fskI+4 + XZDH_SK_SIZE;
				unsigned char* secretSignedState = fskR+4 + XZDH_SK_SIZE;
				secretSignedState += (16 - ((size_t)secretSignedState % 16)) % 16;
				unsigned char* secretOneTimeState = secretSignedState+4 + XZDH_SIGNED_SECRETSTATE_SIZE;
				secretOneTimeState += (16 - ((size_t)secretOneTimeState % 16)) % 16;
				unsigned char* keyI = secretOneTimeState+4 + XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI;
				unsigned char* keyR = keyI+4 + XZDH_KEY_SIZE;
				unsigned char* offlineForgeKey = keyR+4 + XZDH_KEY_SIZE;
				unsigned char* scratch = offlineForgeKey+4 + XZDH_SK_SIZE;
				scratch += (16 - ((size_t)scratch % 16)) % 16;

				prepCanary(skI, XZDH_SK_SIZE);
				prepCanary(skR, XZDH_SK_SIZE);
				prepCanary(fskI, XZDH_SK_SIZE);
				prepCanary(fskR, XZDH_SK_SIZE);
				prepCanary(secretSignedState, XZDH_SIGNED_SECRETSTATE_SIZE);
				prepCanary(secretOneTimeState, XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI);
				prepCanary(keyI, XZDH_KEY_SIZE);
				prepCanary(keyR, XZDH_KEY_SIZE);
				prepCanary(offlineForgeKey, XZDH_SK_SIZE);

				unsigned char pkI[XZDH_PK_SIZE];
				unsigned char pkR[XZDH_PK_SIZE];
				unsigned char fpkI[XZDH_PK_SIZE];
				unsigned char fpkR[XZDH_PK_SIZE];

				unsigned char publicSignedState[XZDH_SIGNED_PUBLICSTATE_SIZE];
				unsigned char ALIGN(16) publicOneTimeState[XZDH_ONETIME_PUBLICSTATE_OVERHEAD + qrKex->sizePqI];

				unsigned char signedPrekey[XZDH_SIGNED_PREKEY_SIZE + 4];
				unsigned char oneTimePrekey[XZDH_ONETIME_PREKEY_OVERHEAD + qrKex->sizePqI + 4];
				unsigned char response[XZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen + 4];
				prepCanary(signedPrekey, XZDH_SIGNED_PREKEY_SIZE);
				prepCanary(oneTimePrekey, XZDH_ONETIME_PREKEY_OVERHEAD + qrKex->sizePqI);
				prepCanary(response, XZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen);

				unsigned char receivedIdR[idLen];

				unsigned char signMode;
				const unsigned char* signSecret;

				bool succeeded = false;
				size_t scratchSize = 0;

				// Generate key pair for I
				if (xzdh_genLongTermKeys(pkI, skI, fpkI, fskI) != XZDH_SUCCESS) {
					problem = "Failed to generate key pair I";
					goto xzdhCleanup;
				}
				if (   !checkCanary(skI, XZDH_SK_SIZE, false)
				    || !checkCanary(fskI, XZDH_SK_SIZE, false)) {
					problem = "Memory corruption from xzdh_genLongTermKeys for I";
					goto xzdhCleanup;
				}

				// Generate key pair for R
				if (xzdh_genLongTermKeys(pkR, skR, fpkR, fskR) != XZDH_SUCCESS) {
					problem = "Failed to generate key pair R";
					goto xzdhCleanup;
				}
				if (   !checkCanary(skR, XZDH_SK_SIZE, false)
				    || !checkCanary(fskR, XZDH_SK_SIZE, false)) {
					problem = "Memory corruption from xzdh_genLongTermKeys for R";
					goto xzdhCleanup;
				}

				// Generate signed prekey
				if (xzdh_genSignedPrekey(pkI, skI, publicSignedState, secretSignedState, signedPrekey) != XZDH_SUCCESS) {
					problem = "Failed to generate signed prekey";
					goto xzdhCleanup;
				}
				if (   !checkCanary(secretSignedState, XZDH_SIGNED_SECRETSTATE_SIZE, false)
				    || !checkCanary(signedPrekey, XZDH_SIGNED_PREKEY_SIZE, false)) {
					problem = "Memory corruption from xzdh_1_makeOneTimePrekey_I";
					goto xzdhCleanup;
				}

				cosmicRay(signedPrekey, XZDH_SIGNED_PREKEY_SIZE, 10, "Alter signed prekey", &shouldSucceed, change);

				// Generate one-time prekey
				scratchSize = qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (xzdh_1_makeOneTimePrekey_I(qrKex, publicOneTimeState, secretOneTimeState, oneTimePrekey, scratch) != XZDH_SUCCESS) {
					problem = "Failed to generate one-time prekey";
					goto xzdhCleanup;
				}
				if (   !checkCanary(secretOneTimeState, XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI, false)
				    || !checkCanary(oneTimePrekey, XZDH_ONETIME_PREKEY_OVERHEAD + qrKex->sizePqI, false)
				    || !checkCanary(scratch, scratchSize, false)) {
					problem = "Memory corruption from xzdh_1_makeOneTimePrekey_I";
					goto xzdhCleanup;
				}
				scratchSize = 0;

				cosmicRay(oneTimePrekey, XZDH_ONETIME_PREKEY_OVERHEAD + qrKex->sizePqI, 10, "Alter one-time prekey", &shouldSucceed, change);

				// Check signed prekey
				if (xzdh_2_verifySignedPrekey_R(pkI, signedPrekey) != XZDH_SUCCESS) {
					problem = "Invalid signature on signed prekey";
					goto xzdhCleanup;
				}

				// Generate response
				switch (rand() % 3) {
				case 0:
					signMode = XZDH_HONEST;
					signSecret = skR;
					break;
				case 1:
					xzdh_forgeOffline(secretOneTimeState, offlineForgeKey);
					if (!checkCanary(offlineForgeKey, XZDH_SK_SIZE, false)) {
						problem = "Memory corruption from xzdh_forgeOffline";
						goto xzdhCleanup;
					}
					signMode = XZDH_OFFLINE_FORGE;
					signSecret = offlineForgeKey;
					break;
				case 2:
					signMode = XZDH_ONLINE_FORGE;
					signSecret = fskI;
					break;
				}
				scratchSize = XZDH_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (xzdh_2_exchange_R(signedPrekey, oneTimePrekey, qrKex, idLen, idI, idR, pkI, fpkI, pkR, signMode, signSecret, ad, adLen, response, keyR, scratch) != XZDH_SUCCESS) {
					problem = "Failed to generate response";
					goto xzdhCleanup;
				}
				if (   !checkCanary(response, XZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from xzdh_2_exchange_R";
					goto xzdhCleanup;
				}
				scratchSize = 0;

				cosmicRay(response, XZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen, 10, "Alter response", &shouldSucceed, change);

				// Receive response
				if (xzdh_3_receive_I(response, idLen, receivedIdR) != XZDH_SUCCESS) {
					problem = "Failed to extract ID from response";
					goto xzdhCleanup;
				}
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from response was incorrect";
					goto xzdhCleanup;
				}

				// Complete exchange
				scratchSize = XZDH_SCRATCH_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeKey + qrKex->sizeScratch;
				prepCanary(scratch, scratchSize);
				if (xzdh_3_exchange_I(response, qrKex, idLen, idI, pkI, fpkI, pkR, skI, ad, adLen, publicSignedState, publicOneTimeState, secretSignedState, secretOneTimeState, keyI, scratch) != XZDH_SUCCESS) {
					problem = "Failed to complete exchange";
					goto xzdhCleanup;
				}
				if (   !checkCanary(secretOneTimeState, XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI, true)
				    || !checkCanary(keyI, XZDH_KEY_SIZE, false)
				    || !checkCanary(scratch, scratchSize, true)) {
					problem = "Memory corruption from xzdh_3_exchange_I";
					goto xzdhCleanup;
				}
				scratchSize = 0;

				if (memcmp(keyI, keyR, XZDH_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto xzdhCleanup;
				}

				succeeded = true;
xzdhCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** XZDH test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("AD Len"); printf("%lu\n", adLen);
					printDesc("AD"); printHex(ad, adLen); printf("\n");
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID I"); printHex(idI, idLen); printf("\n");
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, XZDH_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, XZDH_PK_SIZE); printf("\n");
					printDesc("FPK I"); printHex(fpkI, XZDH_PK_SIZE); printf("\n");
					printDesc("FPK R"); printHex(fpkR, XZDH_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, XZDH_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, XZDH_SK_SIZE+4); printf("\n");
					printDesc("FSK I + canary"); printHex(fskI, XZDH_SK_SIZE+4); printf("\n");
					printDesc("FSK R + canary"); printHex(fskR, XZDH_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, XZDH_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, XZDH_KEY_SIZE+4); printf("\n");
					printDesc("Secret Signed State + canary"); printHex(secretSignedState, XZDH_SIGNED_SECRETSTATE_SIZE + 4); printf("\n");
					printDesc("Secret One-Time State + canary"); printHex(secretOneTimeState, XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI + 4); printf("\n");
					printDesc("Signing mode"); printf("0x%02X\n", signMode);
					if (scratchSize > 0) {
						printDesc("Scratch Space + canary"); printHex(scratch, scratchSize+4); printf("\n");
					}
					exit(1);
				}
			}

			// ECDH
			{
				bool shouldSucceed = true;

				unsigned char* secretState = securePage;
				unsigned char* keyI = secretState+4 + ECDH_SECRETSTATE_SIZE;
				unsigned char* keyR = keyI+4 + ECDH_KEY_SIZE;

				prepCanary(secretState, ECDH_SECRETSTATE_SIZE);
				prepCanary(keyI, ECDH_KEY_SIZE);
				prepCanary(keyR, ECDH_KEY_SIZE);

				unsigned char msg1[ECDH_MSG_SIZE+4];
				unsigned char msg2[ECDH_MSG_SIZE+4];
				prepCanary(msg1, ECDH_MSG_SIZE);
				prepCanary(msg2, ECDH_MSG_SIZE);

				bool succeeded = false;

				// Generate first flow
				ecdh_1_initiate_I(secretState, msg1);
				if (   !checkCanary(secretState, ECDH_SECRETSTATE_SIZE, false)
				    || !checkCanary(msg1, ECDH_MSG_SIZE, false)) {
					problem = "Memory corruption from ecdh_1_initiate_I";
					goto ecdhCleanup;
				}

				cosmicRay(msg1, ECDH_MSG_SIZE, 15, "Alter message 1", &shouldSucceed, change);

				// Generate second flow
				if (!ecdh_2_respond_R(msg1, msg2, keyR)) {
					problem = "Failed to generate response";
					goto ecdhCleanup;
				}
				if (   !checkCanary(msg2, ECDH_MSG_SIZE, false)
				    || !checkCanary(keyR, ECDH_KEY_SIZE, false)) {
					problem = "Memory corruption from ecdh_2_respond_R";
					goto ecdhCleanup;
				}

				cosmicRay(msg2, ECDH_MSG_SIZE, 15, "Alter message 2", &shouldSucceed, change);

				// Receive response
				if (!ecdh_3_finish_I(msg2, secretState, keyI)) {
					problem = "Failed to complete exchange";
					goto ecdhCleanup;
				}
				if (   !checkCanary(secretState, ECDH_SECRETSTATE_SIZE, true)
				    || !checkCanary(keyI, ECDH_KEY_SIZE, false)) {
					problem = "Memory corruption from ecdh_3_finish_I";
					goto ecdhCleanup;
				}

				if (memcmp(keyI, keyR, ECDH_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto ecdhCleanup;
				}

				succeeded = true;
ecdhCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** ECDH test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("Key I + canary"); printHex(keyI, ECDH_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, ECDH_KEY_SIZE+4); printf("\n");
					printDesc("Secret State + canary"); printHex(secretState, ECDH_SECRETSTATE_SIZE + 4); printf("\n");
					exit(1);
				}
			}

			// 3DH
			{
				bool shouldSucceed = true;

				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idR, (int)idLen);
				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + THREEDH_SK_SIZE;
				unsigned char* secretState = skR+4 + THREEDH_SK_SIZE;
				secretState += (16 - ((size_t)secretState % 16)) % 16;
				unsigned char* keyI = secretState+4 + THREEDH_SECRETSTATE_SIZE;
				unsigned char* keyR = keyI+4 + THREEDH_KEY_SIZE;
				unsigned char* scratch = keyR+4 + THREEDH_KEY_SIZE;
				scratch += (16 - ((size_t)scratch % 16)) % 16;

				prepCanary(skI, THREEDH_SK_SIZE);
				prepCanary(skR, THREEDH_SK_SIZE);
				prepCanary(secretState, THREEDH_SECRETSTATE_SIZE);
				prepCanary(keyI, THREEDH_KEY_SIZE);
				prepCanary(keyR, THREEDH_KEY_SIZE);

				unsigned char pkI[THREEDH_PK_SIZE];
				unsigned char pkR[THREEDH_PK_SIZE];

				unsigned char prekey[THREEDH_PREKEY_SIZE + 4];
				unsigned char response[THREEDH_MSG_OVERHEAD + idLen + 4];
				prepCanary(prekey, THREEDH_PREKEY_SIZE);
				prepCanary(response, THREEDH_MSG_OVERHEAD + idLen);

				unsigned char receivedIdR[idLen];

				bool succeeded = false;

				// Generate key pair for I
				prepCanary(scratch, THREEDH_SCRATCH_KEYGEN_SIZE);
				threedh_genLongTermKeys(pkI, skI, scratch);
				if (   !checkCanary(scratch, THREEDH_SCRATCH_KEYGEN_SIZE, true)
				    || !checkCanary(skI, THREEDH_SK_SIZE, false)) {
					problem = "Memory corruption from threedh_genLongTermKeys for I";
					goto threedhCleanup;
				}

				// Generate key pair for R
				prepCanary(scratch, THREEDH_SCRATCH_KEYGEN_SIZE);
				threedh_genLongTermKeys(pkR, skR, scratch);
				if (   !checkCanary(scratch, THREEDH_SCRATCH_KEYGEN_SIZE, true)
				    || !checkCanary(skR, THREEDH_SK_SIZE, false)) {
					problem = "Memory corruption from threedh_genLongTermKeys for R";
					goto threedhCleanup;
				}

				// Generate prekey
				threedh_1_makePreKey_I(secretState, prekey);
				if (   !checkCanary(secretState, THREEDH_SECRETSTATE_SIZE, false)
				    || !checkCanary(prekey, THREEDH_PREKEY_SIZE, false)) {
					problem = "Memory corruption from threedh_1_makePreKey_I";
					goto threedhCleanup;
				}

				cosmicRay(prekey, THREEDH_PREKEY_SIZE, 15, "Alter prekey", &shouldSucceed, change);

				// Generate response
				if (!threedh_2_exchange_R(prekey, idLen, idR, pkI, skR, response, keyR)) {
					problem = "Failed to generate response";
					goto threedhCleanup;
				}
				if (!checkCanary(response, THREEDH_MSG_OVERHEAD + idLen, false)) {
					problem = "Memory corruption from threedh_2_exchange_R";
					goto threedhCleanup;
				}

				cosmicRay(response, THREEDH_MSG_OVERHEAD + idLen, 15, "Alter response", &shouldSucceed, change);

				// Receive response
				threedh_3_receive_I(response, idLen, receivedIdR);
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from response was incorrect";
					goto threedhCleanup;
				}

				// Complete exchange
				if (!threedh_3_exchange_I(response, idLen, pkR, skI, secretState, keyI)) {
					problem = "Failed to complete exchange";
					goto threedhCleanup;
				}
				if (   !checkCanary(secretState, THREEDH_SECRETSTATE_SIZE, true)
				    || !checkCanary(keyI, THREEDH_KEY_SIZE, false)) {
					problem = "Memory corruption from threedh_3_exchange_I";
					goto threedhCleanup;
				}

				if (memcmp(keyI, keyR, THREEDH_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto threedhCleanup;
				}

				succeeded = true;
threedhCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** 3DH test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, THREEDH_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, THREEDH_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, THREEDH_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, THREEDH_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, THREEDH_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, THREEDH_KEY_SIZE+4); printf("\n");
					printDesc("Secret State + canary"); printHex(secretState, THREEDH_SECRETSTATE_SIZE + 4); printf("\n");
					printDesc("Scratch Space + canary"); printHex(scratch, THREEDH_SCRATCH_KEYGEN_SIZE+4); printf("\n");
					exit(1);
				}
			}

			// X3DH
			{
				bool shouldSucceed = true;

				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idR, (int)idLen);
				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + XTHREEDH_SK_SIZE;
				unsigned char* secretOneTimeState = skR+4 + XTHREEDH_SK_SIZE;
				secretOneTimeState += (16 - ((size_t)secretOneTimeState % 16)) % 16;
				unsigned char* secretSignedState = secretOneTimeState+4 + XTHREEDH_SECRETSTATE_SIZE;
				secretSignedState += (16 - ((size_t)secretSignedState % 16)) % 16;
				unsigned char* keyI = secretSignedState+4 + XTHREEDH_SECRETSTATE_SIZE;
				unsigned char* keyR = keyI+4 + XTHREEDH_KEY_SIZE;

				prepCanary(skI, XTHREEDH_SK_SIZE);
				prepCanary(skR, XTHREEDH_SK_SIZE);
				prepCanary(secretOneTimeState, XTHREEDH_SECRETSTATE_SIZE);
				prepCanary(secretSignedState, XTHREEDH_SECRETSTATE_SIZE);
				prepCanary(keyI, XTHREEDH_KEY_SIZE);
				prepCanary(keyR, XTHREEDH_KEY_SIZE);

				unsigned char pkI[XTHREEDH_PK_SIZE];
				unsigned char pkR[XTHREEDH_PK_SIZE];

				unsigned char signedPrekey[XTHREEDH_SIGNED_PREKEY_SIZE + 4];
				unsigned char oneTimePrekey[XTHREEDH_ONETIME_PREKEY_SIZE + 4];
				unsigned char response[XTHREEDH_MSG_OVERHEAD + idLen + 4];
				prepCanary(signedPrekey, XTHREEDH_SIGNED_PREKEY_SIZE);
				prepCanary(oneTimePrekey, XTHREEDH_ONETIME_PREKEY_SIZE);
				prepCanary(response, XTHREEDH_MSG_OVERHEAD + idLen);

				unsigned char receivedIdR[idLen];

				bool succeeded = false;

				// Generate key pair for I
				xthreedh_genLongTermKeys(pkI, skI);
				if (!checkCanary(skI, XTHREEDH_SK_SIZE, false)) {
					problem = "Memory corruption from xthreedh_genLongTermKeys for I";
					goto xthreedhCleanup;
				}

				// Generate key pair for R
				xthreedh_genLongTermKeys(pkR, skR);
				if (!checkCanary(skR, XTHREEDH_SK_SIZE, false)) {
					problem = "Memory corruption from xthreedh_genLongTermKeys for R";
					goto xthreedhCleanup;
				}

				// Generate signed prekey
				xthreedh_genSignedPrekey(pkI, skI, secretSignedState, signedPrekey);
				if (   !checkCanary(secretSignedState, XTHREEDH_SECRETSTATE_SIZE, false)
				    || !checkCanary(signedPrekey, XTHREEDH_SIGNED_PREKEY_SIZE, false)) {
					problem = "Memory corruption from xthreedh_genSignedPrekey";
					goto xthreedhCleanup;
				}

				cosmicRay(signedPrekey, XTHREEDH_SIGNED_PREKEY_SIZE, 10, "Alter signed prekey", &shouldSucceed, change);

				// Generate prekey
				xthreedh_1_makeOneTimePreKey_I(secretOneTimeState, oneTimePrekey);
				if (   !checkCanary(secretOneTimeState, XTHREEDH_SECRETSTATE_SIZE, false)
				    || !checkCanary(oneTimePrekey, XTHREEDH_ONETIME_PREKEY_SIZE, false)) {
					problem = "Memory corruption from xthreedh_1_makeOneTimePreKey_I";
					goto xthreedhCleanup;
				}

				cosmicRay(oneTimePrekey, XTHREEDH_ONETIME_PREKEY_SIZE, 10, "Alter one-time prekey", &shouldSucceed, change);

				// Check signed prekey
				if (!xthreedh_2_verifySignedPrekey_R(pkI, signedPrekey)) {
					problem = "Invalid signature on signed prekey";
					goto xthreedhCleanup;
				}

				// Generate response
				if (!xthreedh_2_exchange_R(signedPrekey, oneTimePrekey, idLen, idR, pkI, skR, response, keyR)) {
					problem = "Failed to generate response";
					goto xthreedhCleanup;
				}
				if (!checkCanary(response, XTHREEDH_MSG_OVERHEAD + idLen, false)) {
					problem = "Memory corruption from xthreedh_2_exchange_R";
					goto xthreedhCleanup;
				}

				cosmicRay(response, XTHREEDH_MSG_OVERHEAD + idLen, 10, "Alter response", &shouldSucceed, change);

				// Receive response
				xthreedh_3_receive_I(response, idLen, receivedIdR);
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from response was incorrect";
					goto xthreedhCleanup;
				}

				// Complete exchange
				if (!xthreedh_3_exchange_I(response, idLen, pkR, skI, secretSignedState, secretOneTimeState, keyI)) {
					problem = "Failed to complete exchange";
					goto xthreedhCleanup;
				}
				if (   !checkCanary(secretOneTimeState, XTHREEDH_SECRETSTATE_SIZE, true)
				    || !checkCanary(keyI, XTHREEDH_KEY_SIZE, false)) {
					problem = "Memory corruption from xthreedh_3_exchange_I";
					goto xthreedhCleanup;
				}

				// Erase signed prekey
				xthreedh_eraseSignedPrekey(secretSignedState);
				if (!checkCanary(secretSignedState, XTHREEDH_SECRETSTATE_SIZE, true)) {
					problem = "Failed to erase signed prekey secret";
					goto xthreedhCleanup;
				}

				if (memcmp(keyI, keyR, XTHREEDH_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto xthreedhCleanup;
				}

				succeeded = true;
xthreedhCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** X3DH test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, XTHREEDH_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, XTHREEDH_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, XTHREEDH_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, XTHREEDH_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, XTHREEDH_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, XTHREEDH_KEY_SIZE+4); printf("\n");
					printDesc("Secret Signed State + canary"); printHex(secretSignedState, XTHREEDH_SECRETSTATE_SIZE + 4); printf("\n");
					printDesc("Secret One-Time State + canary"); printHex(secretOneTimeState, XTHREEDH_SECRETSTATE_SIZE + 4); printf("\n");
					exit(1);
				}
			}

			// SIGMA-R
			{
				bool shouldSucceed = true;

				unsigned char idI[1024];
				unsigned char idR[1024];
				size_t idLen = (rand() % 1023) + 1;
				RAND_bytes(idI, (int)idLen);
				RAND_bytes(idR, (int)idLen);

				unsigned char* skI = securePage;
				unsigned char* skR = skI+4 + SIGMAR_SK_SIZE;
				unsigned char* secretStateI = skR+4 + SIGMAR_SK_SIZE;
				secretStateI += (16 - ((size_t)secretStateI % 16)) % 16;
				unsigned char* secretStateR = secretStateI+4 + SIGMAR_SECRETSTATE_I_SIZE;
				secretStateR += (16 - ((size_t)secretStateR % 16)) % 16;
				unsigned char* keyI = secretStateR+4 + SIGMAR_SECRETSTATE_R_SIZE;
				unsigned char* keyR = keyI+4 + SIGMAR_KEY_SIZE;

				prepCanary(skI, SIGMAR_SK_SIZE);
				prepCanary(skR, SIGMAR_SK_SIZE);
				prepCanary(secretStateI, SIGMAR_SECRETSTATE_I_SIZE);
				prepCanary(secretStateR, SIGMAR_SECRETSTATE_R_SIZE);
				prepCanary(keyI, SIGMAR_KEY_SIZE);
				prepCanary(keyR, SIGMAR_KEY_SIZE);

				unsigned char pkI[SIGMAR_PK_SIZE];
				unsigned char pkR[SIGMAR_PK_SIZE];

				unsigned char message1[SIGMAR_EX1_MSG_SIZE + 4];
				unsigned char message2[SIGMAR_EX1_MSG_SIZE + 4];
				unsigned char message3[SIGMAR_EX2_MSG_OVERHEAD + idLen + 4];
				unsigned char message4[SIGMAR_EX2_MSG_OVERHEAD + idLen + 4];

				prepCanary(message1, SIGMAR_EX1_MSG_SIZE);
				prepCanary(message2, SIGMAR_EX1_MSG_SIZE);
				prepCanary(message3, SIGMAR_EX2_MSG_OVERHEAD + idLen);
				prepCanary(message4, SIGMAR_EX2_MSG_OVERHEAD + idLen);

				unsigned char receivedIdI[idLen];
				unsigned char receivedIdR[idLen];

				bool succeeded = false;

				// Generate key pair for I
				sigmar_genLongTermKeys(pkI, skI);
				if (!checkCanary(skI, SIGMAR_SK_SIZE, false)) {
					problem = "Memory corruption from sigmar_genLongTermKeys for I";
					goto sigmarCleanup;
				}

				// Generate key pair for R
				sigmar_genLongTermKeys(pkR, skR);
				if (!checkCanary(skR, SIGMAR_SK_SIZE, false)) {
					problem = "Memory corruption from sigmar_genLongTermKeys for R";
					goto sigmarCleanup;
				}

				// Generate first message
				sigmar_1_initiateI(secretStateI, message1);
				if (   !checkCanary(secretStateI, SIGMAR_SECRETSTATE_I_SIZE, false)
				    || !checkCanary(message1, SIGMAR_EX1_MSG_SIZE, false)) {
					problem = "Memory corruption from sigmar_1_initiateI";
					goto sigmarCleanup;
				}

				cosmicRay(message1, SIGMAR_EX1_MSG_SIZE, 5, "Alter message 1", &shouldSucceed, change);

				// Generate second message
				sigmar_2_respond_R(message1, secretStateR, message2);
				if (   !checkCanary(secretStateR, SIGMAR_SECRETSTATE_R_SIZE, false)
				    || !checkCanary(message2, SIGMAR_EX1_MSG_SIZE, false)) {
					problem = "Memory corruption from sigmar_2_respond_R";
					goto sigmarCleanup;
				}

				cosmicRay(message2, SIGMAR_EX1_MSG_SIZE, 5, "Alter message 2", &shouldSucceed, change);

				// Generate third message
				if (!sigmar_3_respond_I(message2, idLen, idI, pkI, skI, secretStateI, message3)) {
					problem = "Failed to generate message 3";
					goto sigmarCleanup;
				}
				if (   !checkCanary(secretStateI, SIGMAR_SECRETSTATE_I_SIZE, false)
				    || !checkCanary(message3, SIGMAR_EX2_MSG_OVERHEAD + idLen, false)) {
					problem = "Memory corruption from sigmar_3_respond_I";
					goto sigmarCleanup;
				}

				cosmicRay(message3, SIGMAR_EX2_MSG_OVERHEAD + idLen, 10, "Alter message 3", &shouldSucceed, change);

				// Receive third message
				sigmar_4_receive_R(message3, idLen, receivedIdI);
				if (memcmp(idI, receivedIdI, idLen) != 0) {
					problem = "Extracted ID from message 3 was incorrect";
					goto sigmarCleanup;
				}

				// Generate forth message
				if (!sigmar_4_complete_R(message3, idLen, idI, idR, pkI, pkR, skR, secretStateR, message4, keyR)) {
					problem = "Failed to generate message 4";
					goto sigmarCleanup;
				}
				if (   !checkCanary(secretStateR, SIGMAR_SECRETSTATE_R_SIZE, true)
				    || !checkCanary(message4, SIGMAR_EX2_MSG_OVERHEAD + idLen, false)
				    || !checkCanary(keyR, SIGMAR_KEY_SIZE, false)) {
					problem = "Memory corruption from sigmar_4_complete_R";
					goto sigmarCleanup;
				}

				cosmicRay(message4, SIGMAR_EX2_MSG_OVERHEAD + idLen, 10, "Alter message 4", &shouldSucceed, change);

				// Receive forth message
				sigmar_5_receive_I(message4, idLen, receivedIdR);
				if (memcmp(idR, receivedIdR, idLen) != 0) {
					problem = "Extracted ID from message 4 was incorrect";
					goto sigmarCleanup;
				}

				// Complete the exchange
				if (!sigmar_5_complete_I(message4, idLen, idR, pkR, secretStateI, keyI)) {
					problem = "Failed to process message 4";
					goto sigmarCleanup;
				}
				if (   !checkCanary(secretStateI, SIGMAR_SECRETSTATE_I_SIZE, true)
				    || !checkCanary(keyI, SIGMAR_KEY_SIZE, false)) {
					problem = "Memory corruption from sigmar_5_complete_I";
					goto sigmarCleanup;
				}

				if (memcmp(keyI, keyR, SIGMAR_KEY_SIZE) != 0) {
					problem = "Derived keys were not the same";
					goto sigmarCleanup;
				}

				succeeded = true;
sigmarCleanup:
				if (succeeded != shouldSucceed) {
					printf("\n*** SIGMA-R test failure ***\n");
					if (shouldSucceed) {
						printf("Should have succeeded, but: %s\n", problem);
					} else {
						printf("Should have failed: %s\n", change);
					}
					printDesc("ID Len"); printf("%lu\n", idLen);
					printDesc("ID I"); printHex(idI, idLen); printf("\n");
					printDesc("ID R"); printHex(idR, idLen); printf("\n");
					printDesc("PK I"); printHex(pkI, SIGMAR_PK_SIZE); printf("\n");
					printDesc("PK R"); printHex(pkR, SIGMAR_PK_SIZE); printf("\n");
					printDesc("SK I + canary"); printHex(skI, SIGMAR_SK_SIZE+4); printf("\n");
					printDesc("SK R + canary"); printHex(skR, SIGMAR_SK_SIZE+4); printf("\n");
					printDesc("Key I + canary"); printHex(keyI, SIGMAR_KEY_SIZE+4); printf("\n");
					printDesc("Key R + canary"); printHex(keyR, SIGMAR_KEY_SIZE+4); printf("\n");
					printDesc("Secret State I + canary"); printHex(secretStateI, SIGMAR_SECRETSTATE_I_SIZE + 4); printf("\n");
					printDesc("Secret State R + canary"); printHex(secretStateR, SIGMAR_SECRETSTATE_R_SIZE + 4); printf("\n");
					exit(1);
				}
			}
		}
	}
}

int main() {
	unsigned int seed;
	RAND_bytes((unsigned char*)&seed, sizeof(seed));
	srand(seed);

	ERR_load_crypto_strings();
	OpenSSL_add_all_algorithms();

	bignum256modm ALIGN(16) one = {1,0,0,0,0};
	bignum256modm ALIGN(16) zero = {0,0,0,0,0};
	bignum256modm ALIGN(16) s1, s2;
	modm_rand(s1);
	modm_rand(s2);

	printPoint(&ge25519_basepoint, "B from Donna");

	ge25519 ALIGN(16) b;
	ge25519_scalarmult_base_niels(&b, ge25519_niels_base_multiples, one);
	printPoint(&b, "B from scale Donna");

	ge25519 ALIGN(16) r;
	ge25519_double_scalarmult_vartime(&r, &b, one, zero);
	printPoint(&r, "1*B from Donna");

	ge25519_double_scalarmult_vartime(&r, &b, zero, zero);
	printPoint(&r, "1 from Donna");

	ge25519_double_scalarmult_vartime(&r, &b, one, one);
	printPoint(&r, "1*B + 1*B from Donna");

	ge25519_double_scalarmult_vartime(&r, &b, s1, zero);
	printPoint(&r, "s1*B + 0*B from Donna");

	ge25519_scalarmult_vartime(&r, &b, one);
	printPoint(&r, "Explicit 1*B");

	ge25519_scalarmult_vartime(&r, &b, s1);
	printPoint(&r, "Explicit s1*B");

	printPoint(&ge25519_basepoint1, "g1");

	ge25519_scalarmult_base1(&r, one);
	printPoint(&r, "1*g1");

	ge25519_scalarmult_base1(&r, s1);
	printPoint(&r, "s1*g1");

	ge25519_double_scalarmult_vartime(&r, &ge25519_basepoint1, s1, zero);
	printPoint(&r, "s1*g1 from Donna");

	printPoint(&ge25519_basepoint2, "g2");

	ge25519_scalarmult_base2(&r, one);
	printPoint(&r, "1*g2");

	ge25519_scalarmult_base2(&r, s2);
	printPoint(&r, "s2*g2");

	ge25519_double_scalarmult_vartime(&r, &ge25519_basepoint2, s2, zero);
	printPoint(&r, "s2*g2 from Donna");

	ge25519 ALIGN(16) t1, t2;

	ge25519_double_scalarmult_bases(&r, s1, s2);
	printPoint(&r, "Constant s1*g1+s2*g2");

	ge25519_double_scalarmult_bases_vartime(&r, s1, s2);
	printPoint(&r, "s1*g1+s2*g2");

	ge25519_neg(&t1, &r);
	printPoint(&t1, "-(s1*g1+s2*g2)");

	ge25519_add(&r, &r, &t1);
	printPoint(&r, "(s1*g1+s2*g2)-(s1*g1+s2*g2)");

	unsigned char buffer[32];
	ge25519_pack(buffer, &ge25519_basepoint);
	ge25519_unpack_negative_vartime(&r, buffer);
	printPoint(&r, "Unpacked B");

	ge25519_neg(&t1, &r);
	printPoint(&t1, "-(Unpacked B)");

	ge25519_add(&r, &r, &t1);
	printPoint(&r, "(Unpacked B)-(Unpacked B)");

	ge25519_scalarmult_base1(&t1, s1);
	ge25519_neg(&t2, &t1);
	ge25519_add(&r, &t1, &t2);
	printPoint(&r, "s1*g1 - s1*g1");

	bignum256modm ALIGN(16) st;
	printDesc("s1"); printBigNumModM(s1); printf("\n");
	printDesc("s2"); printBigNumModM(s2); printf("\n");
	modm_neg(st, s1);
	printDesc("-s1"); printBigNumModM(st); printf("\n");
	add256_modm(st, s1, st);
	printDesc("s1+(-s1)"); printBigNumModM(st); printf("\n");
	modm_sub(st, s2, s2);
	printDesc("s2-s2"); printBigNumModM(st); printf("\n");
	add256_modm(st, s1, s2);
	modm_sub(st, st, s2);
	printDesc("(s1+s2)-s2"); printBigNumModM(st); printf("\n");

	fuzz();

	EVP_cleanup();
	CRYPTO_cleanup_all_ex_data();
	ERR_free_strings();
	return 0;
}
