#include <stdlib.h>
#include <stdbool.h>

#include "crypto.h"
#include "ecc.h"
#include "kmac.h"
#include "secmem.h"
#include "other-ake.h"

static unsigned char threedhKdfTag[] = "3DH KDF";
static unsigned char xthreedhKdfTag[] = "X3DH KDF";
static unsigned char sigmarKdfTag[] = "SIGMA-R KDF";
static unsigned char sigmarIMacTag[] = "SIGMA-R KMAC:I";
static unsigned char sigmarRMacTag[] = "SIGMA-R KMAC:R";
static size_t threedhKdfTagLen = 8;
static size_t xthreedhKdfTagLen = 9;
static size_t sigmarKdfTagLen = 12;
static size_t sigmarMacTagLen = 15;

void ecdh_1_initiate_I(unsigned char* secretState, unsigned char* send) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base1(&dhPublic, dhSecret);
	ge25519_pack(send, &dhPublic);
}

bool ecdh_2_respond_R(const unsigned char* received, unsigned char* send, unsigned char* key) {
	ge25519 ALIGN(16) theirDH;
	if (!ge25519_unpack_vartime(&theirDH, received)) return false;

	bignum256modm ALIGN(16) dhSecret;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base1(&dhPublic, dhSecret);
	ge25519_pack(send, &dhPublic);

	ge25519 ALIGN(16) sharedSecret;
	ge25519_scalarmult(&sharedSecret, &theirDH, dhSecret);
	ge25519_pack(key, &sharedSecret);

	return true;
}

bool ecdh_3_finish_I(const unsigned char* received, unsigned char* secretState, unsigned char* key) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;

	ge25519 ALIGN(16) theirDH;
	if (!ge25519_unpack_vartime(&theirDH, received)) return false;

	ge25519 ALIGN(16) sharedSecret;
	ge25519_scalarmult(&sharedSecret, &theirDH, dhSecret);
	ge25519_pack(key, &sharedSecret);

	secureErase(secretState, sizeof(bignum256modm));

	return true;
}

void threedh_genLongTermKeys(unsigned char* pk, unsigned char* sk, unsigned char* scratch) {
	bignum256modm_element_t* mSk = (bignum256modm_element_t*)scratch;

	ge25519 ALIGN(16) gePk;

	RAND_bytes(sk, 32);
	curved25519_make_secret(mSk, sk);
	ge25519_scalarmult_base1(&gePk, mSk);
	ge25519_pack(pk, &gePk);

	secureErase(scratch, sizeof(bignum256modm));
}

void threedh_1_makePreKey_I(unsigned char* secretState, unsigned char* send) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base1(&dhPublic, dhSecret);
	ge25519_pack(send, &dhPublic);
}

static bool threedh_dhIntoKDF(ge25519* pub, bignum256modm secret, Keccak_HashInstance* hasher) {
	unsigned char kdfBuffer[32];

	ge25519 ALIGN(16) shared;
	ge25519_scalarmult(&shared, pub, secret);
	ge25519_pack(kdfBuffer, &shared);

	bool success = (Keccak_HashUpdate(hasher, kdfBuffer, sizeof(kdfBuffer)*8) == SUCCESS);
	secureErase(kdfBuffer, 32);
	return success;
}

static bool threedh_key(const unsigned char* theirEncodedDH, bignum256modm dhSecret,
	const unsigned char* theirPK, const unsigned char* mySK,
	bool meFirst, unsigned char* key) {

	ge25519 ALIGN(16) theirDH;
	if (!ge25519_unpack_vartime(&theirDH, theirEncodedDH)) return false;

	ge25519 ALIGN(16) theirLT;
	if (!ge25519_unpack_vartime(&theirLT, theirPK)) return false;

	bignum256modm ALIGN(16) ltSecret;
	expand256_modm(ltSecret, mySK, 32);

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, threedhKdfTag, threedhKdfTagLen*8) == SUCCESS);
	hashSuccess = hashSuccess && threedh_dhIntoKDF(&theirDH, dhSecret, &hasher);
	hashSuccess = hashSuccess && threedh_dhIntoKDF(meFirst ? &theirDH : &theirLT, meFirst ? ltSecret : dhSecret, &hasher);
	hashSuccess = hashSuccess && threedh_dhIntoKDF(meFirst ? &theirLT : &theirDH, meFirst ? dhSecret : ltSecret, &hasher);
	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, key) == SUCCESS);

	return hashSuccess;
}

bool threedh_2_exchange_R(const unsigned char* received,
	size_t idLen, const unsigned char* idR,
	const unsigned char* pkI,
	const unsigned char* skR,
	unsigned char* send, unsigned char* key) {

	bignum256modm ALIGN(16) dhSecret;
	modm_rand(dhSecret);

	memcpy(send, idR, idLen);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base1(&dhPublic, dhSecret);
	ge25519_pack(send+idLen, &dhPublic);

	return threedh_key(received, dhSecret, pkI, skR, false, key);
}

void threedh_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
}

bool threedh_3_exchange_I(const unsigned char* received, size_t idLen,
	const unsigned char* pkR,
	const unsigned char* skI,
	unsigned char* secretState,
	unsigned char* key) {

	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;

	bool success = threedh_key(received + idLen, dhSecret, pkR, skI, true, key);
	secureErase(secretState, sizeof(bignum256modm));
	return success;
}

void xthreedh_genLongTermKeys(unsigned char* pk, unsigned char* sk) {
	RAND_bytes(sk, 32);
	ed25519_publickey(sk, pk);
}

void xthreedh_genSignedPrekey(const unsigned char* pk, const unsigned char* sk, unsigned char* secretSignedState, unsigned char* send) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretSignedState;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base_niels(&dhPublic, ge25519_niels_base_multiples, dhSecret);
	ge25519_pack(send, &dhPublic);

	ed25519_sign(send, 32, sk, pk, send+32);
}

void xthreedh_eraseSignedPrekey(unsigned char* secretSignedState) {
	secureErase(secretSignedState, sizeof(bignum256modm));
}

void xthreedh_1_makeOneTimePreKey_I(unsigned char* secretOneTimeState, unsigned char* send) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretOneTimeState;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base_niels(&dhPublic, ge25519_niels_base_multiples, dhSecret);
	ge25519_pack(send, &dhPublic);
}

static bool xthreedh_dhIntoKDF(ge25519* pub, bignum256modm secret, Keccak_HashInstance* hasher) {
	unsigned char kdfBuffer[32];

	ge25519 ALIGN(16) shared;
	ge25519_scalarmult(&shared, pub, secret);
	ge25519_pack(kdfBuffer, &shared);

	bool success = (Keccak_HashUpdate(hasher, kdfBuffer, sizeof(kdfBuffer)*8) == SUCCESS);
	secureErase(kdfBuffer, 32);
	return success;
}

static bool xthreedh_key(
	const unsigned char* theirEncodedEphemeralKey, bignum256modm myEphemeralSecret,
	const unsigned char* theirEncodedLinkKey, bignum256modm myLinkSecret,
	const unsigned char* theirPK, const unsigned char* mySK,
	bool meFirst, unsigned char* key) {

	ge25519 ALIGN(16) theirEK;
	if (!ge25519_unpack_vartime(&theirEK, theirEncodedEphemeralKey)) return false;

	ge25519 ALIGN(16) theirLK;
	if (!ge25519_unpack_vartime(&theirLK, theirEncodedLinkKey)) return false;

	ge25519 ALIGN(16) theirLTPubKey;
	if (!ge25519_unpack_vartime(&theirLTPubKey, theirPK)) return false;

	bignum256modm ALIGN(16) ltSecret;
	ed25519_secret_key ed25519sec;
	ed25519_secretkey(ed25519sec, mySK);
	expand256_modm(ltSecret, ed25519sec, 32);

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, xthreedhKdfTag, xthreedhKdfTagLen*8) == SUCCESS);

	// DH(I.LK, R.LK) aka DH(I.SPK, R.EK)
	hashSuccess = hashSuccess && xthreedh_dhIntoKDF(&theirLK, myLinkSecret, &hasher);

	// DH(I.PK, R.LK) aka DH(I.PK, R.EK)
	hashSuccess = hashSuccess && xthreedh_dhIntoKDF(meFirst ? &theirLK : &theirLTPubKey, meFirst ? ltSecret : myLinkSecret, &hasher);

	// DH(R.PK, I.LK) aka DH(R.PK, I.SPK)
	hashSuccess = hashSuccess && xthreedh_dhIntoKDF(meFirst ? &theirLTPubKey : &theirLK, meFirst ? myLinkSecret : ltSecret, &hasher);

	// DH(I.EK, R.LK) aka DH(I.OPK, R.EK)
	hashSuccess = hashSuccess && xthreedh_dhIntoKDF(meFirst ? &theirLK : &theirEK, meFirst ? myEphemeralSecret : myLinkSecret, &hasher);

	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, key) == SUCCESS);
	return hashSuccess;
}

bool xthreedh_2_verifySignedPrekey_R(const unsigned char* pkI, const unsigned char* signedPrekey) {
	return ed25519_sign_open(signedPrekey, 32, pkI, signedPrekey+32) == 0;
}

bool xthreedh_2_exchange_R(
	const unsigned char* signedPrekey, const unsigned char* oneTimePrekey,
	size_t idLen, const unsigned char* idR,
	const unsigned char* pkI,
	const unsigned char* skR,
	unsigned char* send, unsigned char* key) {

	bignum256modm ALIGN(16) dhSecret;
	modm_rand(dhSecret);

	memcpy(send, idR, idLen);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base_niels(&dhPublic, ge25519_niels_base_multiples, dhSecret);
	ge25519_pack(send+idLen, &dhPublic);

	return xthreedh_key(oneTimePrekey, dhSecret, signedPrekey, dhSecret, pkI, skR, false, key);
}

void xthreedh_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
}

bool xthreedh_3_exchange_I(const unsigned char* received, size_t idLen,
	const unsigned char* pkR,
	const unsigned char* skI,
	const unsigned char* secretSignedState,
	unsigned char* secretOneTimeState,
	unsigned char* key) {

	bignum256modm_element_t* oneTimeSecret = (bignum256modm_element_t*)secretOneTimeState;
	bignum256modm_element_t* signedSecret = (bignum256modm_element_t*)secretSignedState;

	bool success = xthreedh_key(received+idLen, oneTimeSecret, received+idLen, signedSecret, pkR, skI, true, key);
	secureErase(secretOneTimeState, sizeof(bignum256modm));
	return success;
}

void sigmar_genLongTermKeys(unsigned char* pk, unsigned char* sk) {
	RAND_bytes(sk, 32);
	ed25519_publickey(sk, pk);
}

// Initiator secret state:
//   [ I ephemeral secret ] [ packed R ephemeral public ] [ packed I ephemeral public ] [ mac key ] [ session key ]
// Responder secret state:
//   [ R ephemeral secret ] [ packed I ephemeral public ] [ packed R ephemeral public ]

void sigmar_1_initiateI(unsigned char* secretState, unsigned char* send) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base1(&dhPublic, dhSecret);
	ge25519_pack(send, &dhPublic);
	memcpy(secretState + sizeof(bignum256modm) + 32, send, 32);
}

void sigmar_2_respond_R(const unsigned char* received, unsigned char* secretState, unsigned char* send) {
	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;
	modm_rand(dhSecret);

	memcpy(secretState + sizeof(bignum256modm), received, 32);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base1(&dhPublic, dhSecret);
	ge25519_pack(send, &dhPublic);
	memcpy(secretState + sizeof(bignum256modm) + 32, send, 32);
}

static bool sigmar_sharedKeys(ge25519* theirDH, bignum256modm dhSecret, unsigned char* key, unsigned char* macKey) {
	ge25519 ALIGN(16) sharedSecret;
	ge25519_scalarmult(&sharedSecret, theirDH, dhSecret);
	ge25519_pack(key, &sharedSecret);

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, sigmarKdfTag, sigmarKdfTagLen*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, key, 32*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, macKey) == SUCCESS);
	return hashSuccess;
}

static bool sigmar_mac(const unsigned char* macKey, bool fromInitiator, const unsigned char* msg, size_t msgLen, unsigned char* mac) {
	return !!kmac256(mac, 32, macKey, 32, msg, msgLen, fromInitiator ? sigmarIMacTag : sigmarRMacTag, sigmarMacTagLen);
}

bool sigmar_3_respond_I(const unsigned char* received,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI,
	const unsigned char* skI,
	unsigned char* secretState,
	unsigned char* send) {

	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;

	ge25519 ALIGN(16) theirDH;
	if (!ge25519_unpack_vartime(&theirDH, received)) return false;

	unsigned char* macKey = secretState + sizeof(bignum256modm) + 64;
	unsigned char* key = macKey + 32;
	if (!sigmar_sharedKeys(&theirDH, dhSecret, key, macKey)) return false;

	unsigned char* signMsg = secretState + sizeof(bignum256modm);
	memcpy(signMsg, received, 32);

	memcpy(send, idI, idLen);
	ed25519_sign(signMsg, 64, skI, pkI, send+idLen);
	sigmar_mac(macKey, true, idI, idLen, send+idLen+64);

	return true;
}

void sigmar_4_receive_R(const unsigned char* received, size_t idLen, unsigned char* idI) {
	memcpy(idI, received, idLen);
}

bool sigmar_4_complete_R(const unsigned char* received,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* pkR,
	const unsigned char* skR,
	unsigned char* secretState,
	unsigned char* send, unsigned char* key) {

	bool success = false;

	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretState;

	const unsigned char* theirEncodedDH = secretState + sizeof(bignum256modm);
	const unsigned char* myEncodedDH = theirEncodedDH + 32;

	const unsigned char* mySignMsg = theirEncodedDH;
	unsigned char theirSignMsg[64];
	memcpy(theirSignMsg, myEncodedDH, 32);
	memcpy(theirSignMsg + 32, theirEncodedDH, 32);
	if (ed25519_sign_open(theirSignMsg, 64, pkI, received+idLen) != 0) goto cleanup;

	ge25519 ALIGN(16) theirDH;
	if (!ge25519_unpack_vartime(&theirDH, theirEncodedDH)) goto cleanup;

	unsigned char macKey[32];
	if (!sigmar_sharedKeys(&theirDH, dhSecret, key, macKey)) goto cleanup;

	unsigned char expectedMac[32];
	sigmar_mac(macKey, true, idI, idLen, expectedMac);
	if (!ed25519_verify(expectedMac, received+idLen+64, 32)) goto cleanup;

	memcpy(send, idR, idLen);
	ed25519_sign(mySignMsg, 64, skR, pkR, send+idLen);
	sigmar_mac(macKey, false, idR, idLen, send+idLen+64);

	success = true;
cleanup:
	secureErase(secretState, sizeof(bignum256modm) + 64);
	return success;
}

void sigmar_5_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
}

bool sigmar_5_complete_I(const unsigned char* received,
	size_t idLen, const unsigned char* idR,
	const unsigned char* pkR,
	unsigned char* secretState,
	unsigned char* key) {

	bool success = false;

	const unsigned char* theirEncodedDH = secretState + sizeof(bignum256modm);
	const unsigned char* myEncodedDH = theirEncodedDH + 32;

	unsigned char theirSignMsg[64];
	memcpy(theirSignMsg, myEncodedDH, 32);
	memcpy(theirSignMsg + 32, theirEncodedDH, 32);
	if (ed25519_sign_open(theirSignMsg, 64, pkR, received+idLen) != 0) goto cleanup;

	unsigned char* macKey = secretState + sizeof(bignum256modm) + 64;
	unsigned char expectedMac[32];
	sigmar_mac(macKey, false, idR, idLen, expectedMac);
	if (!ed25519_verify(expectedMac, received+idLen+64, 32)) goto cleanup;

	memcpy(key, macKey + 32, 32);

	success = true;
cleanup:
	secureErase(secretState, sizeof(bignum256modm) + 128);
	return success;
}
