#include <stdlib.h>
#include <stdbool.h>

#include "qr.h"
#include "crypto.h"

static bool qrNull_Init(void) { return true; }

static void qrNull_Cleanup(void) {}

static bool qrNull_GenI(unsigned char* pqI, unsigned char* sqI, unsigned char* scratch) {
	(void)pqI; (void)sqI; (void)scratch;
	return true;
}

static bool qrNull_GenR(const unsigned char* pqI, unsigned char* qR, unsigned char* key, unsigned char* scratch) {
	(void)pqI; (void)qR; (void)key; (void)scratch;
	return true;
}

static bool qrNull_KeyI(const unsigned char* sqI, const unsigned char* qR, unsigned char* key, unsigned char* scratch) {
	(void)sqI; (void)qR; (void)key; (void)scratch;
	return true;
}

static qrKex qrsKexNone = {
	.sizePqI = 0,
	.sizeSqI = 0,
	.sizeQR = 0,
	.sizeKey = 0,
	.sizeScratch = 0,
	.sizeAlign8 = 0,
	.sizeAlign16 = 0,
	.sizeAlign24 = 0,

	.qrInit = qrNull_Init,
	.qrGenI = qrNull_GenI,
	.qrGenR = qrNull_GenR,
	.qrKeyI = qrNull_KeyI,
	.qrCleanup = qrNull_Cleanup,
};
qrKex* qrKexNone = &qrsKexNone;

#ifdef QRKEX_USE_SIDH

static qrKex qrsKexSIDH;

static PCurveIsogenyStruct sidhData;

static CRYPTO_STATUS qrSIDH_rand(unsigned int nbytes, unsigned char* random_array) {
	unsigned int written = 0;
	while (written < nbytes) {
		unsigned int toWrite = nbytes - written;
		if (toWrite > INT_MAX) toWrite = INT_MAX;
		if (RAND_bytes(random_array, (int)toWrite) != 1) return CRYPTO_ERROR;
		written += toWrite;
	}
	return CRYPTO_SUCCESS;
}

static bool qrSIDH_Init(void) {
	sidhData = SIDH_curve_allocate(&CurveIsogeny_SIDHp751);
	if (sidhData == NULL) return false;
	if (SIDH_curve_initialize(sidhData, qrSIDH_rand, &CurveIsogeny_SIDHp751) != CRYPTO_SUCCESS) return false;
	return true;
}

static void qrSIDH_Cleanup(void) {
	SIDH_curve_free(sidhData);
	sidhData = NULL;
}

static bool qrSIDH_GenI(unsigned char* pqI, unsigned char* sqI, unsigned char* scratch) {
	(void)scratch;
	return (KeyGeneration_A(sqI, pqI, sidhData) == CRYPTO_SUCCESS);
}

static bool qrSIDH_GenR(const unsigned char* pqI, unsigned char* qR, unsigned char* key, unsigned char* scratch) {
	bool valid = false;
	unsigned char pqICopy[qrsKexSIDH.sizePqI];
	memcpy(pqICopy, pqI, qrsKexSIDH.sizePqI);
	if (Validate_PKA(pqICopy, &valid, sidhData) != CRYPTO_SUCCESS || !valid) return false;
	if (KeyGeneration_B(scratch, qR, sidhData) != CRYPTO_SUCCESS) return false;

	if (SecretAgreement_B(scratch, pqICopy, key, sidhData) != CRYPTO_SUCCESS) valid = false;
	return valid;
}

static bool qrSIDH_KeyI(const unsigned char* sqI, const unsigned char* qR, unsigned char* key, unsigned char* scratch) {
	bool valid = false;
	unsigned char qRCopy[qrsKexSIDH.sizeQR];
	memcpy(qRCopy, qR, qrsKexSIDH.sizeQR);
	if (Validate_PKB(qRCopy, &valid, sidhData) != CRYPTO_SUCCESS || !valid) return false;
	memcpy(scratch, sqI, qrsKexSIDH.sizeSqI);

	if (SecretAgreement_A(scratch, qRCopy, key, sidhData) != CRYPTO_SUCCESS) valid = false;
	return valid;
}

static qrKex qrsKexSIDH = {
		.sizePqI = 768,
		.sizeSqI = 48,
		.sizeQR = 768,
		.sizeKey = 192,
		.sizeScratch = 48,
		.sizeAlign8 = 8,
		.sizeAlign16 = 16,
		.sizeAlign24 = 24,

		.qrInit = qrSIDH_Init,
		.qrGenI = qrSIDH_GenI,
		.qrGenR = qrSIDH_GenR,
		.qrKeyI = qrSIDH_KeyI,
		.qrCleanup = qrSIDH_Cleanup,
};
qrKex* qrKexSIDH = &qrsKexSIDH;

#endif

#ifdef QRKEX_USE_NEWHOPE

static bool qrNewHope_Init(void) { return true; }

static void qrNewHope_Cleanup(void) {}

static bool qrNewHope_GenI(unsigned char* pqI, unsigned char* sqI, unsigned char* scratch) {
	(void)scratch;
    newhope_keygen(pqI, (poly*)sqI);
	return true;
}

static bool qrNewHope_GenR(const unsigned char* pqI, unsigned char* qR, unsigned char* key, unsigned char* scratch) {
	(void)scratch;
    newhope_sharedb(key, qR, pqI);
	return true;
}

static bool qrNewHope_KeyI(const unsigned char* sqI, const unsigned char* qR, unsigned char* key, unsigned char* scratch) {
	(void)scratch;
    newhope_shareda(key, (poly*)sqI, qR);
	return true;
}

static qrKex qrsKexNewHope = {
		.sizePqI = NEWHOPE_SENDABYTES,
		.sizeSqI = sizeof(poly),
		.sizeQR = NEWHOPE_SENDBBYTES,
		.sizeKey = 32,
		.sizeScratch = 0,
		.sizeAlign8 = 8,
		.sizeAlign16 = 16,
		.sizeAlign24 = 24,

		.qrInit = qrNewHope_Init,
		.qrGenI = qrNewHope_GenI,
		.qrGenR = qrNewHope_GenR,
		.qrKeyI = qrNewHope_KeyI,
		.qrCleanup = qrNewHope_Cleanup,
};
qrKex* qrKexNewHope = &qrsKexNewHope;

#endif
