#include <stdlib.h>

#include "spawnplus.h"
#include "crypto.h"
#include "ecc.h"
#include "ecelgamal.h"
#include "ecelgamal-internals.h"
#include "secmem.h"
#include "dre.h"
#include "auth.h"

typedef struct {
	ge25519 gi;
	// pqI immediately follows
} spawnplus_publicState;

typedef struct {
	bignum256modm i;
	// sqI immediately follows
} spawnplus_secretState;

static unsigned char implHashTag[] = "Spawn+ key exchange";
static size_t ihtLen = 20;

SPAWNPLUS_STATUS spawnplus_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk,
	unsigned char* scratch) {

	unsigned char buffer[sizeof(ecelgamal_pubkey)];
	ecelgamal_pubkey* public = (ecelgamal_pubkey*)buffer;
	ecelgamal_seckey* secret = (ecelgamal_seckey*)scratch;

	ecelgamal_genPackedKeys(sk, public, secret);
	ecelgamal_packPubKey(forgingPk, public);
	if (forgingSk != NULL) memcpy(forgingSk, sk, ECELGAMAL_SK_SIZE);

	ecelgamal_genPackedKeys(sk, public, secret);
	ecelgamal_packPubKey(pk, public);

	secureErase(scratch, SPAWNPLUS_SCRATCH_KEYGEN_SIZE);

	return SPAWNPLUS_SUCCESS;
}

SPAWNPLUS_STATUS spawnplus_1_makePreKey_I(const qrKex* qrKex,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* prekey,
	unsigned char* scratch) {

	unsigned char* prekeyGi = prekey;
	unsigned char* prekeyPqI = prekeyGi + 32;

	spawnplus_publicState* publicStateEl = (spawnplus_publicState*)publicState;
	spawnplus_secretState* secretStateEl = (spawnplus_secretState*)secretState;
	unsigned char* pqI = publicState + sizeof(spawnplus_publicState);
	unsigned char* sqI = secretState + sizeof(spawnplus_secretState) + qrKex->sizeAlign8;
	if (!qrKex->qrGenI(pqI, sqI, scratch)) return SPAWNPLUS_QRCALC_FAILURE;
	memcpy(prekeyPqI, pqI, qrKex->sizePqI);

	modm_rand(secretStateEl->i);

	ge25519_scalarmult_base1(&publicStateEl->gi, secretStateEl->i);
	ge25519_pack(prekeyGi, &publicStateEl->gi);

	return SPAWNPLUS_SUCCESS;
}

// Scratch space required: sizeof(ge25519) + 32
static bool spawnplus_key(const ge25519* dhPub, const bignum256modm dhSecret,
	const qrKex* qrKex, const unsigned char* qrKey,
	unsigned char* key,
	unsigned char* scratch) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	ge25519* dhKey = (ge25519*)scratch;
	ge25519_scalarmult(dhKey, dhPub, dhSecret);
	unsigned char* dhKeyPack = scratch + sizeof(ge25519);
	ge25519_pack(dhKeyPack, dhKey);

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, dhKeyPack, 32*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, qrKey, qrKex->sizeKey*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, key) == SUCCESS);

	secureErase(scratch, sizeof(ge25519) + 32);

	return hashSuccess;
}

static void spawnplus_makeDreAd(size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const ge25519* gi,
	const qrKex* qrKex,
	const unsigned char* pqI,
	unsigned char* buffer) {

	unsigned char* ptr = buffer;
	memcpy(ptr, idI, idLen);            ptr += idLen;
	memcpy(ptr, idR, idLen);            ptr += idLen;
	ge25519_pack(ptr, gi);              ptr += 32;
	memcpy(ptr, pqI, qrKex->sizePqI);   ptr += qrKex->sizePqI;
}

SPAWNPLUS_STATUS spawnplus_2_exchange_R(const unsigned char* prekey,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkI,
	const unsigned char* pkR, const unsigned char* forgingPkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch) {

	SPAWNPLUS_STATUS result = SPAWNPLUS_SUCCESS;

	size_t ptLen = (32 + qrKex->sizeQR);
	size_t dreAdLen = (2*idLen + 32 + qrKex->sizePqI);
	size_t ctLen = (DRE_CT_OVERHEAD + ptLen);
	size_t bufferLen = dreAdLen + MAX(ctLen, adLen);
	unsigned char buffer[bufferLen];

	const unsigned char* prekeyGi = prekey;
	const unsigned char* prekeyPqI = prekey + 32;

	unsigned char* sendIdR = send;
	unsigned char* sendCT = sendIdR + idLen;
	unsigned char* sendProof = sendCT + ctLen;

	bignum256modm_element_t* mr = (bignum256modm_element_t*)scratch;
	bignum256modm_element_t* signKey = (bignum256modm_element_t*)(scratch + sizeof(bignum256modm));
	unsigned char* scratchKey = scratch + 2*sizeof(bignum256modm);
	unsigned char* ptGr = scratchKey + sizeof(ge25519) + 32;
	unsigned char* ptQR = ptGr + 32;
	unsigned char* qrKey = ptQR + qrKex->sizeQR;
	unsigned char* qrScratch = qrKey + qrKex->sizeKey;

	ge25519 ALIGN(16) gi;
	if (!ge25519_unpack_vartime(&gi, prekeyGi)) {
		result = SPAWNPLUS_INVALID_MESSAGE;
		goto cleanup;
	}

	ecelgamal_pubkey eccrshPkI, eccrshFpkI, eccrshPkR, eccrshFpkR;
	ecelgamal_unpackPubKey(&eccrshPkI, pkI);
	ecelgamal_unpackPubKey(&eccrshFpkI, forgingPkI);
	ecelgamal_unpackPubKey(&eccrshPkR, pkR);
	ecelgamal_unpackPubKey(&eccrshFpkR, forgingPkR);

	memcpy(sendIdR, idR, idLen);

	if (!qrKex->qrGenR(prekeyPqI, ptQR, qrKey, qrScratch)) {
		result = SPAWNPLUS_QRCALC_FAILURE;
		goto cleanup;
	}

	modm_rand(mr);

	ge25519 ALIGN(16) gr;
	ge25519_scalarmult_base1(&gr, mr);
	ge25519_pack(ptGr, &gr);

	spawnplus_makeDreAd(idLen, idI, idR, &gi, qrKex, prekeyPqI, buffer);
	if (adLen > 0) memcpy(buffer + dreAdLen, associatedData, adLen);
	size_t ctUsed;
	if (!dre_encrypt(&eccrshPkI, &eccrshFpkR, ptGr, ptLen, buffer, dreAdLen + adLen, sendCT, ctLen, &ctUsed)) {
		result = SPAWNPLUS_CALC_FAILURE;
		goto cleanup;
	}

	memcpy(buffer + dreAdLen, sendCT, ctLen);

	expand256_modm(signKey, signSecret, 32);
	if (!auth_prove(ge25519_niels_base1_multiples, &eccrshFpkI.h, &eccrshPkR.h, &gi, signMode, signKey, buffer, dreAdLen + ctLen, associatedData, adLen, implHashTag, ihtLen, sendProof)) {
		result = SPAWNPLUS_CALC_FAILURE;
		goto cleanup;
	}

	if (!spawnplus_key(&gi, mr, qrKex, qrKey, key, scratchKey)) {
		result = SPAWNPLUS_CALC_FAILURE;
		goto cleanup;
	}

cleanup:
	if (result != SPAWNPLUS_SUCCESS) secureErase(key, SPAWNPLUS_KEY_SIZE);
	secureErase(scratch, SPAWNPLUS_SCRATCH_R_OVERHEAD + qrKex->sizeQR + qrKex->sizeKey + qrKex->sizeScratch);
	return result;
}

SPAWNPLUS_STATUS spawnplus_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
	return SPAWNPLUS_SUCCESS;
}

SPAWNPLUS_STATUS spawnplus_3_exchange_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI,
	const unsigned char* pkR, const unsigned char* forgingPkR,
	unsigned char finishMode, const unsigned char* finishSecret,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* key,
	unsigned char* scratch) {

	SPAWNPLUS_STATUS result = SPAWNPLUS_SUCCESS;

	size_t ptLen = (32 + qrKex->sizeQR);
	size_t dreAdLen = (2*idLen + 32 + qrKex->sizePqI);
	size_t ctLen = (DRE_CT_OVERHEAD + ptLen);
	size_t bufferLen = dreAdLen + MAX(ctLen, adLen);
	unsigned char buffer[bufferLen];

	const unsigned char* receivedIdR = received;
	const unsigned char* receivedCT = receivedIdR + idLen;
	const unsigned char* receivedProof = receivedCT + ctLen;

	unsigned char* keyScratch = scratch;
	ecelgamal_seckey* dreSk = (ecelgamal_seckey*)(keyScratch + sizeof(ge25519) + 32);
	ge25519* gr = (ge25519*)((unsigned char*)dreSk + sizeof(ecelgamal_seckey));
	unsigned char* ptGr = (unsigned char*)gr + sizeof(ge25519);
	unsigned char* ptQR = ptGr + 32;
	unsigned char* qrKey = ptQR + qrKex->sizeQR;
	unsigned char* qrScratch = qrKey + qrKex->sizeKey;

	ecelgamal_pubkey eccrshPkI, eccrshFpkI, eccrshPkR, eccrshFpkR;
	ecelgamal_unpackPubKey(&eccrshPkI, pkI);
	ecelgamal_unpackPubKey(&eccrshFpkI, forgingPkI);
	ecelgamal_unpackPubKey(&eccrshPkR, pkR);
	ecelgamal_unpackPubKey(&eccrshFpkR, forgingPkR);

	spawnplus_publicState* publicStateEl = (spawnplus_publicState*)publicState;
	spawnplus_secretState* secretStateEl = (spawnplus_secretState*)secretState;

	unsigned char* pqI = ((unsigned char*)publicState) + sizeof(spawnplus_publicState);

	spawnplus_makeDreAd(idLen, idI, receivedIdR, &publicStateEl->gi, qrKex, pqI, buffer);
	memcpy(buffer + dreAdLen, receivedCT, ctLen);
	if (!auth_verify(ge25519_niels_base1_multiples, &eccrshFpkI.h, &eccrshPkR.h, &publicStateEl->gi, receivedProof, buffer, dreAdLen + ctLen, associatedData, adLen, implHashTag, ihtLen)) {
		result = SPAWNPLUS_INVALID_MESSAGE;
		goto cleanup;
	}

	ecelgamal_unpackSecKey(dreSk, finishSecret);

	if (adLen > 0) memcpy(buffer + dreAdLen, associatedData, adLen);
	size_t ptUsed;
	if (!dre_decrypt(&eccrshPkI, &eccrshFpkR, finishMode, dreSk, receivedCT, ctLen, buffer, dreAdLen + adLen, ptGr, ptLen, &ptUsed)) {
		result = SPAWNPLUS_INVALID_MESSAGE;
		goto cleanup;
	}

	if (!ge25519_unpack_vartime(gr, ptGr)) {
		result = SPAWNPLUS_INVALID_MESSAGE;
		goto cleanup;
	}

	unsigned char* sqI = secretState + sizeof(spawnplus_secretState) + qrKex->sizeAlign8;
	if (!qrKex->qrKeyI(sqI, ptQR, qrKey, qrScratch)) {
		result = SPAWNPLUS_QRCALC_FAILURE;
		goto cleanup;
	}

	if (!spawnplus_key(gr, secretStateEl->i, qrKex, qrKey, key, keyScratch)) {
		result = SPAWNPLUS_CALC_FAILURE;
		goto cleanup;
	}

cleanup:
	if (result != SPAWNPLUS_SUCCESS) secureErase(key, SPAWNPLUS_KEY_SIZE);
	secureErase(scratch, SPAWNPLUS_SCRATCH_I_OVERHEAD + qrKex->sizeQR + qrKex->sizeKey + qrKex->sizeScratch);
	if (result == SPAWNPLUS_SUCCESS) secureErase(secretState, SPAWNPLUS_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI);
	return result;
}

void spawnplus_forgeOffline(const unsigned char* secretState, unsigned char* signSecret) {
	contract256_modm(signSecret, ((const spawnplus_secretState*)secretState)->i);
}
