#include <stdlib.h>

#include "xzdh.h"
#include "crypto.h"
#include "ecc.h"
#include "secmem.h"
#include "auth.h"
#include "kmac.h"

typedef struct {
	ed25519_public_key gi;
	// pqI immediately follows
} xzdh_publicOneTimeState;

typedef struct {
	unsigned char iSrc[32];
	// sqI immediately follows
} xzdh_secretOneTimeState;

static unsigned char implHashTag[] = "XZDH key exchange";
static unsigned char xzdhKmacTag[] = "XZDH KMAC";
static unsigned char xzdhKdf1Tag[] = "XZDH KDF1";
static unsigned char xzdhKdf2Tag[] = "XZDH KDF2";
static unsigned char xzdhKdf3Tag[] = "XZDH KDF3";
static size_t ihtLen = 18;
static size_t xzdhKmacTagLen = 10;
static size_t xzdhKdfTagLen = 10;

#define XZDH_KMAC_SIZE 32
#define XZDH_KMAC_KEY_SIZE 32

static void xzdh_extractDHSecret(bignum256modm dhSecret, const unsigned char* sk) {
	ed25519_secret_key ed25519sec;
	ed25519_secretkey(ed25519sec, sk);
	expand256_modm(dhSecret, ed25519sec, 32);
}

XZDH_STATUS xzdh_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk) {

	RAND_bytes(sk, 32);
	ed25519_publickey(sk, forgingPk);
	if (forgingSk != NULL) memcpy(forgingSk, sk, 32);

	RAND_bytes(sk, 32);
	ed25519_publickey(sk, pk);

	return XZDH_SUCCESS;
}

XZDH_STATUS xzdh_genSignedPrekey(const unsigned char* pk, const unsigned char* sk,
	unsigned char* publicSignedState,
	unsigned char* secretSignedState,
	unsigned char* signedPrekey) {

	bignum256modm_element_t* dhSecret = (bignum256modm_element_t*)secretSignedState;
	modm_rand(dhSecret);

	ge25519 ALIGN(16) dhPublic;
	ge25519_scalarmult_base_niels(&dhPublic, ge25519_niels_base_multiples, dhSecret);
	ge25519_pack(signedPrekey, &dhPublic);
	memcpy(publicSignedState, signedPrekey, 32);

	ed25519_sign(signedPrekey, 32, sk, pk, signedPrekey+32);

	return XZDH_SUCCESS;
}

XZDH_STATUS xzdh_1_makeOneTimePrekey_I(const qrKex* qrKex,
	unsigned char* publicOneTimeState,
	unsigned char* secretOneTimeState,
	unsigned char* oneTimePrekey,
	unsigned char* scratch) {

	unsigned char* prekeyGi = oneTimePrekey;
	unsigned char* prekeyPqI = prekeyGi + 32;

	xzdh_publicOneTimeState* publicState = (xzdh_publicOneTimeState*)publicOneTimeState;
	xzdh_secretOneTimeState* secretState = (xzdh_secretOneTimeState*)secretOneTimeState;
	unsigned char* pqI = publicOneTimeState + sizeof(xzdh_publicOneTimeState);
	unsigned char* sqI = secretOneTimeState + sizeof(xzdh_secretOneTimeState);
	if (!qrKex->qrGenI(pqI, sqI, scratch)) return XZDH_QRCALC_FAILURE;
	memcpy(prekeyPqI, pqI, qrKex->sizePqI);

	RAND_bytes(secretState->iSrc, 32);
	ed25519_publickey(secretState->iSrc, prekeyGi);
	memcpy(publicState->gi, prekeyGi, 32);

	return XZDH_SUCCESS;
}

XZDH_STATUS xzdh_2_verifySignedPrekey_R(const unsigned char* pkI, const unsigned char* signedPrekey) {
	if (ed25519_sign_open(signedPrekey, 32, pkI, signedPrekey+32) != 0) {
		return XZDH_INVALID_KEY;
	}
	return XZDH_SUCCESS;
}

static bool xzdh_kdf(unsigned char* key,
	const unsigned char* msg1, size_t msg1Len,
	const unsigned char* msg2, size_t msg2Len,
	const unsigned char* constTag, size_t constTagLen) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, constTag, constTagLen*8) == SUCCESS);
	if (msg1 != NULL) hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, msg1, msg1Len*8) == SUCCESS);
	if (msg2 != NULL) hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, msg2, msg2Len*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, key) == SUCCESS);

	return hashSuccess;
}

// Scratch space required: sizeof(ge25519) + 96
static bool xzdh_keys(const ge25519* dhPub1, const bignum256modm dhSecret1,
	const ge25519* dhPub2, const bignum256modm dhSecret2,
	const ge25519* dhPub3, const bignum256modm dhSecret3,
	const qrKex* qrKex, const unsigned char* qrKey,
	unsigned char* macKey,
	unsigned char* key,
	unsigned char* scratch) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	ge25519* dhKey = (ge25519*)scratch;
	unsigned char* dhKeyPack1 = scratch + sizeof(ge25519);
	unsigned char* dhKeyPack2 = dhKeyPack1 + 32;
	unsigned char* dhKeyPack3 = dhKeyPack2 + 32;
	unsigned char* kappa = scratch; // Intentional overlap

	ge25519_scalarmult(dhKey, dhPub1, dhSecret1);
	ge25519_pack(dhKeyPack1, dhKey);

	ge25519_scalarmult(dhKey, dhPub2, dhSecret2);
	ge25519_pack(dhKeyPack2, dhKey);

	ge25519_scalarmult(dhKey, dhPub3, dhSecret3);
	ge25519_pack(dhKeyPack3, dhKey);

	bool success = true;
	success = success && xzdh_kdf(kappa, dhKeyPack1, 32*3, qrKey, qrKex->sizeKey, xzdhKdf1Tag, xzdhKdfTagLen);
	success = success && xzdh_kdf(macKey, kappa, 32, NULL, 0, xzdhKdf2Tag, xzdhKdfTagLen);
	success = success && xzdh_kdf(key, kappa, 32, NULL, 0, xzdhKdf3Tag, xzdhKdfTagLen);

	secureErase(scratch, sizeof(ge25519) + 32*3 + qrKex->sizeKey);

	return success;
}

#define XZDH_TAG_OVERHEAD (3*32)

static void xzdh_makeTag(size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const ge25519* gi, const ge25519* gr,
	const qrKex* qrKex,
	const unsigned char* pqI, const unsigned char* qR,
	const ge25519* ggamma,
	unsigned char* tag) {

	unsigned char* ptr = tag;
	memcpy(ptr, idI, idLen);          ptr += idLen;
	memcpy(ptr, idR, idLen);          ptr += idLen;
	ge25519_pack(ptr, gi);            ptr += 32;
	ge25519_pack(ptr, gr);            ptr += 32;
	memcpy(ptr, pqI, qrKex->sizePqI); ptr += qrKex->sizePqI;
	memcpy(ptr, qR, qrKex->sizeQR);   ptr += qrKex->sizeQR;
	ge25519_pack(ptr, ggamma);        ptr += 32;
}

XZDH_STATUS xzdh_2_exchange_R(const unsigned char* signedPrekey, const unsigned char* oneTimePrekey,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch) {

	XZDH_STATUS result = XZDH_SUCCESS;

	const unsigned char* prekeyGi = oneTimePrekey;
	const unsigned char* prekeyPqI = oneTimePrekey + 32;

	unsigned char* sendIdR = send;
	unsigned char* sendGr = sendIdR + idLen;
	unsigned char* sendQR = sendGr + 32;
	unsigned char* sendMac = sendQR + qrKex->sizeQR;
	unsigned char* sendProof = sendMac + XZDH_KMAC_SIZE;

	unsigned char ALIGN(16) qrBuffer[qrKex->sizeQR];

	bignum256modm_element_t* mr = (bignum256modm_element_t*)scratch;
	bignum256modm_element_t* signKey = (bignum256modm_element_t*)(scratch + sizeof(bignum256modm));
	unsigned char* scratchKey = scratch + 2*sizeof(bignum256modm);
	unsigned char* qrKey = scratchKey + sizeof(ge25519) + 96;
	unsigned char* macKey = qrKey + qrKex->sizeKey;
	unsigned char* qrScratch = macKey + XZDH_KMAC_KEY_SIZE;

	size_t tagLen = XZDH_TAG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
	unsigned char* tag = malloc(tagLen + adLen);

	ge25519 ALIGN(16) gi;
	if (!ge25519_unpack_vartime(&gi, prekeyGi)) {
		result = XZDH_INVALID_MESSAGE;
		goto cleanup;
	}

	ge25519 ALIGN(16) gePkI, geFpkI, gePkR, geGamma;
	if (   !ge25519_unpack_vartime(&gePkI, pkI)
	    || !ge25519_unpack_vartime(&geFpkI, forgingPkI)
	    || !ge25519_unpack_vartime(&gePkR, pkR)
	    || !ge25519_unpack_vartime(&geGamma, signedPrekey)) {
		result = XZDH_INVALID_KEY;
		goto cleanup;
	}

	memcpy(sendIdR, idR, idLen);

	if (!qrKex->qrGenR(prekeyPqI, qrBuffer, qrKey, qrScratch)) {
		result = XZDH_QRCALC_FAILURE;
		goto cleanup;
	}
	memcpy(sendQR, qrBuffer, qrKex->sizeQR);

	modm_rand(mr);

	ge25519 ALIGN(16) gr;
	ge25519_scalarmult_base_niels(&gr, ge25519_niels_base_multiples, mr);
	ge25519_pack(sendGr, &gr);

	if (!xzdh_keys(&gi, mr, &geGamma, mr, &gePkI, mr, qrKex, qrKey, macKey, key, scratchKey)) {
		result = XZDH_CALC_FAILURE;
		goto cleanup;
	}

	xzdh_makeTag(idLen, idI, idR, &gi, &gr, qrKex, prekeyPqI, sendQR, &geGamma, tag);

	if (adLen > 0) memcpy(tag + tagLen, associatedData, adLen);
	if (!kmac256(sendMac, XZDH_KMAC_SIZE, macKey, XZDH_KMAC_KEY_SIZE, tag, tagLen + adLen, xzdhKmacTag, xzdhKmacTagLen)) {
		result = XZDH_CALC_FAILURE;
		goto cleanup;
	}

	xzdh_extractDHSecret(signKey, signSecret);
	if (!auth_prove(ge25519_niels_base_multiples, &geFpkI, &gePkR, &gi, signMode, signKey, tag, tagLen, associatedData, adLen, implHashTag, ihtLen, sendProof)) {
		result = XZDH_CALC_FAILURE;
		goto cleanup;
	}

cleanup:
	if (result != XZDH_SUCCESS) secureErase(key, XZDH_KEY_SIZE);
	secureErase(scratch, XZDH_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch);
	free(tag);
	return result;
}

XZDH_STATUS xzdh_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
	return XZDH_SUCCESS;
}

XZDH_STATUS xzdh_3_exchange_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	const unsigned char* skI,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicSignedState,
	const unsigned char* publicOneTimeState,
	const unsigned char* secretSignedState,
	unsigned char* secretOneTimeState,
	unsigned char* key,
	unsigned char* scratch) {

	XZDH_STATUS result = XZDH_SUCCESS;

	const unsigned char* receivedIdR = received;
	const unsigned char* receivedGr = receivedIdR + idLen;
	const unsigned char* receivedQR = receivedGr + 32;
	const unsigned char* receivedMac = receivedQR + qrKex->sizeQR;
	const unsigned char* receivedProof = receivedMac + XZDH_KMAC_SIZE;

	unsigned char* keyScratch = scratch;
	unsigned char* qrKey = keyScratch + sizeof(ge25519) + 96;
	unsigned char* macKey = qrKey + qrKex->sizeKey;
	bignum256modm_element_t* mskI = (bignum256modm_element_t*)(macKey + XZDH_KMAC_KEY_SIZE);
	bignum256modm_element_t* mi = (bignum256modm_element_t*)((unsigned char*)mskI + sizeof(bignum256modm) + 24);
	unsigned char* qrScratch = (unsigned char*)mi + sizeof(bignum256modm) + qrKex->sizeAlign8;

	unsigned char expectedMac[XZDH_KMAC_SIZE];

	ge25519 ALIGN(16) gePkI, geFpkI, gePkR, geGamma;
	if (   !ge25519_unpack_vartime(&gePkI, pkI)
	    || !ge25519_unpack_vartime(&geFpkI, forgingPkI)
	    || !ge25519_unpack_vartime(&gePkR, pkR)
	    || !ge25519_unpack_vartime(&geGamma, publicSignedState)) {
		result = XZDH_INVALID_KEY;
		goto cleanup;
	}

	size_t tagLen = XZDH_TAG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
	unsigned char* tag = malloc(tagLen + adLen);

	xzdh_publicOneTimeState* publicState = (xzdh_publicOneTimeState*)publicOneTimeState;
	xzdh_secretOneTimeState* secretState = (xzdh_secretOneTimeState*)secretOneTimeState;

	bignum256modm_element_t* gamma = (bignum256modm_element_t*)secretSignedState;

	unsigned char* pqI = ((unsigned char*)publicOneTimeState) + sizeof(xzdh_publicOneTimeState);

	ge25519 ALIGN(16) gi, gr;
	if (!ge25519_unpack_vartime(&gi, publicState->gi)) {
		result = XZDH_INVALID_KEY;
		goto cleanup;
	}
	if (!ge25519_unpack_vartime(&gr, receivedGr)) {
		result = XZDH_INVALID_MESSAGE;
		goto cleanup;
	}

	xzdh_makeTag(idLen, idI, receivedIdR, &gi, &gr, qrKex, pqI, receivedQR, &geGamma, tag);
	if (!auth_verify(ge25519_niels_base_multiples, &geFpkI, &gePkR, &gi, receivedProof, tag, tagLen, associatedData, adLen, implHashTag, ihtLen)) {
		result = XZDH_INVALID_MESSAGE;
		goto cleanup;
	}

	unsigned char* sqI = secretOneTimeState + sizeof(xzdh_secretOneTimeState);
	if (!qrKex->qrKeyI(sqI, receivedQR, qrKey, qrScratch)) {
		result = XZDH_QRCALC_FAILURE;
		goto cleanup;
	}

	xzdh_extractDHSecret(mskI, skI);
	xzdh_extractDHSecret(mi, secretState->iSrc);
	if (!xzdh_keys(&gr, mi, &gr, gamma, &gr, mskI, qrKex, qrKey, macKey, key, keyScratch)) {
		result = XZDH_CALC_FAILURE;
		goto cleanup;
	}

	if (adLen > 0) memcpy(tag + tagLen, associatedData, adLen);
	if (!kmac256(expectedMac, XZDH_KMAC_SIZE, macKey, XZDH_KMAC_KEY_SIZE, tag, tagLen + adLen, xzdhKmacTag, xzdhKmacTagLen)) {
		result = XZDH_CALC_FAILURE;
		goto cleanup;
	}
	if (memcmp(expectedMac, receivedMac, XZDH_KMAC_SIZE) != 0) {
		result = XZDH_INVALID_MESSAGE;
		goto cleanup;
	}

cleanup:
	if (result != XZDH_SUCCESS) secureErase(key, XZDH_KEY_SIZE);
	secureErase(scratch, XZDH_SCRATCH_I_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeKey + qrKex->sizeScratch);
	if (result == XZDH_SUCCESS) secureErase(secretState, XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI);
	return result;
}

void xzdh_forgeOffline(const unsigned char* secretOneTimeState, unsigned char* signSecret) {
	memcpy(signSecret, ((const xzdh_secretOneTimeState*)secretOneTimeState)->iSrc, 32);
}
