#pragma once

/*
 * This module implements the XZDH key exchange.
 *
 * A key exchange takes place between an initiator, I, and a responder R. XZDH
 * is designed to be used non-interactively; I creates uploads "prekeys" to a
 * central server. This upload consists of a "signed prekey", and multiple
 * "one-time prekeys". I erases the secret for the signed prekey on a roughly
 * weekly basis, and then uploads a new one. This weekly refresh is important
 * for forward secrecy. To send a message to I, R requests a "prekey bundle"
 * containing a signed prekey and a one-time prekey, and uses it to complete a
 * key exchange. This two-flow protocol makes it possible to immediately attach
 * a message. The messages can be cached by store-and-forward servers to allow
 * both parties to go offline during the exchange. It is also possible to use
 * the exchange interactively, which removes the need for I to lookup ephemeral
 * secrets associated with prekeys.
 *
 * xzdh_genLongTermKeys is used to generate long-term key pairs. It is assumed
 * that parties distribute long-term public keys to each other and authenticate
 * them somehow. Authentication *must* cover both the "normal" public keys and
 * the "forging" public keys. This authentication is beyond the scope of this
 * module. Parties are given caller-defined identifiers that are used to locate
 * matching long-term public keys. To simplify key distribution, callers may
 * choose to use long-term public keys and forging public keys AS identifiers,
 * rather than as table lookup indexes, at the cost of communication overhead.
 *
 * The "forging" keys produced by xzdh_genLongTermKeys can be used to forge key
 * exchanges with others when being coerced by a third party to engage in a
 * session (see below for details about how to perform these forgeries).
 * However, if a malicious party ever compromises a long-term forging secret
 * key, then that party can impersonate others to the owner of the key without
 * detection. To prevent this attack, passing forgingSk = NULL to
 * xzdh_genLongTermKeys will prevent this key from ever being generated, at
 * the cost of making the user vulnerable to coercion. In the event that the
 * forging key IS generated, it should be stored more securely than the "normal"
 * long-term secret key (e.g., offline, on paper, in a secure storage facility).
 *
 * Signed prekeys are generated using xzdh_genSignedPrekey. Each party should
 * maintain one active signed prekey on the central server alongside their
 * one-time prekeys. Each prekey bundle contains the signed prekey and a
 * one-time prekey. The server erases one-time prekeys once they have been
 * included in a bundle. Each party is responsible for erasing the secret state
 * associated with their signed prekey after uploading a new one.
 *
 * Given known long-term public keys and the distribution of a signed prekey for
 * I generated with xzdh_genSignedPrekey, a normal non-interactive exchange
 * looks like this:
 *   1) I calls xzdh_1_makeOneTimePrekey_I and uploads the one-time prekey to a
 *      central server. The server maintains multiple one-time prekeys for each
 *      user, and ensures that each prekey is only distributed to one user upon
 *      request.
 *   2) R downloads a prekey bundle containing the signed prekey and a one-time
 *      prekey for I from the central server.
 *   3) R looks up I's long-term public key using some external system.
 *   4) R verifies the signed prekey using xzdh_2_verifySignedPrekey_R.
 *   5) R calls xzdh_2_exchange_R and sends the message to I, along with an
 *      identifier for the signed and one-time prekeys it used. R now knows the
 *      shared session key, which it can use to attach an encrypted message. R
 *      can trust that only I can decrypt the message, assuming that it is
 *      delivered before I uploads a new signed prekey (otherwise, the message
 *      is permanently lost).
 *   6) I calls xzdh_3_receive_I to learn R's asserted identifier.
 *   7) I looks up R's asserted long-term public key using some external system.
 *   8) I looks up secret states associated with the prekeys used by R.
 *   9) I calls xzdh_3_exchange_I. I now trusts R's identity and it knows the
 *      shared session key.
 *
 * xzdh_2_exchange_R takes two parameters---"signMode" and "signSecret"---that
 * specify how to perform authentication. In the normal situation (where both I
 * and R are honest and truly are who they say they are), the following values
 * are used by R:
 *   - signMode = XZDH_HONEST.
 *   - signSecret = skR.
 *
 * For plausible deniability purposes, this module provides the ability to forge
 * communication transcripts. An "offline forger" (someone who knows neither skI
 * nor skR) can produce a false transcript in the following way:
 *   1) Download a legitimate signed prekey for I from the central server.
 *   2) Call xzdh_1_makeOneTimePrekey_I to produce a one-time prekey.
 *   3) Call xzdh_forgeOffline to extract signSecret.
 *   4) Call xzdh_2_exchange_R using XZDH_OFFLINE_FORGE and signSecret to
 *      produce the response message.
 *   4) The shared session key is returned by xzdh_2_exchange_R. It can be used
 *      to forge subsequent messages in higher-level protocols.
 *
 * It is also possible for I to act as an "online forger". If I is being coerced
 * by an external entity J, I can forge an exchange with R (without knowing skR)
 * in the following way:
 *   1) J provides a one-time prekey that I pretends to relay to R.
 *      (If J allows it, then I can call xzdh_1_makeOneTimePrekey_I normally
 *      instead.)
 *   2) I calls xzdh_2_exchange_R with signMode = XZDH_ONLINE_FORGE and
 *      signSecret = forgingSkI, then sends the message to J as if R sent it. I
 *      also receives the shared session key from this function. It can be used
 *      to forge subsequent messages in higher-level protocols.
 *
 * If R is being coerced by an external entity J, there is nothing that can be
 * done to simulate an interaction with I. This is a fundamental limitation of
 * XZDH. The Spawn+ and DAKEZ key exchanges offer this functionality. However,
 * Spawn+ suffers from the same weakness if J has access to legitimate prekeys
 * from I. DAKEZ provides this level of deniability, but cannot be used in
 * non-interactive settings.
 */

#include <stdlib.h>

#include "qr.h"

#define XZDH_PK_SIZE 32
#define XZDH_SK_SIZE 32
#define XZDH_SIGNED_PREKEY_SIZE 96
#define XZDH_SIGNED_PUBLICSTATE_SIZE 32
#define XZDH_SIGNED_SECRETSTATE_SIZE 40
#define XZDH_ONETIME_PREKEY_OVERHEAD 32
#define XZDH_RESP_OVERHEAD 256
#define XZDH_ONETIME_PUBLICSTATE_OVERHEAD 32
#define XZDH_ONETIME_SECRETSTATE_OVERHEAD 32
#define XZDH_SCRATCH_R_OVERHEAD 400
#define XZDH_SCRATCH_I_OVERHEAD 424
#define XZDH_KEY_SIZE 32

typedef enum {
	XZDH_SUCCESS,
	XZDH_CALC_FAILURE,
	XZDH_QRCALC_FAILURE,
	XZDH_INVALID_MESSAGE,
	XZDH_INVALID_KEY,
} XZDH_STATUS;

#define XZDH_HONEST 0x2
#define XZDH_OFFLINE_FORGE 0x4
#define XZDH_ONLINE_FORGE 0x1

/* Generates long-term key pairs. pk and forgingPk are public, while sk and
 * forgingSk are secret.
 * - pk and forgingPk must be XZDH_PK_SIZE.
 * - sk and forgingSk must be XZDH_SK_SIZE (but forgingSk may be NULL).
 * - If forgingSk is NULL, then the secret forging key is securely erased. See
 *   the comment at the top of the file for a discussion of the consequences.
 * - Callers should arrange for secure storage of sk and forgingSk (e.g., using
 *   secmem.h).
 */
XZDH_STATUS xzdh_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk);

/* Generates a XZDH signed prekey. Every user of the system should make one
 * signed prekey available through some central service as part of a prekey
 * bundle. On a roughly weekly basis, the user should securely erase the secret
 * state, generate a new signed prekey, and replace the active one in the
 * central service. Refreshing signed prekeys is important to protect forward
 * secrecy; any messages created using old prekeys will need to be re-sent, but
 * they will be unreadable to any network attacker.
 * - pk is the party's long-term public key.
 * - sk is the party's long-term secret key.
 * - publicSignedState holds public state associated with the signed prekey. It
 *   must point to a buffer of length XZDH_SIGNED_PUBLICSTATE_SIZE.
 * - secretSignedState holds the secret state associated with the signed prekey.
 *   This value is needed to decrypt messages sent using the prekey. It must
 *   point to a buffer of length XZDH_SIGNED_SECRETSTATE_SIZE. secretSignedState
 *   should be aligned on a 16 byte boundary.
 * - signedPrekey holds the signed prekey and must be of length
 *   XZDH_SIGNED_PREKEY_SIZE.
 * - Callers should arrange for secure storage of secretSignedState (e.g., using
 *   secmem.h).
 * - The caller is responsible for securely erasing secretSignedState when
 *   calling this function again to generate a new signed prekey.
 */
XZDH_STATUS xzdh_genSignedPrekey(const unsigned char* pk, const unsigned char* sk,
	unsigned char* publicSignedState,
	unsigned char* secretSignedState,
	unsigned char* signedPrekey);

/* Creates a XZDH one-time prekey. Called by I (initiator). A set of one-time
 * prekeys should be made available through some central service so that
 * responders can obtain a fresh one to complete an exchange.
 * - qrKex is the quantum-resistant key exchange to use.
 * - publicOneTimeState holds temporary public state until R responds. It must
 *   point to a buffer of length
 *   XZDH_ONETIME_PUBLICSTATE_OVERHEAD + qrKex->sizePqI.
 *   publicOneTimeState should be aligned on a 16 byte boundary.
 * - secretOneTimeState holds temporary secret state until R responds. It must
 *   point to a buffer of length
 *   XZDH_ONETIME_SECRETSTATE_OVERHEAD + qrKex->sizeSqI.
 *   secretState should be aligned on a 16 byte boundary.
 * - oneTimePrekey holds the one-time prekey and must be of length
 *   XZDH_ONETIME_PREKEY_OVERHEAD + qrKex->sizePqI.
 * - scratch is secure scratch space of length qrKex->sizeScratch.
 * - Callers should arrange for secure storage of secretState and scratch
 *   (e.g., using secmem.h).
 */
XZDH_STATUS xzdh_1_makeOneTimePrekey_I(const qrKex* qrKex,
	unsigned char* publicOneTimeState,
	unsigned char* secretOneTimeState,
	unsigned char* oneTimePrekey,
	unsigned char* scratch);

/* Verifies a signed prekey. Called by R (responder). After downloading a signed
 * prekey allegedly from I from the central service, R should call this function
 * to verify that the signed prekey was actually created by I. This must be done
 * before using the signed prekey as input to xzdh_2_exchange_R. This
 * verification only needs to be performed once per signed prekey; if the
 * verification is successful, it is safe to cache this result.
 * - pkI is the long-term public key for I.
 * - signedPrekey is the signed prekey allegedly produced by I.
 * - Returns XZDH_SUCCESS if the signed prekey is safe to use, or another status
 *   if the prekey should not be used.
 */
XZDH_STATUS xzdh_2_verifySignedPrekey_R(const unsigned char* pkI, const unsigned char* signedPrekey);

/* Called by R to complete a key exchange with I.
 * - signedPrekey is a signed prekey created by I. This signed prekey *MUST* be
 *   verified using xzdh_2_verifySignedPrekey_R before calling this message.
 * - oneTimePrekey is a one-time prekey created by I.
 * - signedPrekey and oneTimePrekey should be retrieved from some central
 *   service as part of a "prekey bundle" using I's identifier as a lookup
 *   value.
 * - qrKex is the quantum-resistant key exchange in use.
 * - idI and idR, both of length idLen, are identifiers for the parties.
 * - pkI, forgingPkI, and pkR are long-term keys for the parties.
 * - signMode and signSecret specify the key pair used for authentication; see
 *   the comment at the top of the file for details.
 * - associatedData, of length adLen, is any common protocol state known to both
 *   parties in the higher-level protocol. If the key exchange completes
 *   successfully after xzdh_3_exchange_I, then it is guaranteed that both
 *   parties passed the same value to xzdh_2_exchange_R and xzdh_3_exchange_I.
 * - send holds the message to transmit to I and must be of length
 *   XZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen. An identifier for the signed
 *   and one-time prekeys should be sent along with this message so that I can
 *   lookup the associated secret states.
 * - key holds the shared secret for the session and is of size XZDH_KEY_SIZE.
 *   There is no explicit confirmation that I derived the same secret.
 * - scratch is secure scratch space of length
 *   XZDH_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch, and must be
 *   aligned on a 16 byte boundary.
 * - Callers should arrange for secure storage of key and scratch (e.g., using
 *   secmem.h).
 */
XZDH_STATUS xzdh_2_exchange_R(const unsigned char* signedPrekey, const unsigned char* oneTimePrekey,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch);

/* Called by I after receiving a response message from R.
 * - This function returns R's asserted identifier in idR, which must be of
 *   length idLen.
 * - Call this before calling xzdh_3_exchange_I in order to decide whether or
 *   not to complete the connection.
 * - Note that idR is *not* trusted until the key exchange has completed with
 *   xzdh_3_exchange_I; it should only be used to locate the associated
 *   long-term public key for R.
 * - Independently, callers should ensure that I is able to locate the signed
 *   and one-time prekeys that R is responding to. The associated secret states
 *   will be needed to call xzdh_3_exchange_I.
 */
XZDH_STATUS xzdh_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR);

/* Called by I to complete a key exchange using a response message from R.
 * - qrKex is the quantum-resistant key exchange in use.
 * - idI, of length idLen, is the identifier for I.
 * - pkI, forgingPkI, and pkR are long-term keys for the parties.
 * - skI is I's long-term secret key.
 * - associatedData, of length adLen, is any common protocol state known to both
 *   parties in the higher-level protocol. If the key exchange completes
 *   successfully after xzdh_3_exchange_I, then it is guaranteed that both
 *   parties passed the same value to xzdh_2_exchange_R and xzdh_3_exchange_I.
 * - publicSignedState holds the public state allocated by xzdh_genSignedPrekey.
 * - publicOneTimeState holds the temporary public state allocated by
 *   xzdh_1_makeOneTimePrekey_I.
 * - secretSignedState holds the temporary secret state allocated by
 *   xzdh_genSignedPrekey.
 * - secretOneTimeState holds the temporary secret state allocated by
 *   xzdh_1_makeOneTimePrekey_I. It is securely erased if this function
 *   successfully returns the session key.
 * - key holds the shared secret for the session and is of size XZDH_KEY_SIZE.
 *   There is no explicit confirmation that R derived the same secret.
 * - scratch is secure scratch space of length
 *     XZDH_SCRATCH_I_OVERHEAD + qrKex->sizeAlign8 +
 *     qrKex->sizeKey + qrKex->sizeScratch,
 *   and must be aligned on a 16 byte boundary.
 * - Callers should arrange for secure storage of key and scratch (e.g., using
 *   secmem.h).
 */
XZDH_STATUS xzdh_3_exchange_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	const unsigned char* skI,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicSignedState,
	const unsigned char* publicOneTimeState,
	const unsigned char* secretSignedState,
	unsigned char* secretOneTimeState,
	unsigned char* key,
	unsigned char* scratch);

/* Extracts signing key used for an offline forgery.
 * - signSecret must be of length XZDH_SK_SIZE.
 * - To avoid leaving traces of the forgery, callers should use the same
 *   precautions for signSecret that they use for long-term secret keys
 *   (e.g., using secmem.h), and these values should be securely erased after
 *   the forgery.
 */
void xzdh_forgeOffline(const unsigned char* secretOneTimeState, unsigned char* signSecret);
