#include <stdlib.h>

#include "zdh.h"
#include "crypto.h"
#include "ecc.h"
#include "secmem.h"
#include "auth.h"
#include "kmac.h"

typedef struct {
	ge25519 gi;
	// pqI immediately follows
} zdh_publicState;

typedef struct {
	bignum256modm i;
	// sqI immediately follows
} zdh_secretState;

static unsigned char implHashTag[] = "ZDH key exchange";
static unsigned char zdhKmacTag[] = "ZDH KMAC";
static unsigned char zdhKdf1Tag[] = "ZDH KDF1";
static unsigned char zdhKdf2Tag[] = "ZDH KDF2";
static unsigned char zdhKdf3Tag[] = "ZDH KDF3";
static size_t ihtLen = 17;
static size_t zdhKmacTagLen = 9;
static size_t zdhKdfTagLen = 9;

#define ZDH_KMAC_SIZE 32
#define ZDH_KMAC_KEY_SIZE 32

ZDH_STATUS zdh_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk,
	unsigned char* scratch) {

	bignum256modm_element_t* mSk = (bignum256modm_element_t*)scratch;

	ge25519 ALIGN(16) gePk, geFpk;

	RAND_bytes(sk, 32);
	curved25519_make_secret(mSk, sk);
	ge25519_scalarmult_base1(&geFpk, mSk);
	ge25519_pack(forgingPk, &geFpk);
	if (forgingSk != NULL) memcpy(forgingSk, sk, 32);

	RAND_bytes(sk, 32);
	curved25519_make_secret(mSk, sk);
	ge25519_scalarmult_base1(&gePk, mSk);
	ge25519_pack(pk, &gePk);

	secureErase(scratch, sizeof(bignum256modm));

	return ZDH_SUCCESS;
}

ZDH_STATUS zdh_1_makePreKey_I(const qrKex* qrKex,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* prekey,
	unsigned char* scratch) {

	unsigned char* prekeyGi = prekey;
	unsigned char* prekeyPqI = prekeyGi + 32;

	zdh_publicState* publicStateEl = (zdh_publicState*)publicState;
	zdh_secretState* secretStateEl = (zdh_secretState*)secretState;
	unsigned char* pqI = publicState + sizeof(zdh_publicState);
	unsigned char* sqI = secretState + sizeof(zdh_secretState) + qrKex->sizeAlign8;
	if (!qrKex->qrGenI(pqI, sqI, scratch)) return ZDH_QRCALC_FAILURE;
	memcpy(prekeyPqI, pqI, qrKex->sizePqI);

	modm_rand(secretStateEl->i);

	ge25519_scalarmult_base1(&publicStateEl->gi, secretStateEl->i);
	ge25519_pack(prekeyGi, &publicStateEl->gi);

	return ZDH_SUCCESS;
}

static bool zdh_kdf(unsigned char* key,
	const unsigned char* msg1, size_t msg1Len,
	const unsigned char* msg2, size_t msg2Len,
	const unsigned char* constTag, size_t constTagLen) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	bool hashSuccess = true;
	hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, constTag, constTagLen*8) == SUCCESS);
	if (msg1 != NULL) hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, msg1, msg1Len*8) == SUCCESS);
	if (msg2 != NULL) hashSuccess = hashSuccess && (Keccak_HashUpdate(&hasher, msg2, msg2Len*8) == SUCCESS);
	hashSuccess = hashSuccess && (Keccak_HashFinal(&hasher, key) == SUCCESS);

	return hashSuccess;
}

// Scratch space required: sizeof(ge25519) + 64
static bool zdh_keys(const ge25519* dhPub1, const bignum256modm dhSecret1,
	const ge25519* dhPub2, const bignum256modm dhSecret2,
	const qrKex* qrKex, const unsigned char* qrKey,
	unsigned char* macKey,
	unsigned char* key,
	unsigned char* scratch) {

	Keccak_HashInstance hasher;
	Keccak_Must(Keccak_HashInitialize_SHA3_256(&hasher));

	ge25519* dhKey = (ge25519*)scratch;
	unsigned char* dhKeyPack1 = scratch + sizeof(ge25519);
	unsigned char* dhKeyPack2 = dhKeyPack1 + 32;
	unsigned char* kappa = scratch; // Intentional overlap

	ge25519_scalarmult(dhKey, dhPub1, dhSecret1);
	ge25519_pack(dhKeyPack1, dhKey);

	ge25519_scalarmult(dhKey, dhPub2, dhSecret2);
	ge25519_pack(dhKeyPack2, dhKey);

	bool success = true;
	success = success && zdh_kdf(kappa, dhKeyPack1, 32*2, qrKey, qrKex->sizeKey, zdhKdf1Tag, zdhKdfTagLen);
	success = success && zdh_kdf(macKey, kappa, 32, NULL, 0, zdhKdf2Tag, zdhKdfTagLen);
	success = success && zdh_kdf(key, kappa, 32, NULL, 0, zdhKdf3Tag, zdhKdfTagLen);

	secureErase(scratch, sizeof(ge25519) + 32*2 + qrKex->sizeKey);

	return success;
}

#define ZDH_TAG_OVERHEAD (2*32)

static void zdh_makeTag(size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const ge25519* gi, const ge25519* gr,
	const qrKex* qrKex,
	const unsigned char* pqI, const unsigned char* qR,
	unsigned char* tag) {

	unsigned char* ptr = tag;
	memcpy(ptr, idI, idLen);          ptr += idLen;
	memcpy(ptr, idR, idLen);          ptr += idLen;
	ge25519_pack(ptr, gi);            ptr += 32;
	ge25519_pack(ptr, gr);            ptr += 32;
	memcpy(ptr, pqI, qrKex->sizePqI); ptr += qrKex->sizePqI;
	memcpy(ptr, qR, qrKex->sizeQR);   ptr += qrKex->sizeQR;
}

ZDH_STATUS zdh_2_exchange_R(const unsigned char* prekey,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch) {

	ZDH_STATUS result = ZDH_SUCCESS;

	const unsigned char* prekeyGi = prekey;
	const unsigned char* prekeyPqI = prekey + 32;

	unsigned char* sendIdR = send;
	unsigned char* sendGr = sendIdR + idLen;
	unsigned char* sendQR = sendGr + 32;
	unsigned char* sendMac = sendQR + qrKex->sizeQR;
	unsigned char* sendProof = sendMac + ZDH_KMAC_SIZE;

	unsigned char ALIGN(16) qrBuffer[qrKex->sizeQR];

	bignum256modm_element_t* mr = (bignum256modm_element_t*)scratch;
	bignum256modm_element_t* signKey = (bignum256modm_element_t*)(scratch + sizeof(bignum256modm));
	unsigned char* scratchKey = scratch + 2*sizeof(bignum256modm);
	unsigned char* qrKey = scratchKey + sizeof(ge25519) + 64;
	unsigned char* macKey = qrKey + qrKex->sizeKey;
	unsigned char* qrScratch = macKey + ZDH_KMAC_KEY_SIZE;

	size_t tagLen = ZDH_TAG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
	unsigned char* tag = malloc(tagLen + adLen);

	ge25519 ALIGN(16) gi;
	if (!ge25519_unpack_vartime(&gi, prekeyGi)) {
		result = ZDH_INVALID_MESSAGE;
		goto cleanup;
	}

	ge25519 ALIGN(16) gePkI, geFpkI, gePkR;
	if (   !ge25519_unpack_vartime(&gePkI, pkI)
	    || !ge25519_unpack_vartime(&geFpkI, forgingPkI)
	    || !ge25519_unpack_vartime(&gePkR, pkR)) {
		result = ZDH_INVALID_KEY;
		goto cleanup;
	}

	memcpy(sendIdR, idR, idLen);

	if (!qrKex->qrGenR(prekeyPqI, qrBuffer, qrKey, qrScratch)) {
		result = ZDH_QRCALC_FAILURE;
		goto cleanup;
	}
	memcpy(sendQR, qrBuffer, qrKex->sizeQR);

	modm_rand(mr);

	ge25519 ALIGN(16) gr;
	ge25519_scalarmult_base1(&gr, mr);
	ge25519_pack(sendGr, &gr);

	if (!zdh_keys(&gi, mr, &gePkI, mr, qrKex, qrKey, macKey, key, scratchKey)) {
		result = ZDH_CALC_FAILURE;
		goto cleanup;
	}

	zdh_makeTag(idLen, idI, idR, &gi, &gr, qrKex, prekeyPqI, sendQR, tag);

	if (adLen > 0) memcpy(tag + tagLen, associatedData, adLen);
	if (!kmac256(sendMac, ZDH_KMAC_SIZE, macKey, ZDH_KMAC_KEY_SIZE, tag, tagLen + adLen, zdhKmacTag, zdhKmacTagLen)) {
		result = ZDH_CALC_FAILURE;
		goto cleanup;
	}

	expand256_modm(signKey, signSecret, 32);
	if (!auth_prove(ge25519_niels_base1_multiples, &geFpkI, &gePkR, &gi, signMode, signKey, tag, tagLen, associatedData, adLen, implHashTag, ihtLen, sendProof)) {
		result = ZDH_CALC_FAILURE;
		goto cleanup;
	}

cleanup:
	if (result != ZDH_SUCCESS) secureErase(key, ZDH_KEY_SIZE);
	secureErase(scratch, ZDH_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch);
	free(tag);
	return result;
}

ZDH_STATUS zdh_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR) {
	memcpy(idR, received, idLen);
	return ZDH_SUCCESS;
}

ZDH_STATUS zdh_3_exchange_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	const unsigned char* skI,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* key,
	unsigned char* scratch) {

	ZDH_STATUS result = ZDH_SUCCESS;

	const unsigned char* receivedIdR = received;
	const unsigned char* receivedGr = receivedIdR + idLen;
	const unsigned char* receivedQR = receivedGr + 32;
	const unsigned char* receivedMac = receivedQR + qrKex->sizeQR;
	const unsigned char* receivedProof = receivedMac + ZDH_KMAC_SIZE;

	unsigned char* keyScratch = scratch;
	unsigned char* qrKey = keyScratch + sizeof(ge25519) + 64;
	unsigned char* macKey = qrKey + qrKex->sizeKey;
	bignum256modm_element_t* mskI = (bignum256modm_element_t*)(macKey + ZDH_KMAC_KEY_SIZE);
	unsigned char* qrScratch = (unsigned char*)mskI + sizeof(bignum256modm);

	unsigned char expectedMac[ZDH_KMAC_SIZE];

	ge25519 ALIGN(16) gePkI, geFpkI, gePkR;
	if (   !ge25519_unpack_vartime(&gePkI, pkI)
	    || !ge25519_unpack_vartime(&geFpkI, forgingPkI)
	    || !ge25519_unpack_vartime(&gePkR, pkR)) {
		result = ZDH_INVALID_KEY;
		goto cleanup;
	}

	size_t tagLen = ZDH_TAG_OVERHEAD + 2*idLen + qrKex->sizePqI + qrKex->sizeQR;
	unsigned char* tag = malloc(tagLen + adLen);

	zdh_publicState* publicStateEl = (zdh_publicState*)publicState;
	zdh_secretState* secretStateEl = (zdh_secretState*)secretState;

	unsigned char* pqI = ((unsigned char*)publicState) + sizeof(zdh_publicState);

	ge25519 ALIGN(16) gr;
	if (!ge25519_unpack_vartime(&gr, receivedGr)) {
		result = ZDH_INVALID_MESSAGE;
		goto cleanup;
	}

	zdh_makeTag(idLen, idI, receivedIdR, &publicStateEl->gi, &gr, qrKex, pqI, receivedQR, tag);
	if (!auth_verify(ge25519_niels_base1_multiples, &geFpkI, &gePkR, &publicStateEl->gi, receivedProof, tag, tagLen, associatedData, adLen, implHashTag, ihtLen)) {
		result = ZDH_INVALID_MESSAGE;
		goto cleanup;
	}

	unsigned char* sqI = secretState + sizeof(zdh_secretState) + qrKex->sizeAlign8;
	if (!qrKex->qrKeyI(sqI, receivedQR, qrKey, qrScratch)) {
		result = ZDH_QRCALC_FAILURE;
		goto cleanup;
	}

	expand256_modm(mskI, skI, 32);
	if (!zdh_keys(&gr, secretStateEl->i, &gr, mskI, qrKex, qrKey, macKey, key, keyScratch)) {
		result = ZDH_CALC_FAILURE;
		goto cleanup;
	}

	if (adLen > 0) memcpy(tag + tagLen, associatedData, adLen);
	if (!kmac256(expectedMac, ZDH_KMAC_SIZE, macKey, ZDH_KMAC_KEY_SIZE, tag, tagLen + adLen, zdhKmacTag, zdhKmacTagLen)) {
		result = ZDH_CALC_FAILURE;
		goto cleanup;
	}
	if (memcmp(expectedMac, receivedMac, ZDH_KMAC_SIZE) != 0) {
		result = ZDH_INVALID_MESSAGE;
		goto cleanup;
	}

cleanup:
	if (result != ZDH_SUCCESS) secureErase(key, ZDH_KEY_SIZE);
	secureErase(scratch, ZDH_SCRATCH_I_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch);
	if (result == ZDH_SUCCESS) secureErase(secretState, ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI);
	return result;
}

void zdh_forgeOffline(const unsigned char* secretState, unsigned char* signSecret) {
	contract256_modm(signSecret, ((const zdh_secretState*)secretState)->i);
}
