#pragma once

/*
 * This module implements the ZDH key exchange.
 *
 * A key exchange takes place between an initiator, I, and a responder R. ZDH is
 * designed to be used non-interactively; I creates a "prekey" message that is
 * uploaded to a central server. Later, R requests a prekey and uses it to
 * complete a key exchange. This two-flow protocol makes it possible to
 * immediately attach a message. The messages can be cached by store-and-forward
 * servers to allow both parties to go offline during the exchange. It is also
 * possible to use the exchange interactively, which removes the need for I to
 * lookup ephemeral secrets associated with prekeys.
 *
 * zdh_genLongTermKeys is used to generate long-term key pairs. It is assumed
 * that parties distribute long-term public keys to each other and authenticate
 * them somehow. Authentication *must* cover both the "normal" public keys and
 * the "forging" public keys. This authentication is beyond the scope of this
 * module. Parties are given caller-defined identifiers that are used to locate
 * matching long-term public keys. To simplify key distribution, callers may
 * choose to use long-term public keys and forging public keys AS identifiers,
 * rather than as table lookup indexes, at the cost of communication overhead.
 *
 * The "forging" keys produced by zdh_genLongTermKeys can be used to forge key
 * exchanges with others when being coerced by a third party to engage in a
 * session (see below for details about how to perform these forgeries).
 * However, if a malicious party ever compromises a long-term forging secret
 * key, then that party can impersonate others to the owner of the key without
 * detection. To prevent this attack, passing forgingSk = NULL to
 * zdh_genLongTermKeys will prevent this key from ever being generated, at
 * the cost of making the user vulnerable to coercion. In the event that the
 * forging key IS generated, it should be stored more securely than the "normal"
 * long-term secret key (e.g., offline, on paper, in a secure storage facility).
 *
 * Given known long-term public keys, a normal non-interactive exchange looks
 * like this:
 *   1) I calls zsh_1_makePreKey_I and uploads the prekey to a central server.
 *      The server maintains multiple prekeys for each user, and ensures that
 *      each prekey is only distributed to one user upon request.
 *   2) R downloads a prekey for I from the central server.
 *   3) R looks up I's long-term public key using some external system.
 *   4) R calls zdh_2_exchange_R and sends the message to I, along with an
 *      identifier for the prekey it used. R now knows the shared session key,
 *      which it can use to attach an encrypted message. R can trust that only I
 *      can decrypt the message.
 *   5) I calls zdh_3_receive_I to learn R's asserted identifier.
 *   6) I looks up R's asserted long-term public key using some external system.
 *   7) I looks up secret state associated with the prekey used by R.
 *   8) I calls zdh_3_exchange_I. I now trusts R's identity and it knows the
 *      shared session key.
 *
 * zdh_2_exchange_R takes two parameters---"signMode" and "signSecret"---that
 * specify how to perform authentication. In the normal situation (where both I
 * and R are honest and truly are who they say they are), the following values
 * are used by R:
 *   - signMode = ZDH_HONEST.
 *   - signSecret = skR.
 *
 * For plausible deniability purposes, this module provides the ability to forge
 * communication transcripts. An "offline forger" (someone who knows neither skI
 * nor skR) can produce a false transcript in the following way:
 *   1) Call zdh_1_makePreKey_I to produce the prekey.
 *   2) Call zdh_forgeOffline to extract signSecret.
 *   3) Call zdh_2_exchange_R using ZDH_OFFLINE_FORGE and signSecret to produce
 *      the response message.
 *   4) The shared session key is returned by zdh_2_exchange_R. It can be used
 *      to forge subsequent messages in higher-level protocols.
 *
 * It is also possible for I to act as an "online forger". If I is being coerced
 * by an external entity J, I can forge an exchange with R (without knowing skR)
 * in the following way:
 *   1) J provides a prekey that I pretends to relay to R.
 *      (If J allows it, then I can call zdh_1_makePreKey_I normally instead.)
 *   2) I calls zdh_2_exchange_R with signMode = ZDH_ONLINE_FORGE and
 *      signSecret = forgingSkI, then sends the message to J as if R sent it. I
 *      also receives the shared session key from this function. It can be used
 *      to forge subsequent messages in higher-level protocols.
 *
 * If R is being coerced by an external entity J, there is nothing that can be
 * done to simulate an interaction with I. This is a fundamental limitation of
 * ZDH. The Spawn+ and DAKEZ key exchanges offer this functionality. However,
 * Spawn+ suffers from the same weakness if J has access to legitimate prekeys
 * from I. DAKEZ provides this level of deniability, but cannot be used in
 * non-interactive settings.
 */

#include <stdlib.h>

#include "qr.h"

#define ZDH_PK_SIZE 32
#define ZDH_SK_SIZE 32
#define ZDH_PREKEY_OVERHEAD 32
#define ZDH_RESP_OVERHEAD 256
#define ZDH_PUBLICSTATE_OVERHEAD 192
#define ZDH_SECRETSTATE_OVERHEAD 40
#define ZDH_SCRATCH_KEYGEN_SIZE 40
#define ZDH_SCRATCH_R_OVERHEAD 368
#define ZDH_SCRATCH_I_OVERHEAD 328
#define ZDH_KEY_SIZE 32

typedef enum {
	ZDH_SUCCESS,
	ZDH_CALC_FAILURE,
	ZDH_QRCALC_FAILURE,
	ZDH_INVALID_MESSAGE,
	ZDH_INVALID_KEY,
} ZDH_STATUS;

#define ZDH_HONEST 0x2
#define ZDH_OFFLINE_FORGE 0x4
#define ZDH_ONLINE_FORGE 0x1

/* Generates long-term key pairs. pk and forgingPk are public, while sk and
 * forgingSk are secret.
 * - pk and forgingPk must be ZDH_PK_SIZE.
 * - sk and forgingSk must be ZDH_SK_SIZE (but forgingSk may be NULL).
 * - If forgingSk is NULL, then the secret forging key is securely erased. See
 *   the comment at the top of the file for a discussion of the consequences.
 * - scratch is secure scratch space of length ZDH_SCRATCH_KEYGEN_SIZE.
 * - Callers should arrange for secure storage of sk, forgingSk, and scratch
 *   (e.g., using secmem.h).
 */
ZDH_STATUS zdh_genLongTermKeys(unsigned char* pk, unsigned char* sk,
	unsigned char* forgingPk, unsigned char* forgingSk,
	unsigned char* scratch);

/* Creates a ZDH prekey. Called by I (initiator). A set of prekeys should be
 * made available through some central service so that responders can obtain a
 * fresh one to complete an exchange.
 * - qrKex is the quantum-resistant key exchange to use.
 * - publicState holds temporary public state until R responds. It must point to
 *   a buffer of length ZDH_PUBLICSTATE_OVERHEAD + qrKex->sizePqI.
 *   publicState should be aligned on a 16 byte boundary.
 * - secretState holds temporary secret state until R responds. It must point to
 *   a buffer of length
 *   ZDH_SECRETSTATE_OVERHEAD + qrKex->sizeAlign8 + qrKex->sizeSqI.
 *   secretState should be aligned on a 16 byte boundary.
 * - prekey holds the prekey and must be of length
 *   ZDH_PREKEY_OVERHEAD + qrKex->sizePqI.
 * - scratch is secure scratch space of length qrKex->sizeScratch.
 * - Callers should arrange for secure storage of secretState and scratch
 *   (e.g., using secmem.h).
 */
ZDH_STATUS zdh_1_makePreKey_I(const qrKex* qrKex,
	unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* prekey,
	unsigned char* scratch);

/* Called by R to complete a key exchange with I.
 * - prekey is a prekey created by I. It should be retrieved from some central
 *   service using I's identifier as a lookup value.
 * - qrKex is the quantum-resistant key exchange in use.
 * - idI and idR, both of length idLen, are identifiers for the parties.
 * - pkI, forgingPkI, and pkR are long-term keys for the parties.
 * - signMode and signSecret specify the key pair used for authentication; see
 *   the comment at the top of the file for details.
 * - associatedData, of length adLen, is any common protocol state known to both
 *   parties in the higher-level protocol. If the key exchange completes
 *   successfully after zdh_3_exchange_I, then it is guaranteed that both
 *   parties passed the same value to zdh_2_exchange_R and zdh_3_exchange_I.
 * - send holds the message to transmit to I and must be of length
 *   ZDH_RESP_OVERHEAD + qrKex->sizeQR + idLen. An identifier for the prekey
 *   should be sent along with this message so that I can lookup the associated
 *   secret state.
 * - key holds the shared secret for the session and is of size ZDH_KEY_SIZE.
 *   There is no explicit confirmation that I derived the same secret.
 * - scratch is secure scratch space of length
 *   ZDH_SCRATCH_R_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch, and must be
 *   aligned on a 16 byte boundary.
 * - Callers should arrange for secure storage of key and scratch (e.g., using
 *   secmem.h).
 */
ZDH_STATUS zdh_2_exchange_R(const unsigned char* prekey,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI, const unsigned char* idR,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	unsigned char signMode, const unsigned char* signSecret,
	const unsigned char* associatedData, size_t adLen,
	unsigned char* send,
	unsigned char* key,
	unsigned char* scratch);

/* Called by I after receiving a response message from R.
 * - This function returns R's asserted identifier in idR, which must be of
 *   length idLen.
 * - Call this before calling zdh_3_exchange_I in order to decide whether or not
 *   to complete the connection.
 * - Note that idR is *not* trusted until the key exchange has completed with
 *   zdh_3_exchange_I; it should only be used to locate the associated long-term
 *   public key for R.
 * - Independently, callers should ensure that I is able to locate the prekey
 *   that R is responding to. The associated secret state will be needed to call
 *   zdh_3_exchange_I.
 */
ZDH_STATUS zdh_3_receive_I(const unsigned char* received, size_t idLen, unsigned char* idR);

/* Called by I to complete a key exchange using a response message from R.
 * - qrKex is the quantum-resistant key exchange in use.
 * - idI, of length idLen, is the identifier for I.
 * - pkI, forgingPkI, and pkR are long-term keys for the parties.
 * - skI is I's long-term secret key.
 * - associatedData, of length adLen, is any common protocol state known to both
 *   parties in the higher-level protocol. If the key exchange completes
 *   successfully after zdh_3_exchange_I, then it is guaranteed that both
 *   parties passed the same value to zdh_2_exchange_R and zdh_3_exchange_I.
 * - publicState holds the temporary public state allocated by
 *   zdh_1_makePreKey_I.
 * - secretState holds the temporary secret state allocated by
 *   zdh_1_makePreKey_I. It is securely erased if this function successfully
 *   returns the session key.
 * - key holds the shared secret for the session and is of size ZDH_KEY_SIZE.
 *   There is no explicit confirmation that R derived the same secret.
 * - scratch is secure scratch space of length
 *   ZDH_SCRATCH_I_OVERHEAD + qrKex->sizeKey + qrKex->sizeScratch, and must be
 *   aligned on a 16 byte boundary.
 * - Callers should arrange for secure storage of key and scratch (e.g., using
 *   secmem.h).
 */
ZDH_STATUS zdh_3_exchange_I(const unsigned char* received,
	const qrKex* qrKex,
	size_t idLen, const unsigned char* idI,
	const unsigned char* pkI, const unsigned char* forgingPkI, const unsigned char* pkR,
	const unsigned char* skI,
	const unsigned char* associatedData, size_t adLen,
	const unsigned char* publicState,
	unsigned char* secretState,
	unsigned char* key,
	unsigned char* scratch);

/* Extracts signing key used for an offline forgery.
 * - signSecret must be of length ZDH_SK_SIZE.
 * - To avoid leaving traces of the forgery, callers should use the same
 *   precautions for signSecret that they use for long-term secret keys
 *   (e.g., using secmem.h), and these values should be securely erased after
 *   the forgery.
 */
void zdh_forgeOffline(const unsigned char* secretState, unsigned char* signSecret);
